%{
/*
 *	Copyright (c) 1999 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*#define DEBUG_CALC*/

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#ifdef	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifdef	__STDC__
void yyerror(const char *fmt, ...);
#else
void yyerror();
#endif

static char *yyStart, *yyLast, *yyInput;
static double yyResult;

#ifdef	DEBUG_CALC
int syslog_initialized = 0;
#else
extern int syslog_initialized;
#endif

struct op_ent {
	char *token;
	int op;
};

%}

%union {
	double number;
	char *string;
}

%token <number> NUMBER
%token <string> STRING

%type <number> expr

%left OR AND
%left EQ NEQ GTE LTE '>' '<' '!'
%left '|' '&' '^' '~'
%left '+' '-'
%left '*' '/' '%'
%left UMINUS UNOT UINV	/* supplies precedence for unary minus, not, inverse */

%%	/* beginning of rules section */

spec	:	expr
		{
			yyResult = $1;
		}
	;

expr	:	'(' expr ')'
		{
			$$ = $2;
		}
	|	expr OR expr
		{
			$$ = $1 || $3;
		}
	|	expr AND expr
		{
			$$ = $1 && $3;
		}
	|	expr EQ expr
		{
			$$ = $1 == $3;
		}
	|	expr NEQ expr
		{
			$$ = $1 != $3;
		}
	|	expr GTE expr
		{
			$$ = $1 >= $3;
		}
	|	expr LTE expr
		{
			$$ = $1 <= $3;
		}
	|	expr '>' expr
		{
			$$ = $1 > $3;
		}
	|	expr '<' expr
		{
			$$ = $1 < $3;
		}
	|	expr '|' expr
		{
			register unsigned long l1 = $1, l2 = $3;
			$$ = l1 | l2;
		}
	|	expr '&' expr
		{
			register unsigned long l1 = $1, l2 = $3;
			$$ = l1 & l2;
		}
	|	expr '^' expr
		{
			register unsigned long l1 = $1, l2 = $3;
			$$ = l1 ^ l2;
		}
	|	expr '+' expr
		{
			$$ = $1 + $3;
		}
	|	expr '-' expr
		{
			$$ = $1 - $3;
		}
	|	expr '*' expr
		{
			$$ = $1 * $3;
		}
	|	expr '/' expr
		{
			$$ = $1 / $3;
		}
	|	expr '%' expr
		{
			register unsigned long l1 = $1, l2 = $3;
			$$ = l1 % l2;
		}
	|	'!' expr %prec UNOT
		{
			$$ = !$2;
		}
	|	'~' expr %prec UINV
		{
			register unsigned long l = $2;
			$$ = ~l;
		}
	|	'-' expr %prec UMINUS
		{
			$$ = -$2;
		}
	|	NUMBER
		{
			$$ = $1;
		}
	|	STRING
		{
			double value = 0;
			char *cp;
			for (cp = $1; *cp; cp++)
				value += (*cp << 8) | *(cp+1);
			$$ = value;
		}
	;

%%	/* beginning of subroutines section */

void
#ifdef	__STDC__
yyerror(const char *fmt, ...)
#else
yyerror(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;
#ifdef	__STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (!syslog_initialized) {
		fprintf(stderr, "\"%s\": at `%c': ", yyStart, *yyLast);
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		fflush(stderr);
	} else {
		char ebuf[1024];
		vsprintf(ebuf, fmt, ap);
		syslog(LOG_WARNING, "\"%s\": at `%c': %s", yyStart, *yyLast, ebuf);
	}
	va_end(ap);
}

static int
get_compose_op()
{
	int len;
	struct op_ent *p;
	static struct op_ent op_tab[] = {
		{ "==", EQ },
		{ "!=", NEQ },
		{ "||", OR },
		{ "&&", AND },
		{ ">=", GTE },
		{ "<=", LTE },
		{ 0 },
	};

	for (p = op_tab; p && p->token; p++) {
		len = strlen(p->token);
		if (!strncmp(p->token, yyInput, len)) {
			yyInput += len;
			return p->op;
		}
	}
	return -1;
}

int
yylex()
{
	register char *cp;
	int op;
	static char buf[100];

	/* skip blanks */
	while (isascii(*yyInput) && isspace(*yyInput)) yyInput++;

	yyLast = yyInput;

	/* is there a number? */
	for (cp = yyInput; *cp == '.' || isdigit(*cp); cp++);
	if (cp > yyInput) {
		memcpy(buf, yyInput, cp - yyInput);
		buf[cp - yyInput] = '\0';
		yyInput = cp;
		if (buf[0] == '.' && buf[1] == '\0')
			return buf[0];	/* unmarried dot */
		yylval.number = (double)atof(buf);
		return NUMBER;
	}

	/* is there a string? */
	if (*yyInput == '"') {
		for (cp = ++yyInput, op = 0; *cp && op < sizeof(buf)-1; cp++, op++) {
			if (*cp == '"') {
				cp++;
				break;
			} else if (*cp == '\\') {
				cp++;
				switch(*cp) {
				case 'r': buf[op] = '\r'; break;
				case 'n': buf[op] = '\n'; break;
				case 't': buf[op] = '\t'; break;
				case '\n': buf[op] = '\\'; break;
				default:
					buf[op++] = '\\';
					buf[op] = *cp;
				}
			} else buf[op] = *cp;
		}
		buf[op] = '\0';
		yyInput = cp;
		yylval.string = buf;
		return STRING;
	}

	/* is there a Unused variable? */
	if (!strncmp(yyInput, "Unused", 6)) {
		yyInput += 6;
		yylval.number = 0;
		return NUMBER;
	}

	if ((op = get_compose_op()) != -1)
		return op;

	return (*yyInput++);
}

double
calculate(expression)
	char *expression;
{
	yyStart = yyLast = yyInput = expression;
	yyResult = 0;

	return (yyparse() ? 0 : yyResult);
}

#ifdef	DEBUG_CALC

int
main(argc, argv)
	int argc;
	char *argv[];
{
	if (argc != 2) return;

	printf("%g\n", calculate(argv[1]));
}

#endif
