/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <arpa/nameser.h>
#ifdef	HAVE_RESOLV_H
#include <resolv.h>
#endif
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"

extern int errno;

static void dns_reply __P((int, SESSION *, int));

static int
dns_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	u_char buf[PACKETSZ];
	HEADER *hp = (HEADER *)buf;
	int len;
	u_char *cp, *dname, *dnptrs[20], **dpp, **lastdnptr;

/*	dprintf(("dns_send\n"));
*/
	memset(hp, 0, HFIXEDSZ);
	hp->id = htons(request->reqid);
	hp->opcode = QUERY;
	hp->rcode = NOERROR;

	cp = buf + HFIXEDSZ;
	len = PACKETSZ - (HFIXEDSZ + QFIXEDSZ);
	dpp = dnptrs;
	*dpp++ = buf;
	*dpp = NULL;
	lastdnptr = dnptrs + sizeof(dnptrs) / sizeof(dnptrs[0]);

	dname = (u_char *)sd->method->argument;
	if (!dname || !*dname) dname = (u_char *)".";

	if ((len = dn_comp((const char *)dname, cp, len, dnptrs, lastdnptr)) < 0)
		return -1;

	cp += len;
	if (!strcmp((const char *)dname, "."))
		len = T_A;
	else	len = T_SOA;
	PUTSHORT(len, cp);
	PUTSHORT(C_IN, cp);

	hp->qdcount = htons(1);
	len = cp - buf;

	len = send(sd->sock, (char *)buf, len, 0);
	if (len < 0 && errno == EINPROGRESS)
		len = 0;
	return len;
}

static int
dns_recv(sd)
	SESSION *sd;
{
	REQUEST *request;
	u_char buf[PACKETSZ * 2];
	HEADER *hp = (HEADER *)buf;
	struct sockaddr_in from;
	int len, tmpval;
	socklen_t from_len;

/*	dprintf(("dns_recv\n"));
*/
	if ((request = get_request(sd, 0 /* no reqid yet */)) == NULL) {
		trace(("\t dns_recv: internal error?\n"));
		return -1;
	}

	from_len = sizeof(from);
	if ((len = recvfrom(sd->sock, (char *)buf, sizeof(buf), 0,
			    (struct sockaddr *)&from, &from_len)) < 0) {
		if (errno != EAGAIN)
			dns_reply(errno, sd, request->reqid);
		return -1;
	}

	if (len < HFIXEDSZ) {
		trace(("\t dns_recv: undersized packet, len=%d\n", len));
		return -1;
	}
	if (hp->id != htons(request->reqid)) {
		trace(("\t dns_recv: invalid reqid\n"));
		return -1;
	}
	if (!hp->qr) {
		trace(("\t dns_recv: not a response\n"));
		return -1;
	}
	if (hp->opcode != QUERY) {
		trace(("\t dns_recv: response not a QUERY\n"));
		return -1;
	}
	if (hp->qdcount != htons(1)) {
		trace(("\t dns_recv: qdcount missmatched\n"));
		return -1;
	}

	if (sd->data_ptr) free(sd->data_ptr);
	if ((sd->data_ptr = (char *)malloc(len)) != NULL) {
		memmove(sd->data_ptr, buf, len);
		sd->data_int = len;
	} else 	sd->data_int = 0;

	return 0;
}

static void
dns_stop(method)
	METHOD *method;
{
	session_stop(method->sd, 1);
}

static void
dns_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int reqid;

	/* sanity check */
	if (!sd) return;

	dns_stop(method); /* stop the method if still running */

	/* open datagram socket near before UDP session */
	if ((sd->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		dns_reply(errno, sd, 0);
		return;
	}

	/* turn on non-blocking I/O */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		dns_reply(errno, sd, 0);
		return;
	}

	/*
	 * Make datagram socket `connected', so error on this socket
	 * will be returned asynchronously without timing out.
	 */
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		dns_reply(errno, sd, 0);
		return;
	}

	if ((reqid = session_start(sd, NULL)) != 0) {
		dprintf(("dns_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, reqid));
	}
}

static char *dns_error[] = {
	/* standart return codes */
	"No error",		/* NOERROR */
	"Format error",		/* FORMERR */
	"Server failure",	/* SERVFAIL */
	"Non existend domain",	/* NXDOAMIN */
	"Not implemented",	/* NOTIMP */
	"Query refused",	/* REFUSED */

	/* private return codes */
#define	ERR_TRUNCATED		(REFUSED + 1)
	"Truncated answer",
#define	ERR_UNEXP_END		(REFUSED + 2)
	"Unexpected end of answer",
#define	ERR_BAD_CLASS		(REFUSED + 3)
	"Bad class in answer",
#define	ERR_BAD_TYPE		(REFUSED + 4)
	"Bad type in answer",
#define	ERR_NON_AUTH		(REFUSED + 5)
	"Non-authoritative answer",
#define	ERR_INVALID		(REFUSED + 6)
	"Invalid answer"
#define	DNS_ERROR_LAST		ERR_INVALID
};

static int
dns_parse(packet, length)
	u_char *packet;
	int length;
{
	HEADER *hp = (HEADER *)packet;
	u_char *cp = packet + HFIXEDSZ;
	int n, non_auth = 0, no_answer = 0, qdcount = ntohs(hp->qdcount);
	u_short type, class;
	char name[MAXDNAME+1];

	if (hp->tc)
		return ERR_TRUNCATED;

	while (qdcount-- > 0) {
		if ((n = dn_expand(packet, packet + length, cp, name, sizeof(name))) < 0)
			return ERR_UNEXP_END;
		cp += n;
		if (cp + 2 * INT16SZ > packet + length)
			return ERR_UNEXP_END;
		GETSHORT(type, cp);
		GETSHORT(class, cp);
		if (class != C_IN)
			return ERR_BAD_CLASS;
		if (type == T_SOA) {
			if (!hp->aa) non_auth++;
			if (!hp->ancount) no_answer++;
		} else if (type != T_A)
			return ERR_BAD_TYPE;
	}
	/* need more checking here? */

	if (hp->rcode != NOERROR)
		return hp->rcode;
	if (non_auth)
		return ERR_NON_AUTH;
	if (no_answer)
		return ERR_INVALID;
	return NOERROR;
}

static void
dns_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[200];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		op = dns_parse(sd->data_ptr, sd->data_int);

		if (op >= NOERROR && op <= DNS_ERROR_LAST)
			diag = dns_error[op];
		else	diag = "Unknown error";
	} else	diag = strerror(op);

	if (sd->method->argument && *sd->method->argument) {
		sprintf(buf, "%s (%s)", diag, sd->method->argument);
		diag = buf;
	}

	dprintf(("dns_reply(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	dns_stop(method);

	method_finished(target, method, diag, !op);
}

static int
dns_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("dns_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = dns_send;
	template.recv = dns_recv;
	template.read = dns_reply;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_dns = {
	dns_init,	/* initialize dns protocol session */
	dns_start,	/* start dns session */
	dns_stop,	/* stop dns session */

	"DNS",		/* method name */
	".",		/* get root servers by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_UDP,	/* network protocol */
	NAMESERVER_PORT, /* server port */
	0, 0,		/* timeout and retries undefined yet */
	{ 0, 0 },	/* no parameters used */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
