/*
 *	Copyright (c) 1999 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"

extern CONFIG *cf;

static EVENT *first_event = NULL;	/* first system event in table */

/*
 * Subtract 2 timeval structs: out = out - in. Out result always greater 0.
 */
void
tv_sub(out, in)
	register TIMEVAL *out, *in;
{
	if (out->tv_sec > in->tv_sec) {
		out->tv_sec -= in->tv_sec;
		if (out->tv_usec >= in->tv_usec) {
			out->tv_usec -= in->tv_usec;
		} else {
			out->tv_sec--;
			out->tv_usec += 1000000L - in->tv_usec;
		}
	} else if (out->tv_sec == in->tv_sec) {
		out->tv_sec = 0L;
		out->tv_usec -= in->tv_usec;
		if (out->tv_usec < 0)
			out->tv_usec = 100L;	/* not zero anyway */
	} else {
		out->tv_sec = 0L;
		out->tv_usec = 100L;	/* not zero anyway */
	}
}

void
tv_add(out, in)
	register TIMEVAL *out, *in;
{
	out->tv_sec += in->tv_sec;
	out->tv_usec += in->tv_usec;
	if (out->tv_usec >= 1000000L) {
		out->tv_sec++;
		out->tv_usec -= 1000000L;
	}
}

u_long
tv_round(tvp)
	TIMEVAL *tvp;
{
	u_long sec = tvp->tv_sec;
	if (tvp->tv_usec >= 500000L)
		sec++;
	return sec;
}

/*
 * Execute pending event and schedule the next nearest.
 */
int
select_event(tvp)
	TIMEVAL *tvp;
{
	EVENT *ep, *next_event;
	TIMEVAL now, gap, earliest;

	gettimeofday(&now, NULL);
	gap.tv_sec = 0;
	gap.tv_usec = 100000L; /* 0.1sec grip */
	tv_add(&gap, &now);
again:
	next_event = NULL;
	timerclear(&earliest);

	for (ep = first_event; ep; ep = ep->next) {
		/* skip over the empty slots */
		if (ep->func == NULL) continue;

		if (timercmp(&gap, &ep->tv, >)) {
			void (*func)() = ep->func;
			ep->func = NULL;        /* free event slot before */

			dprintf(("-call_event(%p/%p)\n", func, ep->arg));

			(*func)(ep->arg);       /* call event function */
			goto again;

		} else if (!timerisset(&earliest) ||
			   timercmp(&ep->tv, &earliest, <)) {
			earliest = ep->tv;
			next_event = ep;
		}
	}

	if (!next_event) {	/* no more awaiting events */

		dprintf(("select_event: no timeout\n"));

		return 1;	/* timeout undefined */
	}

	tv_sub(&earliest, &now);
	*tvp = earliest;

	dprintf(("=wait_event(%p/%p): timeout=%d.%03d\n",
		 next_event->func, next_event->arg,
		 tvp->tv_sec, tvp->tv_usec / 1000));

	return 0;	/* timeout defined */
}

/*
 * Add the new system event to be executed at the given time.
 */
void
add_event(tvp, func, arg)
	TIMEVAL *tvp;
	void (*func)(void *);
	void *arg;
{
	register EVENT *ep;
	EVENT *prev = NULL, *next = NULL;
#ifdef	DEBUG
	char at_time[50];
	strftime(at_time, sizeof(at_time), "%T", localtime(&tvp->tv_sec));
#endif

	/*
	 * The same event in queue may cause a looping! Prevent it.
	 */
	for (ep = first_event; ep; ep = ep->next) {
		if (ep->func == func && ep->arg == arg) {
			dprintf(("?add_event(%p/%p): already awaiting!\n", func, arg));
			return;
		}
	}

	/*
	 * Search for first empty or last event slot.
	 */
	for (ep = first_event; ep; ep = ep->next) {
		if (ep->func == NULL) {
			next = ep->next;
			break;
		}
		prev = ep;
	}
	if (!ep && (ep = (EVENT *)malloc(sizeof(EVENT))) == NULL) {
		report(LOG_ERR, "add_event: Out of memory");
		return;
	}
	memset(ep, 0, sizeof(EVENT));
	ep->tv = *tvp;
	ep->func = func;
	ep->arg = arg;
	if (next) ep->next = next;
	else if (prev) prev->next = ep;
	if (!first_event) first_event = ep;

	dprintf(("+add_event(%p/%p): schedule=%s.%03d\n",
		 func, arg, at_time, tvp->tv_usec / 1000));
}

/*
 * Remove system event from queue if any.
 */
int
remove_event(func, arg)
	void (*func)(void *);
	void *arg;
{
	int found = 0;
	EVENT *ep;

/*	dprintf(("remove_event(%p/%p)\n", func, arg));
*/
	for (ep = first_event; ep; ep = ep->next) {
		if ((!func || ep->func == func) && ep->arg == arg) {
			ep->func = NULL;
			found++;
		}
	}
	return found;
}

/*
 * Modify existing system event in queue for the new function argument.
 */
int
change_event(func, arg, new_arg)
	void (*func)(void *);
	void *arg, *new_arg;
{
	int found = 0;
	EVENT *ep;

/*	dprintf(("change_event(%p/%p -> %p/%p)\n", func, arg, func, new_arg));
*/
	for (ep = first_event; ep; ep = ep->next) {
		if ((!func || ep->func == func) && ep->arg == arg) {
			ep->arg = new_arg;
			found++;
		}
	}
	return found;
}

void
free_events()
{
	EVENT *ep, *next;

	ep = first_event;
	while (ep) {
		next = ep->next;
		free(ep);
		ep = next;
	}
	first_event = NULL;
}

/*
 * Return time offset for the object.
 */
int
object_offset(object, period)
	OBJECT *object;
	int period;	/* seconds */
{
	OBJECT *objp;
	int off, curr, all;

	/* find current host index and number of all hosts */
	off = curr = all = 0;
	for (objp = cf->target; objp; objp = objp->next) {
		if (!off) {
			if (objp != object) curr++;
			else off = 1;
		}
		all++;
	}
	if (!all) return 0;	/* should not happen */

	/* make object time_offset depending of current network topology */
	if (object->hop_count > 0)
		curr = object->hop_count - 1;

	off = period / all;
	if (off < 1) off = 1;
	return (off * curr);
}

/*
 * Shift the time to be sharp at (12am + N * period), local time.
 */
void
tv_sharp(tvp, period, off)
	TIMEVAL *tvp;
	int period, off;
{
	struct tm *tm;
	time_t defect;

#ifdef	HAVE_PTHREAD
	pthread_mutex_lock(&localtime_lock);
#endif
	tm = localtime(&tvp->tv_sec);
	defect = tm->tm_sec + 60 * tm->tm_min + 3600 * tm->tm_hour + off;
#ifdef	HAVE_PTHREAD
	pthread_mutex_unlock(&localtime_lock);
#endif
	defect %= period;
	period -= defect;
	if (period < 1) period = 1;
	tvp->tv_sec += period;
	tvp->tv_usec = (long)random() % 1000000L;
}
