/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_PATHS_H
#include <paths.h>
#endif

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"
#include "mib.h"

char copyright[] = "Copyright (c) 1999-2004 Rinet Corp., Novosibirsk, Russia";

/*
 * Forward declarations.
 */

void mainloop __P((sigset_t *));
void vers __P((void));
void usage __P((void));

static void cleanup_handler __P((int));
static void childend_handler __P((int));
static void reconfig_handler __P((int));
static void brokenpipe_handler __P((int));
static void watchdog_handler __P((int));

/*
 * Global varibales.
 */
char *program_name;		/* my self name */
char *config_file;		/* netmond config file */
CONFIG *cf;			/* current configuration */
int Cflag;			/* command line switches */
pid_t mypid;			/* my self PID */
int syslog_initialized;		/* syslog ready to use */
int syslog_facility;		/* current syslog facility */

#ifdef	HAVE_PTHREAD
pthread_t main_thr;
int busy_threads;
pthread_attr_t detached_attr;
pthread_attr_t joinable_attr;
pthread_mutex_t gethostby_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t localtime_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t variables_lock = PTHREAD_MUTEX_INITIALIZER;
#else
static char **Argv;		/* pointer to argument vector */
static char *LastArgv;		/* end of argv */
#endif

static int reconfig_pending;
static int watchdog_timeout;
static int watchdog_pending;

static struct sighandler_ent {
	int sig;
	int flags;
	void (*handler)();
} sighandler_tab[] = {
	{ SIGCHLD,	(SA_RESTART | SA_NOCLDSTOP),	childend_handler },
#ifdef	PTHREAD_SIGNALS_BROKEN
	{ SIGHUP,	SA_RESTART,			reconfig_handler },
	{ SIGINT,	SA_RESTART,			reconfig_handler },
#else
	{ SIGHUP,	0,				reconfig_handler },
	{ SIGINT,	0,				reconfig_handler },
#endif
	{ SIGPIPE,	0,				brokenpipe_handler },
	{ SIGTERM,	0,				cleanup_handler },
	{ SIGALRM,	SA_RESTART,			watchdog_handler },
	{ 0 },
};

int
main(argc, argv, envp)
	int argc;
	char **argv, **envp;
{
	int op;
	char *cp, buf[1024];
	void *parsed_cf;
	FILE *fp;
	struct sighandler_ent *sp;
	sigset_t sigmask;
	struct sigaction sigact;
#ifdef	HAVE_PTHREAD
	struct sched_param param;
#endif
	extern int optind, opterr;
	extern char *optarg;

	/*
	 * Set the variables to it default values.
	 */
	program_name = strippath(argv[0]);
	config_file = DEFAULT_CONFIG;
	syslog_initialized = 0;
	syslog_facility = 0;
	Cflag = 0;
	reconfig_pending = 0;
	cf = NULL;
	watchdog_timeout = DEFAULT_WATCHDOG;
	watchdog_pending = 0;

	/*
	 * Go through command line switches and arguments.
	 */
	opterr = 0;
	while ((op = getopt(argc, argv, "c:CMw:vh?")) != EOF) {
		switch (op) {
		case 'c':	/* config file */
			config_file = optarg;
			break;
		case 'C':	/* parse config and exit */
			Cflag++;
			break;
		case 'M':	/* dump MIB tables */
			dump_mib_table(0); /* all tables */
			exit(0);
		case 'w':	/* watchdog timer */
			watchdog_timeout = atoi(optarg);
			break;
		case 'v':
			vers();
		case 'h':
		case '?':
		default:
			usage();
		}
	}
	cp = NULL;
	if (argv[optind] != NULL) {
		;	/* nope yet */
#ifdef	DEBUG
		cp = argv[optind];
#endif
	}

	/* restrict access for `other' */
	(void)umask(027);

	/*
	 * Initialize configurations.
	 */
	if (*config_file != '/') {
		if (getcwd(buf, sizeof(buf)) == NULL) {
			report(LOG_ERR, "getcwd: %m");
			exit(1);
		}
		(void)strcat(buf, "/");
		(void)strcat(buf, config_file);
		config_file = strdup(buf);
	}
	if ((parsed_cf = parse_config((void *)reconfig_pending)) == NULL)
		exit(1);
	if ((cf = (CONFIG *)malloc(sizeof(CONFIG))) == NULL) {
		report(LOG_ERR, "malloc: Out of memory");
		exit(1);
	}
	memcpy(cf, parsed_cf, sizeof(CONFIG));

#ifdef	DEBUG_MEMORY
	set_cksum();
#endif
	if (Cflag) {
		dump_config(cf);
		if (cp) printf("%s->%s\n", cp,
			       insert_variables(buf, sizeof(buf), cp, 0, NULL));
		exit(0);
	}

#ifdef	HAVE_PTHREAD
	/*
	 * Initialize threads.
	 */
	busy_threads = 0;
	main_thr = pthread_self();
	pthread_getschedparam(main_thr, &op, &param);

	pthread_attr_init(&joinable_attr);
	pthread_attr_setdetachstate(&joinable_attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setschedparam(&joinable_attr, &param);

	pthread_attr_init(&detached_attr);
	pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);
	param.sched_priority /= 2; /* lower priority for detached childs */
	pthread_attr_setschedparam(&detached_attr, &param);
#endif

	(void)mkdir(cf->rootdir, 0755);
	if (chdir(cf->rootdir) < 0) {
		report(LOG_ERR, "chdir \"%s\": %m", cf->rootdir);
		exit(1);
	}

#ifndef	DEBUG
	/*
	 * Start as daemon.
	 */
#ifdef	HAVE_DAEMON
	daemon(1, 0);
#else
	if (fork() > 0) _exit(0);
	setsid();
	if ((op = open("/dev/null", O_RDWR)) != -1) {
		dup2(op, 0);
		dup2(op, 1);
		dup2(op, 2);
		if (op > 2) close(op);
	}
#endif
	if (syslog_initialized == 0) {
		if (syslog_facility == 0)
			syslog_facility = LOG_DAEMON;
		openlog(program_name, LOG_PID|LOG_CONS, syslog_facility);
		syslog_initialized = 1;
	}
#endif
	mypid = getpid();

	if (trap_init(cf->enable_traps > 0) < 0)
		exit(1);
	if (netstate_init(cf->ns_port) < 0)
		exit(1);

	/*
	 * Make pid file.
	 */
	(void)strcpy(buf, program_name);
	(void)strcat(buf, ".pid");
	if ((fp = fopen(buf, "w")) != NULL) {
		fprintf(fp, "%d\n", (int)mypid);
		fclose(fp);
	}

#ifndef	HAVE_PTHREAD
	/*
	 * Save start and extend of argv.
	 */
	Argv = argv;
	while (*envp) envp++;
	LastArgv = envp[-1] + strlen(envp[-1]);
#endif

	report(LOG_NOTICE, "started");

	/*
	 * Prepare signal mask and handlers.
	 */
	sigemptyset(&sigmask);
	for (sp = sighandler_tab; sp && sp->sig; sp++)
		sigaddset(&sigmask, sp->sig);
#if defined(HAVE_PTHREAD) && !defined(PTHREAD_SIGNALS_BROKEN)
	pthread_sigmask(SIG_SETMASK, &sigmask, NULL);
#else
	sigprocmask(SIG_SETMASK, &sigmask, NULL);
#endif
	for (sp = sighandler_tab; sp && sp->sig; sp++) {
		sigact.sa_handler = sp->handler;
		memcpy(&sigact.sa_mask, &sigmask, sizeof(sigset_t));
		sigact.sa_flags = sp->flags;
		sigaction(sp->sig, &sigact, NULL);
	}
#ifdef	HAVE_SIGINTERRUPT
	for (sp = sighandler_tab; sp && sp->sig; sp++) {
		if ((sp->flags & SA_RESTART) == 0)
			siginterrupt(sp->sig, 1);
	}
#endif

	if (watchdog_timeout)
		set_timer(watchdog_timeout, watchdog_handler);
	start_object_list(cf->target);
	mainloop(&sigmask);

	/* NOT REACHED */
	cleanup_handler(0);
}

void
watchdog(passed)
	int passed; /* seconds since previous call */
{
	OBJECT *target, *service;
	TIMEVAL now, tv;

	dprintf(("watchdog(%d) called\n", passed));

	if (!passed) return;

	gettimeofday(&now, NULL);
	for (target = cf->target; target; target = target->next) {
		if (target->polling > 0 && target->method_list) {
			tv = now;
			tv_sub(&tv, &target->last_request);
			if (tv.tv_sec >= passed) {
				int lost = !remove_event(NULL, target);

				target->state = 0; /* set NONE state */
				reset_statistic(target->var_list, 1);

				object_stop(target);
				for (service = target->service; service; service = service->next)
					object_stop(service);

				set_none_state(target);
				save_object_state(target);

				report(LOG_WARNING, "watchdog %s polling at \"%s\"",
				       lost ? "resume" : "restart", target->name);

				object_init(target);
				for (service = target->service; service; service = service->next) {
					service->ip_addr = target->ip_addr;
					service->parent = target;
					object_init(service);
				}
				start_method_list(target);
			}
		} else if (trap_sock < 0) {
			report(LOG_WARNING, "No polling configured at \"%s\"?",
			       target->name);
		}
	}
}

void
mainloop(sigmask)
	sigset_t *sigmask;
{
	void *parsed_cf = NULL;
	fd_set readfds, writefds;
	int op, nfds, save_errno;
	TIMEVAL timeout;
	TRAP *trap;
	struct sockaddr_in from;
#ifdef	HAVE_PTHREAD
	pthread_t reconfig_thr, netstate_thr;
#endif

	for (;;) {
		if (reconfig_pending > 0) {
			if (!parsed_cf) {
#ifdef	HAVE_PTHREAD
				pthread_create(&reconfig_thr, &joinable_attr, parse_config, (void *)reconfig_pending);
				pthread_join(reconfig_thr, &parsed_cf);
#else
				parsed_cf = parse_config((void *)reconfig_pending);
#endif
				if (!parsed_cf) {
					report(LOG_NOTICE, "reconfig canceled");
#ifdef	HAVE_PTHREAD
					reconfig_pending = 0;
#else
					reconfig_pending = -1;
#endif
				}
			}
			if (parsed_cf && !session_busy(NULL)
#ifdef	HAVE_PTHREAD
			    && !busy_threads
#endif
			    ) {
				if (reconfig_pending == SIGHUP)
					soft_reconfig((CONFIG *)parsed_cf);
				else	hard_reconfig((CONFIG *)parsed_cf);
				parsed_cf = NULL;
				report(LOG_NOTICE, "reconfig completed");
				reconfig_pending = 0;
			}
		}
		if (watchdog_pending && !session_busy(NULL)
#ifdef	HAVE_PTHREAD
		    && !busy_threads
#endif
		    ) {
			watchdog(watchdog_pending * watchdog_timeout);
			watchdog_pending = 0;
		}

		nfds = 0;
		FD_ZERO(&readfds);
		if (trap_sock != -1) {
			nfds = trap_sock + 1;
			FD_SET(trap_sock, &readfds);
		}
		if (netstate_sock != -1) {
			nfds = netstate_sock + 1;
			FD_SET(netstate_sock, &readfds);
		}
		FD_ZERO(&writefds);
		op = select_event(&timeout);
		session_select(&nfds, &readfds, &writefds, &timeout, &op);
#if defined(HAVE_PTHREAD) && !defined(PTHREAD_SIGNALS_BROKEN)
		pthread_sigmask(SIG_UNBLOCK, sigmask, NULL);
#else
		sigprocmask(SIG_UNBLOCK, sigmask, NULL);
#endif
		op = select(nfds, &readfds, &writefds, 0, op ? 0 : &timeout);
		save_errno = errno;
#if defined(HAVE_PTHREAD) && !defined(PTHREAD_SIGNALS_BROKEN)
		pthread_sigmask(SIG_BLOCK, sigmask, NULL);
#else
		sigprocmask(SIG_BLOCK, sigmask, NULL);
#endif
		if (op > 0) {
			trace(("\t select operate\n"));

			session_operate(&readfds, &writefds);

			if (trap_sock != -1 &&
			    FD_ISSET(trap_sock, &readfds)) {
				if ((trap = snmp_trap(trap_sock, cf->source_traps > 0)) != NULL) {
					trap_parse(trap);
					trap_free(trap);
				}
			}
			if (netstate_sock != -1 &&
			    FD_ISSET(netstate_sock, &readfds)) {
				socklen_t from_len = sizeof(from);
				if ((op = accept(netstate_sock, (struct sockaddr *)&from, &from_len)) != -1) {
#ifdef	HAVE_PTHREAD
					pthread_create(&netstate_thr, &detached_attr, netstate_serve, (void *)op);
#else
					if (fork() == 0) { /* child */
						close(netstate_sock);
						set_timer(0, NULL);
						netstate_serve((void *)op);
						_exit(0);
					}
					close(op); /* parent */
#endif
				} else	report(LOG_ERR, "accept: %m");
			}
		} else if (op == 0) {
			trace(("\t select timedout\n"));

			session_timeout();

		} else if (save_errno != EINTR) {
			trace(("\t select error\n"));

			report(LOG_ERR, "select: %s", strerror(save_errno));
			/* cleanup_handler(0); */
		} else {
			trace(("\t select interrupted\n"));
		}
	}
	/* NOT REACHED */
}

int
set_timer(timeout, func)
	int timeout;	/* seconds */
	void (*func)();
{
	struct sigaction sigact;
	struct itimerval it, rest;

	sigaction(SIGALRM, NULL, &sigact);
	if (sigact.sa_handler != func) {
		if (func)
			sigact.sa_handler = func;
		else	sigact.sa_handler = cleanup_handler;
		sigaction(SIGALRM, &sigact, NULL);
	}
	it.it_value.tv_sec = it.it_interval.tv_sec = timeout;
	it.it_value.tv_usec = it.it_interval.tv_usec = 0;
	setitimer(ITIMER_REAL, &it, &rest);

	return (rest.it_value.tv_sec);
}

void
object_init(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (!method->init) {
			errno = EINVAL;
			report(LOG_ERR, "object_init(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
		if ((*method->init)(object, method) < 0) {
			report(LOG_ERR, "object_init(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
	}
}

void
object_stop(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (!method->stop) {
			errno = EINVAL;
			report(LOG_ERR, "object_stop(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
		(*method->stop)(method);
	}
}

int
object_busy(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (method->sd && session_busy(method->sd))
			return 1;
	}
	return 0;
}

void
start_method_list(arg)
	void *arg;
{
	OBJECT *object = (OBJECT *)arg;
	METHOD *method;
	TIMEVAL tv;
	int busy;

	/* sanity check */
	if (!object) {
		errno = EINVAL;
		report(LOG_ERR, "start_method_list(%s): %m", object->name);
		cleanup_handler(0);
	}

	method = object->method_list;
	if (!method) {
		report(LOG_NOTICE, "No polling methods at \"%s\"", object->name);
		/* save_object_data(object); -- no periodic saving for traps */
		return;
	}

	/* current timestamp */
	gettimeofday(&tv, NULL);

	busy = object_busy(object);
	if (busy || (!object->parent && !check_netpath(object))) {
		if (busy) {
			if (object->parent)
				report(LOG_DEBUG, "Polling methods at \"%s/%s\" still busy -- aborted!",
				       object->parent->name, object->name);
			else	report(LOG_DEBUG, "Polling methods at \"%s\" still busy -- aborted!",
				       object->name);
			object_stop(object);
		}

		/* set the NONE state on the object */
		object->state = 0;
		reset_statistic(object->var_list, 1);

		/* set the NONE state to all subobjects */
		set_none_state(object);

		/* save state changes if required */
		save_object_state(object);

		/* stop data saving on the object */
		remove_event(save_object_data, object);

		/* keep normal polling */
		if (object->polling > 0) {
			tv_sharp(&tv, object->polling, object_offset(object, object->polling));
			add_event(&tv, start_method_list, object);
		}
		return;
	}

	dprintf(("start_method_list(%s)\n", object->name));

	object->prev_request = object->last_request;
	object->last_request = tv; /* last request timestamp */

	(*method->start)(method);
}

void
method_finished(object, method, diag, ok)
	OBJECT *object;
	METHOD *method;
	char *diag;
	int ok;
{
	TIMEVAL tv;
	VARIABLE *var;
	OBJECT *service;

	/* current timestamp */
	gettimeofday(&tv, NULL);

	/* save method diagnostic */
	if (method && diag && (var = get_var(object->var_list, method->name)) != NULL)
		str2var(var, diag);

	if (!method || !ok) {
		/*
		 * Method list aborted or Start Trap received.
		 */

		/* update object operational status */
		object->prev_state = object->state;
		object->state = STATE_DOWN;

		if (object->state != object->prev_state)
			object->last_change = tv;

#ifdef	DEBUG
		if (object->prev_state != STATE_DOWN)
			dprintf(("object \"%s\" change state to DOWN\n", object->name));
#endif

		/* stop anything here */
		object_stop(object);

		for (service = object->service; service; service = service->next)
			object_stop(service);

		set_none_state(object);

		save_object_state(object);

		/* stop data saving on the object */
		remove_event(save_object_data, object);

		/* keep touching the object if required */
		if (object->polling > 0) {
			tv.tv_sec += object->polling / 2;
			add_event(&tv, start_method_list, object);
		}
		return;
	}

	if (method->next) {
		/*
		 * Advance to next object method.
		 */

		method = method->next;
		(*method->start)(method);
		return;
	}

	/*
	 * Method list done.
	 */

	object->prev_reply = object->last_reply;
	object->last_reply = tv; /* last reply timestamp */

	/* update object operational status */
	object->prev_state = object->state;
	object->state = STATE_UP;

	if (object->state != object->prev_state)
		object->last_change = tv;

#ifdef	DEBUG
	if (object->prev_state != STATE_UP)
		dprintf(("object \"%s\" change state to UP\n", object->name));
#endif

	save_object_state(object);

	/* to avoid network flooding at startup */
	if (timerisset(&object->prev_request)) {
		/* if list of subobject present start it here */
		for (service = object->service; service; service = service->next)
			start_method_list(service);
	}

	if (object->polling > 0) {
		/* start data saving if required */
		if (object->saving > 0 && !find_event(save_object_data, object)) {
			if (!object->sync) {
				save_object_data(object);
			} else if (--object->sync == 0) {
				object->sync = object->saving / object->polling;
				save_object_data(object);
			}
		}
		/* keep polling the object */
		tv_sharp(&tv, object->polling, object_offset(object, object->polling));
		add_event(&tv, start_method_list, object);
	}
}

void
start_object_list(object)
	OBJECT *object;
{
	OBJECT *target, *service;

	/* initialize sessions for configured objects */
	for (target = object; target; target = target->next) {
		object_init(target);

		/* and it subobjects */
		for (service = target->service; service; service = service->next) {
			service->ip_addr = target->ip_addr;
			service->parent = target;

			object_init(service);
		}
	}

	/* start sessions for configured objects */
	for (target = object; target; target = target->next)
		start_method_list(target);
}

#ifndef	HAVE_PTHREAD
/*
 * clobber argv so ps will show what we're doing.
 * stolen from sendmail
 */
void
proctitle(str)
	char *str;
{
	int i;
	char *cp, buf[1024];

	cp = Argv[0];
	sprintf(buf, "%s: %s", program_name, str);
	i = strlen(buf);
	if (i > LastArgv - cp - 1)
		i = LastArgv - cp - 1;
	(void)strncpy(cp, buf, i);
	cp[i] = '\0';
	cp += i;
	while (cp < LastArgv) *cp++ = ' ';
}
#endif

#define	MAX_ARGS	50

void
execfile(ispipe, file, args)
	int ispipe;
	char *file, *args;
{
	sigset_t sigmask;
	int ac = 0, fildes[2];
	char *av[MAX_ARGS+2];
	extern char **environ;

	dprintf(("execfile: %s %s\n", file, args ? args : "null"));

	if (ispipe && pipe(fildes) < 0) {
		report(LOG_ERR, "pipe: %m");
		return;
	}

	if (*file != '/') {
		av[ac++] = "sh";
		av[ac++] = "-c";
		av[ac++] = file;
#ifdef	_PATH_BSHELL
		file = _PATH_BSHELL;
#else
		file = "/bin/sh";
#endif
	} else	av[ac++] = strippath(file);
	if (!ispipe)
		ac += make_argv(args, (char ***) &av[ac], MAX_ARGS);
	av[ac] = NULL;

	switch (vfork()) {
	case -1:
		report(LOG_ERR, "fork: %m");
		if (ispipe) {
			close(fildes[0]);
			close(fildes[1]);
		}
		return;
	case 0:		/* child */
		if (ispipe) {
			if (fildes[0] != STDIN_FILENO) {
				dup2(fildes[0], STDIN_FILENO);
				close(fildes[0]);
			}
			close(fildes[1]);
		}

		/* child would be terminated by signals */
		sigprocmask(SIG_SETMASK, NULL, &sigmask);
		sigprocmask(SIG_UNBLOCK, &sigmask, NULL);

		/* make session leader to be able killpg() latter */
		setsid();

		execve(file, av, environ);
		report(LOG_ERR, "execve %s: %m", file);
		_exit(127);
	}
	/* parent */

	if (ispipe) {
		FILE *fp;

		close(fildes[0]);
		if ((fp = fdopen(fildes[1], "w")) != NULL) {
			fprintf(fp, "%s\n", args);
			fclose(fp);
		} else	report(LOG_ERR, "fdopen: %m");
	}
}

static void
childend_handler(sig)
	int sig;
{
	pid_t pid;
	int status, done;

	dprintf(("childend: entered\n"));

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {

		/* set the exit status of executed program */
		if (WIFEXITED(status))
			done = WEXITSTATUS(status);
		else	done = -1;

		if (done)
			report(LOG_INFO, "child %d exit %d\n", (int)pid, done);
	}
}

static void
reconfig_handler(sig)
	int sig;
{
#ifdef	HAVE_PTHREAD
	if (pthread_self() == main_thr)
#else
	if (getpid() == mypid)
#endif
	{
		if (reconfig_pending < 0)
			report(LOG_WARNING, "reconfig denied");
		else if (!reconfig_pending) {
			reconfig_pending = sig;
			report(LOG_NOTICE, "%s reconfig requested",
			       sig == SIGHUP ? "soft" : "hard");
		} else	report(LOG_WARNING, "reconfig requested too frequently");
	}
}

static void
brokenpipe_handler(sig)
	int sig;
{
#ifdef	HAVE_PTHREAD
	if (pthread_self() != main_thr)
		pthread_exit(0); // file descriptors leak? XXX
#else
	if (getpid() != mypid)
		cleanup_handler(sig);
#endif
}

static void
watchdog_handler(sig)
	int sig;
{
	if (
#ifdef	HAVE_PTHREAD
	    pthread_self() == main_thr
#else
	    getpid() == mypid
#endif
	    && watchdog_timeout) {
		watchdog_pending++;
	}
}

static void
cleanup_handler(sig)
	int sig;
{
#ifdef	HAVE_PTHREAD
	if (pthread_self() == main_thr)
#else
	if (getpid() == mypid)
#endif
	{
		char pidfile[100];
		(void)strcpy(pidfile, program_name);
		(void)strcat(pidfile, ".pid");
		(void)unlink(pidfile);
		report(LOG_CRIT, "aborted by signal %d", sig);
	} else	report(LOG_INFO, "interrupted by signal %d", sig);
	exit(1);
}

#include <zlib.h>

void
vers()
{
	fprintf(stderr, "\n%s Version %s\n\tcompiled for %s with\n %s\n",
		package, version, target, compiled);
#ifdef	HAVE_PTHREAD
	fprintf(stderr, "\tusing POSIX threads\n");
#endif
#ifdef	HAVE_DMALLOC_H
	fprintf(stderr, "\twith dmalloc library\n");
#endif
	fprintf(stderr, "\twith zlib version %s\n", ZLIB_VERSION);
	fprintf(stderr, "\n%s\n", copyright);
	fprintf(stderr, "%s support list: netmon@service.risp.ru\n\
\tsend `subscribe netmon' to majordomo@service.risp.ru\n", package);
	exit(0);
}

void
usage()
{
	fprintf(stderr, "usage:\n\
  %s [-c file] [-C] [-M] [-v] [-w]\n", program_name);

	fprintf(stderr, "where:\n\
  -c file  Config file, default %s\n\
  -C       Check config file for syntax errors\n\
  -M       Dump known SNMP MIB tables and exit\n\
  -v       Print version number, compile-time options, and exit\n\
  -w sec   Watch dog timer (0=disable), default %d sec.\n",
		DEFAULT_CONFIG, DEFAULT_WATCHDOG);

	exit(1);
}
