/*
 *	Copyright (c) 2000-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <netdb.h>
#include <ctype.h>

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"
#include "regex.h"

#include <zlib.h>

#ifndef	MAX_BAD_INPUT
#define	MAX_BAD_INPUT	3
#endif
#ifndef	MAX_STR_LEN
#define	MAX_STR_LEN	1024
#endif

extern CONFIG *cf;
extern VAR_TABLE main_table[];
extern VAR_TABLE object_table[];
extern VAR_TABLE service_table[];
extern VAR_TABLE iface_table[];
extern VAR_TABLE bgp_table[];
extern VAR_TABLE peer_table[];
extern VAR_TABLE envmon_table[];
extern VAR_TABLE envgauge_table[];

int netstate_sock = -1;		/* netstate server socket */

static struct object_type_ent {
	char *name;
	int type;
} object_types[] = {
	{ "ANY",	TYPE_ANY	},
	{ "OBJECT",	TYPE_OBJECT	},
	{ "SERVICE",	TYPE_SERVICE	},
	{ "INTERFACE",	TYPE_INTERFACE	},
	{ "BGPAS",	TYPE_BGPAS	},
	{ "BGPPEER",	TYPE_BGPPEER	},
	{ "ENVMON",	TYPE_ENVMON	},
	{ "ENVTEMP",	TYPE_ENVTEMP	},
	{ "ENVVOLT",	TYPE_ENVVOLT	},
	{ "ENVFAN",	TYPE_ENVFAN	},
	{ "ENVPS",	TYPE_ENVPS	},
	{ 0 }
};

int
get_type_by_name(name)
	char *name;
{
	struct object_type_ent *p;

	for (p = object_types; p && p->name; p++) {
		if (!strcasecmp(name, p->name))
			return p->type;
	}
	return -1;
}

char *
get_name_by_type(type)
	int type;
{
	struct object_type_ent *p;

	for (p = object_types; p && p->name; p++) {
		if (type == p->type)
			return p->name;
	}
	return NULL;
}


/*
 * [Re]initialize netstate server socket.
 */
int
netstate_init(port)
	int port;
{
	static struct sockaddr_in sin;

	if (netstate_sock != -1)
		close(netstate_sock);

	if (port) {
		int on = 1;

		if ((netstate_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			report(LOG_ERR, "socket: %m");
			return -1;
		}

		/* by cause of reconfig, it lets reopen server socket */
		if (setsockopt(netstate_sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on)) < 0) {
			report(LOG_ERR, "setsockopt SO_REUSEADDR: %m");
			close(netstate_sock);
			netstate_sock = -1;
			return -1;
		}

		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(cf->ns_port);
		sin.sin_addr.s_addr = INADDR_ANY;
		if (bind(netstate_sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
			report(LOG_ERR, "bind port %d: %m", ntohs(sin.sin_port));
			close(netstate_sock);
			netstate_sock = -1;
			return -1;
		}
		listen(netstate_sock, 5);
	} else
		netstate_sock = -1;

	return 0;
}

static void
send_hard_var(fp, gz, re, var_table, object, buf, old)
	FILE *fp;
	gzFile gz;
	yre_t *re;
	VAR_TABLE *var_table;
	void *object;
	char *buf;
	int old;
{
	VAR_TABLE *var;
	char *cp, *vp = &buf[strlen(buf)];

	for (var = var_table; var && var->func; var++) {
		if (!*var->name || var->old != old)
			continue;

		(void)strcpy(vp, var->name);
		if (yre_exec(re, buf)) {
#ifdef	HAVE_PTHREAD
			pthread_mutex_lock(&variables_lock);
#endif
			cp = (*var->func)(object);
			if (*cp && strcmp(cp, "\"\"") && strcmp(cp, "0")) {
				if (gz) {
					gzprintf(gz, "%s = %s\r\n", buf, cp);
				} else if (fp) {
					fprintf(fp, "%s = %s\r\n", buf, cp);
				}
			}
#ifdef	HAVE_PTHREAD
			pthread_mutex_unlock(&variables_lock);
#endif
		}
	}
	*vp = '\0';
}

static void
send_soft_var(fp, gz, re, var_list, buf, old, mtime)
	FILE *fp;
	gzFile gz;
	yre_t *re;
	VARIABLE *var_list;
	char *buf;
	int old, mtime;
{
	VARIABLE *var, *curr;
	char *vp = &buf[strlen(buf)];
	char strval[MAX_STR_LEN];

	for (var = var_list; var; var = var->next) {
		if (!old) curr = var;
		else curr = var->old;
		if (!curr || !curr->value.integer)
			continue; /* skip unused variables */

		(void)strcpy(vp, var->name);
		if (yre_exec(re, buf)) {
			if (gz) {
				gzprintf(gz, "%s = ", buf);
			} else if (fp) {
				fprintf(fp, "%s = ", buf);
			}
			if (mtime) {
#ifdef	USE_VAR_MTIME
				if (curr) {
					if (gz) {
						gzprintf(gz, "%u", tv_round(&curr->mtime));
					} else if (fp) {
						fprintf(fp, "%u", tv_round(&curr->mtime));
					}
				} else
#endif
				if (gz) {
					gzprintf(gz, "Unused");
				} else if (fp) {
					fprintf(fp, "Unused");
				}
			} else if (gz) {
				gzprintf(gz, "%s", var2str(strval, curr));
			} else if (fp) {
				fprintf(fp, "%s", var2str(strval, curr));
			}
			if (gz) {
				gzprintf(gz, "\r\n");
			} else if (fp) {
				fprintf(fp, "\r\n");
			}
		}
	}
	*vp = '\0';
}

/*
 * Send requested variable(s) using gz of fp output stream.
 */
static void
send_variables(fp, gz, re, obj_type, old, mtime, addr, name)
	FILE *fp;
	gzFile gz;
	yre_t *re;
	int obj_type, old, mtime;
	char *addr, *name; /* host access */
{
	OBJECT *target, *service;
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	ENV_MON *env;
	ENV_GAUGE *gauge;
	char buf[BUFSIZ], tmpbuf[100];

#ifdef	HAVE_PTHREAD
	busy_threads++;
#endif
	if (obj_type == TYPE_ANY) {
		buf[0] = '\0';
		if (!mtime)
			send_hard_var(fp, gz, re, main_table, 0, buf, old);
	}

	for (target = cf->target; target; target = target->next) {
		if (target->ns_acl &&
		    !check_group_ref(target->ns_acl, addr, name))
			continue;

		if (obj_type == TYPE_ANY || obj_type == TYPE_OBJECT) {
			if (gz) {
				gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_OBJECT));
			} else if (fp) {
				fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_OBJECT));
			}
			sprintf(buf, "%s!", target->name);
			if (!mtime)
				send_hard_var(fp, gz, re, object_table, target, buf, old);
			send_soft_var(fp, gz, re, target->var_list, buf, old, mtime);
		}

		for (iface = target->interface; iface; iface = iface->next) {
			if (iface->ns_acl &&
			    !check_group_ref(iface->ns_acl, addr, name))
				continue;

			if (obj_type == TYPE_ANY || obj_type == TYPE_INTERFACE) {
				if (gz) {
					gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_INTERFACE));
				} else if (fp) {
					fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_INTERFACE));
				}
				sprintf(buf, "%s!%s!", target->name, iface->name ? iface->name : itoa(iface->index));
				if (!mtime)
					send_hard_var(fp, gz, re, iface_table, iface, buf, old);
				send_soft_var(fp, gz, re, iface->var_list, buf, old, mtime);
			}
		}

		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			if (bgp->ns_acl &&
			    !check_group_ref(bgp->ns_acl, addr, name))
				continue;

			if (obj_type == TYPE_ANY || obj_type == TYPE_BGPAS) {
				if (gz) {
					gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_BGPAS));
				} else if (fp) {
					fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_BGPAS));
				}
				sprintf(buf, "%s!AS%d!", target->name, bgp->asn);
				if (!mtime)
					send_hard_var(fp, gz, re, bgp_table, bgp, buf, old);
			}

			for (peer = bgp->peer; peer; peer = peer->next) {

				if (obj_type == TYPE_ANY || obj_type == TYPE_BGPPEER) {
					if (gz) {
						gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_BGPPEER));
					} else if (fp) {
						fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_BGPPEER));
					}
					sprintf(buf, "%s!AS%d!%s!", target->name,
						bgp->asn, get_bgppeer_name(tmpbuf, peer->ip_addr));
					if (!mtime)
						send_hard_var(fp, gz, re, peer_table, peer, buf, old);
					send_soft_var(fp, gz, re, peer->var_list, buf, old, mtime);
				}
			}
		}

		for (env = target->env; env; env = env->next) {
			if (env->ns_acl &&
			    !check_group_ref(env->ns_acl, addr, name))
				continue;

			if (obj_type == TYPE_ANY || obj_type == TYPE_ENVMON) {
				if (gz) {
					gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_ENVMON));
				} else if (fp) {
					fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_ENVMON));
				}
				sprintf(buf, "%s!%s!", target->name, get_name_by_type(env->type));
				if (!mtime)
					send_hard_var(fp, gz, re, envmon_table, env, buf, old);
			}

			for (gauge = env->gauge; gauge; gauge = gauge->next) {

				if (obj_type == TYPE_ANY || obj_type == env->type) {
					if (gz) {
						gzprintf(gz, "!%s\r\n", get_name_by_type(env->type));
					} else if (fp) {
						fprintf(fp, "!%s\r\n", get_name_by_type(env->type));
					}
					sprintf(buf, "%s!%s!%s!", target->name,
						get_name_by_type(env->type), gauge->name);
					if (!mtime)
						send_hard_var(fp, gz, re, envgauge_table, gauge, buf, old);
					send_soft_var(fp, gz, re, gauge->var_list, buf, old, mtime);
				}
			}
		}

		for (service = target->service; service; service = service->next) {
			if (service->ns_acl &&
			    !check_group_ref(service->ns_acl, addr, name))
				continue;

			if (obj_type == TYPE_ANY || obj_type == TYPE_SERVICE) {
				if (gz) {
					gzprintf(gz, "!%s\r\n", get_name_by_type(TYPE_SERVICE));
				} else if (fp) {
					fprintf(fp, "!%s\r\n", get_name_by_type(TYPE_SERVICE));
				}
				sprintf(buf, "%s!%s!", target->name, service->name);
				if (!mtime)
					send_hard_var(fp, gz, re, service_table, service, buf, old);
				send_soft_var(fp, gz, re, service->var_list, buf, old, mtime);
			}
		}
#ifdef	notdef
		if (gz) {
			if (gzflush(gz, Z_FINISH) != Z_OK)
				break;
		} else if (fp) {
			if (fflush(fp) < 0)
				break;
		}
#endif
	}
#ifdef	HAVE_PTHREAD
	busy_threads--;
#endif
}

#ifndef	HAVE_PTHREAD
static void
interrupt()
{
	report(LOG_INFO, "netstate timeout");
	_exit(1);
}
#endif

void *
netstate_serve(arg)
	void *arg;
{
	int sock = (int)arg;
	struct sockaddr_in from;
	FILE *fp_in, *fp_out;
	struct hostent *hp;
	int bad_input, obj_type, old, mtime, gz_level;
	socklen_t from_len;
	sigset_t sigmask;
	yre_t re;
	char client_addr[50], client_name[100];
	char *cp, *next, input[MAX_STR_LEN];
	char inbuf[MAX_STR_LEN], outbuf[MAX_STR_LEN];

	from_len = sizeof(from);
	if (getpeername(sock, (struct sockaddr *)&from, &from_len) < 0) {
		report(LOG_ERR, "getpeername: %m");
		close(sock);
		return;
	}
#ifdef	SO_KEEPALIVE
	from_len = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *)&from_len, sizeof(int)) < 0) {
		report(LOG_ERR, "setsockopt SO_KEEPALIVE: %m");
		close(sock);
		return;
	}
#endif
	if ((fp_in = fdopen(sock, "r")) == NULL) {
		report(LOG_ERR, "fdopen(r): %m");
		close(sock);
		return;
	}
	setvbuf(fp_in, inbuf, _IOLBF, sizeof(inbuf)); /* setlinebuf */

	if ((fp_out = fdopen(sock, "w")) == NULL) {
		report(LOG_ERR, "fdopen(w): %m");
		fclose(fp_in);
		close(sock);
		return;
	}
	setvbuf(fp_out, outbuf, _IOLBF, sizeof(outbuf)); /* setlinebuf */

#if defined(HAVE_PTHREAD) && !defined(PTHREAD_SIGNALS_BROKEN)
	pthread_sigmask(SIG_SETMASK, NULL, &sigmask);
	pthread_sigmask(SIG_UNBLOCK, &sigmask, NULL);
#else
	sigprocmask(SIG_SETMASK, NULL, &sigmask);
	sigprocmask(SIG_UNBLOCK, &sigmask, NULL);
#endif

#ifdef	HAVE_PTHREAD
	pthread_mutex_lock(&gethostby_lock); /* unsafe calls follow */
#endif
	hp = gethostbyaddr((char *)&from.sin_addr, 4, AF_INET);
	if (hp && hp->h_name) {
		(void)strncpy(client_name, hp->h_name, sizeof(client_name)-1);
		client_name[sizeof(client_name)-1] = '\0';
	} else	client_name[0] = '\0';
#ifdef	HAVE_PTHREAD
	pthread_mutex_unlock(&gethostby_lock);
#endif
	intoa(client_addr, from.sin_addr);
	if (!check_group_ref(cf->ns_acl, client_addr, client_name)) {
		report(LOG_NOTICE, "connect from \"%s\" [%s] refused",
		       client_name, client_addr);
		fprintf(fp_out, "Access for \"%s\" [%s] denied\r\n",
			client_name, client_addr);
		fclose(fp_in);
		fclose(fp_out);
		close(sock);
		return;
	}
#ifdef	HAVE_PTHREAD
	fprintf(fp_out, "NetState server ready (no timeout)\r\n");
#else
	fprintf(fp_out, "NetState server ready (timeout %d sec.)\r\n",
		cf->ns_timo);
#endif
	fprintf(fp_out, "Compress: ZLIB %s\r\n", zlibVersion());

	/* proccess the client commands */
	bad_input = 0;
	gz_level = Z_NO_COMPRESSION;
	while (!get_socket_error(sock) && bad_input < MAX_BAD_INPUT) {
		fprintf(fp_out, "!\r\n");
#ifndef	HAVE_PTHREAD
		if (cf->ns_timo) /* timeout for command */
			set_timer(cf->ns_timo, interrupt);
#endif
		cp = fgets(input, sizeof(input), fp_in);
		input[sizeof(input)-1] = '\0';
#ifndef	HAVE_PTHREAD
		if (cf->ns_timo)
			set_timer(0, interrupt);
#endif
		if (!cp) break;
		while (isprint(*cp)) cp++;
		*cp = '\0';

		next = input;
		if ((cp = my_strsep(&next, " ")) == NULL) {
			bad_input++;
			fprintf(fp_out, "! Invalid command \"%s\"\r\n", input);
			continue;
		}
		if (!strcasecmp(cp, "QUIT"))
			break;

		if (!strcasecmp(cp, "COMPRESS")) {
			gz_level = Z_NO_COMPRESSION;
			if ((cp = my_strsep(&next, "")) != NULL) {
				gz_level = atoi(cp);
				if (gz_level < Z_BEST_SPEED)
					gz_level = Z_NO_COMPRESSION;
				else if (gz_level > Z_BEST_COMPRESSION)
					gz_level = Z_BEST_COMPRESSION;
			}
			bad_input = 0;
			continue;
		}

		old = mtime = 0;
		if (!strcasecmp(cp, "OLD")) {
			if ((cp = my_strsep(&next, " ")) == NULL) {
				bad_input++;
				fprintf(fp_out, "! Invalid command \"%s\"\r\n",
					input);
				continue;
			}
			old++;
		}
#ifdef	USE_VAR_MTIME
		if (!strcasecmp(cp, "MTIME")) {
			if ((cp = my_strsep(&next, " ")) == NULL) {
				bad_input++;
				fprintf(fp_out, "! Invalid command \"%s\"\r\n",
					input);
				continue;
			}
			mtime++;
		}
#endif
		if ((obj_type = get_type_by_name(cp)) < 0) {
			bad_input++;
			fprintf(fp_out, "! Unknown type \"%s\"\r\n", cp);
			continue;
		}
		if ((cp = my_strsep(&next, "")) == NULL) {
			bad_input++;
			fprintf(fp_out, "! Missing regex pattern\r\n");
			continue;
		}

		if ((next = yre_comp(&re, cp)) != NULL) {
			bad_input++;
			fprintf(fp_out, "! Bad regex pattern \"%s\": %s\r\n",
				cp, next);
			continue;
		}

		if (gz_level != Z_NO_COMPRESSION) {
			gzFile gz_out = gzdopen(dup(sock), "w");
			if (!gz_out) {
				bad_input++;
				report(LOG_ERR, "gzdopen: %m");
				fprintf(fp_out, "! Can't gzdopen this socket\r\n");
				continue;
			}
			gzsetparams(gz_out, gz_level, Z_DEFAULT_STRATEGY);
			send_variables(0, gz_out, &re, obj_type, old, mtime,
				       client_addr, client_name);
			gzclose(gz_out);
		} else {
			send_variables(fp_out, 0, &re, obj_type, old, mtime,
				       client_addr, client_name);
		}
		bad_input = 0;
	}
	fclose(fp_in);
	fclose(fp_out);
	close(sock); /* already closed? */

	return; /* request done */
}
