%{
/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <netdb.h>
#include <ctype.h>

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"
#include "snmp.h"
#include "regex.h"
#include "mib.h"

extern int errno;

#define	ILLEGAL_CHARS	"/!,"

static CONFIG config;

static OBJECT object, subobject;
static int add_object __P((OBJECT **chain, OBJECT *item));

static INTERFACE interface;
static int add_interface __P((INTERFACE **chain, INTERFACE *item));

static BGP_AS bgp_as;
static int add_bgp_as __P((BGP_AS **chain, BGP_AS *item));
static int add_bgp_peer __P((BGP_AS *bgp, char *peer));

static ENV_MON env_mon;
static int add_env_mon __P((ENV_MON **chain, ENV_MON *item));
static int add_env_gauge __P((ENV_MON *env, char *gauge));

static CHATSCRIPT chatscript, *first_chatscript;
static REFERENCE *first_reference;

static METHOD method, *first_method;
static int add_method __P((METHOD **chain, METHOD *item));

static TRAP trap, *first_trap;
static int add_trap __P((TRAP **chain, TRAP *item));

static SAVE save, *first_save;
static int add_save __P((SAVE **chain, SAVE *item));

static VARIABLE *global_var_list;
static int add_variable __P((VARIABLE **list, VARIABLE *var));
static int add_var_ref __P((VARIABLE **list, REFERENCE **chain, char *name));

static GROUP_LIST group_list;
static GROUP_LIST *find_group_list __P((GROUP_LIST *list, char *name));
static int add_group_list __P((GROUP_LIST **chain, GROUP_LIST *group));
static int add_group_entry __P((GROUP_LIST *list, int permit, char *patt));
static int add_group_ref __P((GROUP_LIST *list, GROUP_REF **chain, char *name));

static int gethostaddr __P((struct in_addr *ip, const char *host));

#define	CLEAR_CONFIG		memset(&config, 0, sizeof(CONFIG))
#define	CLEAR_OBJECT(o)		memset(o, 0, sizeof(OBJECT))
#define	CLEAR_INTERFACE(i)	memset(i, 0, sizeof(INTERFACE))
#define	CLEAR_BGP_AS(b)		memset(b, 0, sizeof(BGP_AS))
#define	CLEAR_ENV_MON(e)	memset(e, 0, sizeof(ENV_MON))
#define	CLEAR_METHOD(m)		memset(m, 0, sizeof(METHOD))
#define	CLEAR_TRAP(t)		memset(t, 0, sizeof(TRAP))
#define	CLEAR_CHATSCRIPT(c)	memset(c, 0, sizeof(CHATSCRIPT))
#define	CLEAR_SAVE(d)		memset(d, 0, sizeof(SAVE))
#define	CLEAR_GROUP_LIST(g)	memset(g, 0, sizeof(GROUP_LIST))

static void
clear_config()
{
	CLEAR_CONFIG;

	CLEAR_GROUP_LIST(&group_list);

	CLEAR_OBJECT(&object);
	CLEAR_OBJECT(&subobject);

	CLEAR_INTERFACE(&interface);
	CLEAR_BGP_AS(&bgp_as);
	CLEAR_ENV_MON(&env_mon);

	CLEAR_METHOD(&method);
	first_method = NULL;

	CLEAR_TRAP(&trap);
	first_trap = NULL;

	CLEAR_CHATSCRIPT(&chatscript);
	first_chatscript = NULL;
	first_reference = NULL;

	CLEAR_SAVE(&save);
	first_save = NULL;

	global_var_list = 0;
}

static char *
fix_file_name(name)
	char *name;
{
	static char buf[100];
	register char *cp = buf;

	while (*name && cp < buf + sizeof(buf) - 1) {
		if (strchr(ILLEGAL_CHARS, *name))
			*cp++ = '_';
		else if (!isspace(*name))
			*cp++ = *name;
		name++;
	}
	*cp = '\0';
	return buf;
}

/* make by full path of requested directory */
static int
make_dir(dir)
	char *dir;
{
	char *cp, cwd[1024], buf[1024];

	if (dir[0] != '/' || dir[1] == '\0') {
		errno = ENOTDIR;
		return -1;
	}
	if (!getcwd(cwd, sizeof(cwd))) return -1;
	if (chdir("/") < 0) return -1;
	for (cp = strcpy(buf, dir); (cp = strtok(cp, "/")) != NULL; cp = NULL) {
		errno = 0;
		if (mkdir(cp, 0755) < 0 && errno && errno != EEXIST) return -1;
		if (chdir(cp) < 0) return -1;
	}
	chdir(cwd);
	return 0;
}

static int
make_save_path(dir, ds)
	char *dir;
	register SAVE *ds;
{
	char *cp, buf[1024];

	for (; ds; ds = ds->next) {
		if (ds->method != SAVE_FILE || !ds->file)
			continue;

		/* make directory tree */
		if (*ds->file != '/') {
			(void)strcpy(buf, dir);
			(void)strcat(buf, "/");
		} else	*buf = '\0';
		(void)strcat(buf, ds->file);
		if ((cp = strrchr(buf, '/')) != NULL) *cp = '\0';
		if (make_dir(buf) < 0) return -1;

		/* replace save file name to reflect full path */
		if (*ds->file != '/') {
			if (cp) *cp = '/'; /* return it back */
			free(ds->file);
			if ((ds->file = strdup(buf)) == NULL) return -1;
		}
	}
	return 0;
}

static CONFIG *
set_config()
{
	OBJECT *target, *service;
	INTERFACE *iface;
	BGP_AS *bgp;
	ENV_MON *env;
	char *cp, buf[1024];

	if (!config.rootdir) {
		report(LOG_ERR, "%s: rootdir unspecified", config_file);
		return NULL;
	}
	if (config.polling) {
		if (!config.timeout)
			config.timeout = TIMEOUT_DEFAULT;
		if (!config.retries)
			config.retries = RETRIES_DEFAULT;
	}
	for (target = config.target; target; target = target->next) {
		target->parent = 0; /* root object */

		sprintf(buf, "%s/%s", config.rootdir, target->datadir ?
			target->datadir : fix_file_name(target->name));
		if (make_save_path(buf, target->save_list) < 0) {
			report(LOG_ERR, "%s: %s: %m", config_file, buf);
			return NULL;
		}
		cp = &buf[strlen(buf)]; /* end of parent dir */

		for (iface = target->interface; iface; iface = iface->next) {
			iface->parent = target;

			(void)strcpy(cp, "/");
			if (iface->datadir)
				(void)strcat(cp, iface->datadir);
			else if (iface->name)
				(void)strcat(cp, fix_file_name(iface->name));
			else	sprintf(cp+1, "Interface%d", iface->index);
			if (make_save_path(buf, iface->save_list) < 0) {
				report(LOG_ERR, "%s: %s: %m", config_file, buf);
				return NULL;
			}
		}
		for (iface = target->ifgroup; iface; iface = iface->next) {
			iface->parent = target;

			(void)strcpy(cp, "/");
			if (iface->datadir)
				(void)strcat(cp, iface->datadir);
			else	(void)strcat(cp, fix_file_name(iface->name));
			if (make_save_path(buf, iface->save_list) < 0) {
				report(LOG_ERR, "%s: %s: %m", config_file, buf);
				return NULL;
			}
		}

		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			bgp->parent = target;

			(void)strcpy(cp, "/");
			if (bgp->datadir)
				(void)strcat(cp, bgp->datadir);
			else	sprintf(cp+1, "AS%d", bgp->asn);
			if (make_save_path(buf, bgp->save_list) < 0) {
				report(LOG_ERR, "%s: %s: %m", config_file, buf);
				return NULL;
			}
		}

		for (env = target->env; env; env = env->next) {
			env->parent = target;

			(void)strcpy(cp, "/");
			(void)strcat(cp, env->datadir ? env->datadir :
				     fix_file_name(get_name_by_type(env->type)));
			if (make_save_path(buf, env->save_list) < 0) {
				report(LOG_ERR, "%s: %s: %m", config_file, buf);
				return NULL;
			}
		}

		for (service = target->service; service; service = service->next) {
			service->ip_addr = target->ip_addr;
			service->parent = target;

			(void)strcpy(cp, "/");
			(void)strcat(cp, service->datadir ? service->datadir :
				     fix_file_name(service->name));
			if (make_save_path(buf, service->save_list) < 0) {
				report(LOG_ERR, "%s: %s: %m", config_file, buf);
				return NULL;
			}
		}
	}

	free_save_list(first_save);
	free_method_list(first_method);
	free_trap_list(first_trap);
	free_var_list(global_var_list);

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return &config;
}

static int
insert_instance(var, instance, instance_len)
	VARIABLE *var;
	const oid *instance;
	int instance_len;
{
	int i, cnt = 0;

	/* sanity check */
	if (!var || !var->objid || !instance || instance_len < 1)
		return 0;

	/* is there room at objid to place instance */
	for (i = 0; i < var->objid_len; i++) {
		if (cnt || var->objid[i] == 0)
			cnt++;
	}
	if (cnt != instance_len) {
		yyerror("no room for Instance at $%s OID", var->name);
		return 0;
	}
	set_instance(var, instance, instance_len);
	return 1;
}

static int
add_object(chain, item)
	OBJECT **chain, *item;
{
	OBJECT *curr, *last = 0;
	METHOD *method;
	CHATSCRIPT *cs;
	REFERENCE *ref;

	/* sanity check */
	if (!chain || !item) {
		yyerror("add_object: internal error");
		return 0;
	}
	if (!item->name || !*item->name) {
		yyerror("object name is empty");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, item->name)) {
			yyerror("object name \"%s\" duplicated", item->name);
			return 0;
		}
		last = curr;
	}
	curr = (OBJECT *)malloc(sizeof(OBJECT));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(OBJECT));
	curr->next = 0;
	CLEAR_OBJECT(item);

	/* make object variable by method name */
	for (method = curr->method_list; method; method = method->next) {
		if (!add_var(&curr->var_list, method->name)) {
			yyerror("Out of memory");
			return 0;
		}
		/* bind variable references */
		for (cs = method->chatscript; cs; cs = cs->next) {
			for (ref = cs->var_ref; ref; ref = ref->next) {
				ref->var = add_var(&curr->var_list, ref->name);
				if (!ref->var) {
					yyerror("Out of memory");
					return 0;
				}
			}
		}
		if (method->snmpreq) {
			VARIABLE *var, *dict;
			const SNMP_MIB_ENTRY *mib;
			char *cp;
			oid instance[100];
			int instance_len;

			for (ref = method->snmpreq->var_ref; ref; ref = ref->next) {
				instance_len = 0;
				if ((cp = strchr(ref->name, '.')) != NULL) {
					*cp++ = '\0';
					instance_len = 100;
					str2oid(instance, &instance_len, cp);
				}
				var = get_var(curr->var_list, ref->name);
				if (var && var->objid) {
					if (instance_len &&
					    !insert_instance(var, instance, instance_len))
						return 0;
					ref->var = var;
					continue;
				}
				mib = get_mib_entry(ref->name, 0);
				if (mib && mib->objid) {
					if (!var && (var = add_var(&curr->var_list, ref->name)) == 0) {
						yyerror("Out of memory");
						return 0;
					}
					set_objid(var, mib->objid, mib->objid_len);
					if (instance_len &&
					    !insert_instance(var, instance, instance_len))
						return 0;
					ref->var = var;
					continue;
				}
				dict = get_var(global_var_list, ref->name);
				if (dict && dict->objid) {
					if (!var && (var = add_var(&curr->var_list, ref->name)) == 0) {
						yyerror("Out of memory");
						return 0;
					}
					set_objid(var, dict->objid, dict->objid_len);
					if (instance_len &&
					    !insert_instance(var, instance, instance_len))
						return 0;
					ref->var = var;
					continue;
				}
				yyerror("OID of variable $%s is unknown", ref->name);
				return 0;
			}
		}
	}
	if (last)
		last->next = curr;
	else	*chain = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return 1;
}

static int
add_interface(chain, item)
	INTERFACE **chain, *item;
{
	INTERFACE *curr, *last = 0;

	/* sanity check */
	if (!chain || !item) {
		yyerror("add_interface: internal error");
		return 0;
	}
	if ((!item->name || !*item->name) && item->index < 1) {
		yyerror("interface name (index) is empty");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (item->name) {
			if (curr->name && !strcasecmp(curr->name, item->name)) {
				yyerror("interface name \"%s\" duplicated", item->name);
				return 0;
			}
		} else if (item->index > 0) {
			if (curr->index > 0 && curr->index == item->index) {
				yyerror("interface index %d duplicated", item->index);
				return 0;
			}
		}
		last = curr;
	}
	curr = (INTERFACE *)malloc(sizeof(INTERFACE));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(INTERFACE));
	curr->next = 0;
	CLEAR_INTERFACE(item);
	if (last)
		last->next = curr;
	else	*chain = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return 1;
}

static int
add_bgp_as(chain, item)
	BGP_AS **chain, *item;
{
	BGP_AS *curr, *last = 0;
	BGP_PEER *peer;

	/* sanity check */
	if (!chain || !item || item->asn < 1) {
		yyerror("add_bgp_as: internal error");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (curr->asn == item->asn) {
			yyerror("bgp ASN %d duplicated", item->asn);
			return 0;
		}
		last = curr;
	}
	curr = (BGP_AS *)malloc(sizeof(BGP_AS));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(BGP_AS));
	curr->next = 0;
	CLEAR_BGP_AS(item);
	if (last)
		last->next = curr;
	else	*chain = curr;
	for (peer = curr->peer; peer; peer = peer->next) {
		peer->parent = curr;
		curr->handconf++;
	}
	return 1;
}

static int
add_bgp_peer(bgp, peer)
	BGP_AS *bgp;
	char *peer;
{
	struct in_addr ip_addr;
	BGP_PEER *curr, *last = 0;

	/* sanity check */
	if (!bgp) {
		yyerror("add_bgp_peer: internal error");
		return 0;
	}
	if (!peer || !*peer) {
		yyerror("bgp peer address is empty");
		return 0;
	}
	if (!gethostaddr(&ip_addr, peer)) {
		return 0;
	}
	/* check for duplicates */
	for (curr = bgp->peer; curr; curr = curr->next) {
		if (!memcmp(&curr->ip_addr, &ip_addr, sizeof(curr->ip_addr))) {
			yyerror("bgp peer \"%s\" duplicated", peer);
			return 0;
		}
		last = curr;
	}
	curr = (BGP_PEER *)malloc(sizeof(BGP_PEER));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memset(curr, 0, sizeof(BGP_PEER));
	curr->ip_addr = ip_addr;
	if (last)
		last->next = curr;
	else	bgp->peer = curr;
	return 1;
}

static int
add_env_mon(chain, item)
	ENV_MON **chain, *item;
{
	ENV_MON *curr, *last = 0;
	ENV_GAUGE *gauge;

	/* sanity check */
	if (!chain || !item || item->type < 1) {
		yyerror("add_env_mon: internal error");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (curr->type == item->type) {
			yyerror("env type %d duplicated", item->type);
			return 0;
		}
		last = curr;
	}
	curr = (ENV_MON *)malloc(sizeof(ENV_MON));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(ENV_MON));
	curr->next = 0;
	CLEAR_ENV_MON(item);
	if (last)
		last->next = curr;
	else	*chain = curr;
	for (gauge = curr->gauge; gauge; gauge = gauge->next) {
		gauge->parent = curr;
		curr->handconf++;
	}
	return 1;
}

static int
add_env_gauge(env, gauge)
	ENV_MON *env;
	char *gauge;
{
	ENV_GAUGE *curr, *last = 0;

	/* sanity check */
	if (!env) {
		yyerror("add_env_gauge: internal error");
		return 0;
	}
	if (!gauge || !*gauge) {
		yyerror("env gauge name is empty");
		return 0;
	}
	/* check for duplicates */
	for (curr = env->gauge; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, gauge)) {
			yyerror("env gauge \"%s\" duplicated", gauge);
			return 0;
		}
		last = curr;
	}
	curr = (ENV_GAUGE *)malloc(sizeof(ENV_GAUGE));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memset(curr, 0, sizeof(ENV_GAUGE));
	curr->name = gauge;
	if (last)
		last->next = curr;
	else	env->gauge = curr;
	return 1;
}

#define	MAX_ARGS	20

static char *
insert_args(src, av, ac)
	register char *src;
	char *av[];
	int ac;
{
	register char *dst;
	int idx;
	char *vb, *ve, dst_buf[1024];

	dst = dst_buf;
	while (src && *src && dst < &dst_buf[sizeof(dst_buf)-1]) {
		if (*src != '$') {
			*dst++ = *src++;
			continue;
		}
		vb = src + 1;
		for (ve = vb; *ve && isdigit(*ve); ve++);
		if (ve - vb != 1) {
			*dst++ = *src++;
			continue;
		}
		idx = *vb - '0';
		if (idx >= ac) {
			yyerror("missing argument for $%d at \"%s\"", idx, av[0]);
			return NULL;
		}
		vb = av[idx];
		if (!vb) {
			yyerror("missing argument for $%d at \"%s\"", idx, av[0]);
			return NULL;
		}
		if (dst + strlen(vb) + 1 >= &dst_buf[sizeof(dst_buf)-1]) {
			yyerror("argument $%d at \"%s\" too long", idx, av[0]);
			return NULL;
		}
		dst += strlen(strcpy(dst, vb));
		src = ve;
	}
	*dst = '\0';
	if ((dst = strdup(dst_buf)) == NULL)
		yyerror("Out of memory");
	return dst;
}

/* make copy of chatscript and return it */
static CHATSCRIPT *
dup_chatscript(self, arg, orig)
	char *self, *arg;
	CHATSCRIPT *orig;
{
	CHATSCRIPT *first, *curr, *last;
	yre_t prog;
	char *yre_error, arg_list[1024], *av[MAX_ARGS+2];
	int ac = 0;

	/* make array of argument vectors */
	if (arg) {
		(void)strncpy(arg_list, arg, sizeof(arg_list));
		arg_list[sizeof(arg_list)-1] = '\0';
	} else	arg_list[0] = '\0';
	av[ac++] = self;
	ac += make_argv(arg_list, (char ***)&av[ac], MAX_ARGS);
	av[ac] = NULL;

	for (first = NULL; orig; orig = orig->next) {
		if ((curr = (CHATSCRIPT *)malloc(sizeof(CHATSCRIPT))) == NULL) {
			yyerror("Out of memory");
			return NULL;
		}
		memcpy(curr, orig, sizeof(CHATSCRIPT));
		if (curr->send) {
			if ((curr->send = insert_args(curr->send, av, ac)) == NULL)
				return NULL;
		}
		if (curr->expect) {
			if ((curr->expect = insert_args(curr->expect, av, ac)) == NULL)
				return NULL;
			if (curr->expect && *curr->expect) {
				if ((yre_error = yre_comp(&prog, curr->expect)) != NULL) {
					yyerror("\"%s\": %s", curr->expect, yre_error);
					return NULL;
				}
			}
		}
		if (curr->var_ref) {
			if ((curr->var_ref = dup_var_ref(curr->var_ref)) == NULL) {
				yyerror("Out of memory");
				return NULL;
			}
		}
		curr->next = NULL;

		if (first)
			last->next = curr;
		else	first = curr;
		last = curr;
	}
	return first;
}

/* add chatscript to chain */
static CHATSCRIPT *
add_chatscript()
{
	CHATSCRIPT *curr;
	static CHATSCRIPT *last;

	if ((curr = (CHATSCRIPT *)malloc(sizeof(CHATSCRIPT))) == NULL) {
		yyerror("Out of memory");
		return NULL;
	}
	memcpy(curr, &chatscript, sizeof(CHATSCRIPT));
	curr->next = NULL;
	CLEAR_CHATSCRIPT(&chatscript);

	if (first_chatscript)
		last->next = curr;
	else	first_chatscript = curr;
	last = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return first_chatscript;
}

/* find save in chain by name or simply return last chain save */
static SAVE *
find_save(chain, name)
	SAVE *chain;
	char *name;
{
	SAVE *curr, *last = NULL;

	for (curr = chain; curr; curr = curr->next) {
		if (name && !strcasecmp(curr->name, name))
			break;
		last = curr;
	}
	return (name ? curr : last);
}

/* make a copy of save and return it */
static SAVE *
dup_save_args(old, argument)
	SAVE *old;
	char *argument;
{
	SAVE *new;
	char arg_list[1024], *av[MAX_ARGS+2];
	int ac = 0;

	if ((new = (SAVE *)malloc(sizeof(SAVE))) == NULL) {
		yyerror("Out of memory");
		return NULL;
	}
	memcpy(new, old, sizeof(SAVE));
	if (new->name && (new->name = strdup(new->name)) == NULL) {
		yyerror("Out of memory");
		return NULL;
	}
	/* make array of argument vectors */
	if (argument) {
		(void)strncpy(arg_list, argument, sizeof(arg_list));
		arg_list[sizeof(arg_list)-1] = '\0';
	} else	arg_list[0] = '\0';
	av[ac++] = new->name;
	ac += make_argv(arg_list, (char ***)&av[ac], MAX_ARGS);
	av[ac] = NULL;

	if ((new->file = insert_args(new->file, av, ac)) == NULL)
		return NULL;
	if (new->saving && (new->saving = insert_args(new->saving, av, ac)) == NULL)
		return NULL;
	if (new->state && (new->state = insert_args(new->state, av, ac)) == NULL)
		return NULL;
	if (new->when && (new->when = insert_args(new->when, av, ac)) == NULL)
		return NULL;
	if (new->when_fmt && (new->when_fmt = insert_args(new->when_fmt, av, ac)) == NULL)
		return NULL;

	new->next = NULL;
	return new;
}

static int
add_save(chain, item)
	SAVE **chain, *item;
{
	SAVE *curr, *last = 0;

	/* sanity check */
	if (!chain || !item) {
		yyerror("add_save: internal error");
		return 0;
	}
	if (!item->name || !*item->name) {
		yyerror("save name is empty");
		return 0;
	}
	if (!item->method) {
		yyerror("save method unspecified");
		return 0;
	}
	if (!item->file) {
		yyerror("save file unspecified");
		return 0;
	}
	if (!item->saving && !item->state && !item->when) {
		yyerror("save action must be specified");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, item->name)) {
			yyerror("save name \"%s\" duplicated", item->name);
			return 0;
		}
		last = curr;
	}
	curr = (SAVE *)malloc(sizeof(SAVE));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(SAVE));
	curr->next = 0;
	CLEAR_SAVE(item);
	if (last)
		last->next = curr;
	else	*chain = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return 1;
}

/* find method in chain by name or simply return last chain method */
static METHOD *
find_method(chain, name)
	METHOD *chain;
	char *name;
{
	METHOD *curr, *last = NULL;

	for (curr = chain; curr; curr = curr->next) {
		if (name && !strcasecmp(curr->name, name))
			break;
		last = curr;
	}
	return (name ? curr : last);
}

/* make a copy of method and return it */
static METHOD *
dup_method_args(old, argument)
	METHOD *old;
	char *argument;
{
	METHOD *new;

	if ((new = (METHOD *)malloc(sizeof(METHOD))) == NULL) {
		yyerror("Out of memory");
		return NULL;
	}
	memcpy(new, old, sizeof(METHOD));
	if (new->name && (new->name = strdup(new->name)) == 0) {
		yyerror("Out of memory");
		return 0;
	}
	if (new->address && (new->address = strdup(new->address)) == 0) {
		yyerror("Out of memory");
		return 0;
	}
	if (argument)
		new->argument = argument;
	else if (new->argument)
		new->argument = strdup(new->argument);
	if (new->chatscript) {
		new->chatscript = dup_chatscript(new->name, new->argument, new->chatscript);
		if (!new->chatscript) return NULL;
	}
	if (new->snmpreq) {
		if ((new->snmpreq = dup_snmp_request(new->snmpreq)) == NULL) {
			yyerror("Out of memory");
			return NULL;
		}
	}
	new->sd = 0;
	new->next = 0;
	return new;
}

static int
add_method(chain, item)
	METHOD **chain, *item;
{
	METHOD *curr, *last = 0;

        /* sanity check */
	if (!chain || !item) {
		yyerror("add_method: internal error");
		return 0;
	}
	if (!item->name || !*item->name) {
		yyerror("method name is empty");
		return 0;
	}
	if (!item->protocol) {
		yyerror("method protocol unspecified");
		return 0;
	}
	if (!item->rport && (item->protocol == IPPROTO_TCP ||
			     item->protocol == IPPROTO_UDP)) {
		yyerror("method port unspecified");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (chain == &first_method &&
		    !strcasecmp(curr->name, item->name)) {
			yyerror("method name \"%s\" duplicated", item->name);
			return 0;
		}
		last = curr;
	}
	curr = (METHOD *)malloc(sizeof(METHOD));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(METHOD));
	curr->next = 0;
	if (curr->timeout < 1) {
		if (object.timeout)
			curr->timeout = object.timeout;
		else if (config.timeout)
			curr->timeout = config.timeout;
		else	curr->timeout = TIMEOUT_DEFAULT;
	}
	if (curr->retries < 1) {
		if (object.retries)
			curr->retries = object.retries;
		else if (config.retries)
			curr->retries = config.retries;
		else	curr->retries = RETRIES_DEFAULT;
	}
	if (curr->protocol == IPPROTO_ICMP &&
	    (!curr->ping_expect || curr->ping_expect > curr->ping_send)) {
		curr->ping_expect = curr->ping_send;
	}
	CLEAR_METHOD(item);
	if (last)
		last->next = curr;
	else	*chain = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return 1;
}

/* find trap in chain by name or simply return last chain trap */
static TRAP *
find_trap(chain, name)
	TRAP *chain;
	char *name;
{
	TRAP *curr, *last = NULL;

	for (curr = chain; curr; curr = curr->next) {
		if (name && !strcasecmp(curr->name, name))
			break;
		last = curr;
	}
	return (name ? curr : last);
}

/* make a copy of trap and return it */
static TRAP *
dup_trap(old, community)
	TRAP *old;
	char *community;
{
	TRAP *new;

	if ((new = (TRAP *)malloc(sizeof(TRAP))) == NULL) {
		yyerror("Out of memory");
		return NULL;
	}
	memcpy(new, old, sizeof(TRAP));
	if (new->name) new->name = strdup(new->name);
	if (community)
		new->community = community;
	else if (new->community)
		new->community = strdup(new->community);
	if (new->objid && new->objid_len) {
		if ((new->objid = (oid *)malloc(new->objid_len * sizeof(oid))) == NULL) {
			yyerror("Out of memory");
			return NULL;
		}
		memcpy(new->objid, old->objid, new->objid_len * sizeof(oid));
	}
	new->next = NULL;
	return new;
}

static int
add_trap(chain, item)
	TRAP **chain, *item;
{
	TRAP *curr, *last = 0;

	/* sanity check */
	if (!chain || !item) {
		yyerror("add_trap: internal error");
		return 0;
	}
	if (!item->name || !*item->name) {
		yyerror("trap name is empty");
		return 0;
	}
	if (item->type && !item->objid) { /* Enterprise Specific trap */
		yyerror("trap enterprise unspecified");
		return 0;
	}
	/* check for duplicates */
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, item->name)) {
			yyerror("trap name \"%s\" duplicated", item->name);
			return 0;
		}
		last = curr;
	}
	curr = (TRAP *)malloc(sizeof(TRAP));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, item, sizeof(TRAP));
	curr->next = 0;
	if (!curr->community) {
		curr->community = strdup("public");
		if (!curr->community) {
			yyerror("Out of memory");
			return 0;
		}
	}
	CLEAR_TRAP(item);
	if (last)
		last->next = curr;
	else	*chain = curr;

#ifdef	DEBUG_MEMORY
	assert(get_cksum());
#endif
	return 1;
}

static int
add_variable(chain, var)
	VARIABLE **chain, *var;
{
	VARIABLE *curr, *last = 0;

	/* should not happen */
	if (!chain || !var || !var->name || !*var->name) {
		yyerror("add_variable: internal error");
		return 0;
	}
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, var->name)) {
			yyerror("variable $%s redefined", var->name);
			return 0;
		}
		last = curr;
	}
	if (last)
		last->next = var;
	else	*chain = var;
	return 1;
}

static int
gethostaddr(addr, host)
	struct in_addr *addr;
	const char *host;
{
	if ((addr->s_addr = inet_addr(host)) == -1) {
		struct hostent *hp;
#ifdef	HAVE_PTHREAD
		pthread_mutex_lock(&gethostby_lock);
#endif
		if ((hp = gethostbyname(host)) == NULL) {
			yyerror("\"%s\": Unknown host", host);
			return 0;
		}
		memcpy(addr, hp->h_addr, hp->h_length);
#ifdef	HAVE_PTHREAD
		pthread_mutex_unlock(&gethostby_lock);
#endif
	}
	return 1;
}

static int
add_var_ref(list, chain, name)
	VARIABLE **list;
	REFERENCE **chain;
	char *name;
{
	REFERENCE *curr, *last = 0;

	/* sanity check */
	if (!chain || !name || !*name) {
		yyerror("add_var_ref: internal error");
		return 0;
	}
	if (list && !get_var(*list, name)) {
		yyerror("variable $%s undefined", name);
		return 0;
	}
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, name)) {
			yyerror("variable reference $%s duplicated", name);
			return 0;
		}
		last = curr;
	}
	curr = (REFERENCE *)malloc(sizeof(REFERENCE));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memset(curr, 0, sizeof(REFERENCE));
	curr->name = strdup(name);
	if (!curr->name) {
		free(curr);
		yyerror("Out of memory");
		return 0;
	}
	if (last)
		last->next = curr;
	else	*chain = curr;
	return 1;
}

static GROUP_LIST *
find_group_list(list, name)
	GROUP_LIST *list;
	char *name;
{
	GROUP_LIST *curr;
	for (curr = list; curr; curr = curr->next) {
		if (name && !strcasecmp(curr->name, name))
			return curr;
	}
	return 0;
}

static int
add_group_list(chain, group)
	GROUP_LIST **chain, *group;
{
	GROUP_LIST *curr, *last = 0;

	/* sanity check */
	if (!chain || !group || !group->name || !*group->name) {
		yyerror("add_group_list: internal error");
		return 0;
	}
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, group->name)) {
			yyerror("group name \"%s\" duplicated", group->name);
			return 0;
		}
		last = curr;
	}
	curr = (GROUP_LIST *)malloc(sizeof(GROUP_LIST));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memcpy(curr, group, sizeof(GROUP_LIST));
	curr->next = 0;
	CLEAR_GROUP_LIST(group);
	if (last)
		last->next = curr;
	else	*chain = curr;
	return 1;
}

static int
add_group_entry(list, permit, patt)
	GROUP_LIST *list;
	int permit;
	char *patt;
{
	yre_t prog;
	char *yre_error;
	GROUP_ENTRY *curr, *last;

	if (!list) { /* for backward compatibility */
		char *name = "Default Group";
		list = find_group_list(config.group_list, name);
		if (!list) {
			GROUP_LIST ml;
			memset(&ml, 0, sizeof(GROUP_LIST));
			ml.name = strdup(name);
			if (!add_group_list(&config.group_list, &ml))
				return 0;
			list = find_group_list(config.group_list, name);
			if (!list) {
				yyerror("add_group_entry: internal error");
				return 0;
			}
			if (!add_group_ref(list, &config.ns_acl, strdup(name)))
				return 0;
		}
	}
	if (!patt || !*patt) {
		yyerror("group pattern is empty");
		return 0;
	}
	if ((yre_error = yre_comp(&prog, patt)) != NULL) {
		yyerror("\"%s\": %s", patt, yre_error);
		return 0;
	}
	last = 0;
	for (curr = list->entry; curr; curr = curr->next) {
		if (!strcmp(curr->pattern, patt)) {
			yyerror("group pattern \"%s\" duplicated", patt);
			return 0;
		}
		last = curr;
	}
	curr = (GROUP_ENTRY *)malloc(sizeof(GROUP_ENTRY));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memset(curr, 0, sizeof(GROUP_ENTRY));
	curr->permit = permit;
	curr->pattern = patt;
	if (last)
		last->next = curr;
	else	list->entry = curr;
	return 1;
}

static int
add_group_ref(list, chain, name)
	GROUP_LIST *list;
	GROUP_REF **chain;
	char *name;
{
	GROUP_REF *curr, *last = 0;

	/* sanity check */
	if (!chain || !name || !*name) {
		yyerror("add_group_ref: internal error");
		return 0;
	}
	if (!find_group_list(list, name)) {
		yyerror("group name \"%s\" undefined", name);
		return 0;
	}
	for (curr = *chain; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, name)) {
			yyerror("group reference \"%s\" duplicated", name);
			return 0;
		}
		last = curr;
	}
	curr = (GROUP_REF *)malloc(sizeof(GROUP_REF));
	if (!curr) {
		yyerror("Out of memory");
		return 0;
	}
	memset(curr, 0, sizeof(GROUP_REF));
	curr->name = name;
	if (last)
		last->next = curr;
	else	*chain = curr;
	return 1;
}

%}

%union
{
	unsigned long number;
	char *string;
	SAVE *save;
	METHOD *method;
	TRAP *trap;
	VARIABLE *variable;
}

/* Lexical analyzer return values */
%token TOKEN_ROOTDIR
%token TOKEN_TIMEFMT
%token TOKEN_POLLING
%token TOKEN_SAVING
%token TOKEN_TIMEOUT
%token TOKEN_RETRIES

%token TOKEN_GROUP
%token TOKEN_PERMIT
%token TOKEN_DENY

%token TOKEN_NETSTATE
%token TOKEN_PORT

%token TOKEN_SAVE
%token TOKEN_FILE
%token TOKEN_EXEC
%token TOKEN_PIPE
%token TOKEN_DATA
%token TOKEN_STATE
%token TOKEN_WHEN

%token TOKEN_OBJECT
%token TOKEN_ADDRESS
%token TOKEN_DESCRIPTION
%token TOKEN_SERVICE
%token TOKEN_INTERFACE
%token TOKEN_DATADIR
%token TOKEN_BANDWIDTH
%token TOKEN_BGP
%token TOKEN_PEER
%token TOKEN_ENV
%token TOKEN_TEMP
%token TOKEN_VOLT
%token TOKEN_FAN
%token TOKEN_PS
%token TOKEN_GAUGE

%token TOKEN_METHOD
%token TOKEN_TCP
%token TOKEN_UDP
%token TOKEN_ICMP
%token TOKEN_LOCALPORT
%token TOKEN_CHATSCRIPT
%token TOKEN_SEND
%token TOKEN_EXPECT
%token TOKEN_SIZE
%token TOKEN_BATCH
%token METHOD_PING
%token METHOD_SNMP
%token METHOD_ROUTER
%token METHOD_DNS
%token METHOD_RADIUS
%token METHOD_TACACS
%token TOKEN_V2

%token TOKEN_TRAP
%token TOKEN_SOURCECHECK
%token TOKEN_COMMUNITY
%token TOKEN_ENTERPRISE
%token TOKEN_SPECIFIC
%token TRAP_GENERIC

%token <number> TOKEN_NUMBER
%token <string> TOKEN_STRING
%token <string> VAR_NAME
%token <string> VAR_OID

%type <number> optional_number
%type <string> optional_string quoted_string legal_string multiline_string
%type <save> get_save
%type <method> get_method
%type <trap> get_trap
%type <variable> make_variable init_variable

%%

config_file:	/* empty */
	|	config_file config_statement
	;

config_statement:
		TOKEN_ROOTDIR quoted_string
		{
			register char *cp;

			if (config.rootdir) {
				yyerror("rootdir statement duplicated");
				YYABORT;
			}
			/* strip off trailing slashes */
			cp = $2 + strlen($2) - 1;
			while (cp > $2 && *cp == '/')
				*cp-- = '\0';
			if (cp > $2) {
				config.rootdir = $2;
			} else {
				yyerror("bad rootdir specification");
				YYABORT;
			}
		}
	|	TOKEN_TIMEFMT quoted_string
		{
			if (config.timefmt) {
				yyerror("timefmt statement duplicated");
				YYABORT;
			}
			config.timefmt = $2;
		}
	|	TOKEN_POLLING TOKEN_NUMBER
		{
			if ($2 < POLLING_MIN) {
				yyerror("polling period too small (min %d sec.)", POLLING_MIN);
				YYABORT;
			}
			config.polling = $2;
		}
	|	TOKEN_SAVING TOKEN_NUMBER
		{
			if ($2 < POLLING_MIN) {
				yyerror("saving period too small (min %d sec.)",
					POLLING_MIN);
				YYABORT;
			}
			config.saving = $2;
		}
	|	TOKEN_TIMEOUT TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid timeout value (min 1 max %d sec.)",
					POLLING_MIN);
				YYABORT;
			}
			config.timeout = $2;
		}
	|	TOKEN_RETRIES TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid retries number (min 1 max %d)",
					POLLING_MIN);
				YYABORT;
			}
			config.retries = $2;
		}
	|	TOKEN_NETSTATE '{' netstate_config '}'
		{
			if (!config.ns_port) {
				yyerror("netstate port undefined");
				YYABORT;
			}
			if (!config.ns_acl) {
				yyerror("netstate group list undefined");
				YYABORT;
			}
			if (!config.ns_timo)
				config.ns_timo = TIMEOUT_DEFAULT;
		}
	|	TOKEN_PORT TOKEN_NUMBER '{' netstate_config '}'
		{
			/* for backward compatibility */
			if (config.ns_port) {
				yyerror("port statement duplicated");
				YYABORT;
			}
			if (!config.ns_timo)
				config.ns_timo = TIMEOUT_DEFAULT;
			config.ns_port = $2;
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&first_save, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_METHOD legal_string '{' method_config '}'
		{
			method.name = $2;
			if (!add_method(&first_method, &method)) {
				YYABORT;
			}
		}
	|	TOKEN_TRAP TOKEN_DENY
		{
			config.enable_traps = -1;
		}
	|	TOKEN_TRAP TOKEN_SOURCECHECK
		{
			config.source_traps = 1;
		}
	|	TOKEN_TRAP legal_string '{' trap_config '}'
		{
			trap.name = $2;
			trap.type = 6;	/* Enterprise Specific trap */
			if (!add_trap(&first_trap, &trap)) {
				YYABORT;
			}
		}
	|	TOKEN_GROUP legal_string '{' group_list_config '}'
		{
			if (!group_list.entry) {
				yyerror("group list \"%s\" is empty", $2);
				YYABORT;
			}
			group_list.name = $2;
			if (!add_group_list(&config.group_list, &group_list)) {
				YYABORT;
			}
		}
	|	TOKEN_OBJECT legal_string '{' object_config '}'
		{
			if (!object.address) {
				yyerror("object address unspecified");
				YYABORT;
			}
/*			if ((object.interface || object.ifgroup ||
			     object.bgp || object.env) &&
			    !find_method(object.method_list, "ROUTER")) {
				yyerror("object lacks method ROUTER");
				YYABORT;
			}
*/
			if (!object.method_list)
				object.polling = 0;
			else if (!object.polling)
				object.polling = config.polling;

			if (!object.saving)
				object.saving = config.saving;

			if (object.polling > 0 && object.saving > 0) {
				if (object.polling > object.saving) {
					yyerror("saving rate less then polling");
					YYABORT;
				}
				/* can be saving synchronized with polling? */
				if ((object.saving % object.polling) == 0)
					object.sync = object.saving / object.polling;
			}

			if (!object.timeout) {
				if (config.timeout)
					object.timeout = config.timeout;
				else	object.timeout = TIMEOUT_DEFAULT;
			}
			if (!object.retries) {
				if (config.retries)
					object.retries = config.retries;
				else	object.retries = RETRIES_DEFAULT;
			}

			object.name = $2;
			if (!add_object(&config.target, &object)) {
				YYABORT;
			}
		}
	|	make_variable
		{
			if (!add_variable(&global_var_list, $1)) {
				YYABORT;
			}
		}
	;

netstate_config:	/* empty */
	|	netstate_config netstate_statement
	;

netstate_statement:
		TOKEN_PORT TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > 65535) {
				yyerror("port number out of range");
				YYABORT;
			}
			if (config.ns_port) {
				yyerror("port statement duplicated");
				YYABORT;
			}
			config.ns_port = $2;
		}
	|	TOKEN_TIMEOUT TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid timeout value (min 1 max %d sec.)",
					POLLING_MIN);
				YYABORT;
			}
			config.ns_timo = $2;
		}
	|	TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &config.ns_acl, $2)) {
				YYABORT;
			}
		}
	|	TOKEN_PERMIT quoted_string
		{
			/* for backward compatibility */
			if (!add_group_entry(0, 1, $2)) {
				YYABORT;
			}
		}
	|	TOKEN_DENY quoted_string
		{
			/* for backward compatibility */
			if (!add_group_entry(0, 0, $2)) {
				YYABORT;
			}
		}
	;

save_config:	/* empty */
	|	save_config save_statement
	;

save_statement:
		TOKEN_FILE quoted_string
		{
			if (save.method) {
				yyerror("save method duplicated");
				YYABORT;
			}
			save.method = SAVE_FILE;
			save.file = $2;
		}
	|	TOKEN_EXEC quoted_string
		{
			if (save.method) {
				yyerror("save method duplicated");
				YYABORT;
			}
			save.method = SAVE_EXEC;
			save.file = $2;
		}
	|	TOKEN_PIPE quoted_string
		{
			if (save.method) {
				yyerror("save method duplicated");
				YYABORT;
			}
			save.method = SAVE_PIPE;
			save.file = $2;
		}
	|	TOKEN_DATA multiline_string
		{
			if (save.saving) {
				yyerror("save data format duplicated");
				YYABORT;
			}
			save.saving = $2;
		}
	|	TOKEN_STATE multiline_string
		{
			if (save.state) {
				yyerror("save state format duplicated");
				YYABORT;
			}
			save.state = $2;
		}
	|	TOKEN_WHEN multiline_string TOKEN_NUMBER optional_string
		{
			if (save.when) {
				yyerror("save when condition duplicated");
				YYABORT;
			}
			save.when = $2;
			save.delay = $3;
			save.when_fmt = $4;
		}
	;

method_config:	/* empty */
	|	method_config method_statement
	;

method_statement:
		TOKEN_ADDRESS quoted_string
		{
			if (method.address) {
				yyerror("method address duplicated");
				YYABORT;
			}
			if (!gethostaddr(&method.ip_addr, $2)) {
				YYABORT;
			}
			method.address = $2;
		}
	|	TOKEN_TCP
		{
			if (method.protocol) {
				yyerror("method protocol duplicated");
				YYABORT;
			}
			method.protocol = IPPROTO_TCP;
			method.init = tcp_init;
			method.start = tcp_start;
			method.stop = tcp_stop;
		}
	|	TOKEN_UDP
		{
			if (method.protocol) {
				yyerror("method protocol duplicated");
				YYABORT;
			}
			method.protocol = IPPROTO_UDP;
			method.init = udp_init;
			method.start = udp_start;
			method.stop = udp_stop;
		}
	|	TOKEN_ICMP
		{
			if (method.protocol) {
				yyerror("method protocol duplicated");
				YYABORT;
			}
			method.protocol = IPPROTO_ICMP;
			method.rport = sizeof(TIMEVAL *); /* min data size */
			method.init = echo_init;
			method.start = echo_start;
			method.stop = echo_stop;
		}
	|	TOKEN_PORT TOKEN_NUMBER
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_TCP &&
			    method.protocol != IPPROTO_UDP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if (method.rport) {
				yyerror("port statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > 65535) {
				yyerror("port number out of range");
				YYABORT;
			}
			method.rport = $2;
		}
	|	TOKEN_LOCALPORT TOKEN_NUMBER optional_number
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_TCP &&
			    method.protocol != IPPROTO_UDP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if (method.lport_min) {
				yyerror("localport statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > 65535) {
				yyerror("localport number out of range");
				YYABORT;
			}
			method.lport_min = $2;
			if ($3 != -1) {
				method.lport_max = $3;
				if (method.lport_max < method.lport_min) {
					unsigned short swap_port;
					swap_port = method.lport_max;
					method.lport_max = method.lport_min;
					method.lport_min = swap_port;
				}
			}
		}
	|	TOKEN_TIMEOUT TOKEN_NUMBER
		{
			if (method.timeout) {
				yyerror("timeout statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid timeout value (min 1 max %d sec.)",
					POLLING_MIN);
				YYABORT;
			}
			method.timeout = $2;
		}
	|	TOKEN_RETRIES TOKEN_NUMBER
		{
			if (method.retries) {
				yyerror("retries statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid retries number (min 1 max %d)",
					POLLING_MIN);
				YYABORT;
			}
			method.retries = $2;
		}
	|	TOKEN_CHATSCRIPT '{' chatscript_config '}'
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_TCP &&
			    method.protocol != IPPROTO_UDP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if ((!chatscript.send || !*chatscript.send) &&
			    (!chatscript.expect || !*chatscript.expect)) {
				yyerror("chatscript defined but empty");
				YYABORT;
			}
			if (method.chatscript) {
				yyerror("method chat-script duplicated");
				YYABORT;
			}
			if ((method.chatscript = add_chatscript()) == NULL) {
				YYABORT;
			}
			first_chatscript = NULL;
		}
	|	TOKEN_SIZE TOKEN_NUMBER
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_ICMP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if ($2 < sizeof(TIMEVAL *) || $2 > 1500) {
				yyerror("invalid size value (min %d max 1500 bytes)",
					sizeof(TIMEVAL *));
				YYABORT;
			}
			method.rport = $2;
		}
	|	TOKEN_SEND TOKEN_NUMBER
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_ICMP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if (method.ping_send) {
				yyerror("send statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > 100) {
				yyerror("invalid send value (min 1 max 100)");
				YYABORT;
			}
			method.ping_send = $2;
		}
	|	TOKEN_EXPECT TOKEN_NUMBER
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_ICMP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if (method.ping_expect) {
				yyerror("expect statement duplicated");
				YYABORT;
			}
			if ($2 < 1 || $2 > 100) {
				yyerror("invalid expect value (min 1 max 100)");
				YYABORT;
			}
			method.ping_expect = $2;
		}
	|	TOKEN_BATCH TOKEN_NUMBER
		{
			if (method.protocol &&
			    method.protocol != IPPROTO_UDP) {
				yyerror("no suitable method protocol");
				YYABORT;
			}
			if ($2 < 1 || $2 > 100) {
				yyerror("invalid batch value (min 1 max 100)");
				YYABORT;
			}
			method.router_batch = $2;
		}
	|	VAR_NAME
		{
			if (!method.snmpreq) {
				method.snmpreq = (SNMP_REQ *)malloc(sizeof(SNMP_REQ));
				if (!method.snmpreq) {
					yyerror("Out of memory");
					YYABORT;
				}
				memset(method.snmpreq, 0, sizeof(SNMP_REQ));
			}
			if (!add_var_ref(0, &method.snmpreq->var_ref, $1)) {
				YYABORT;
			}
		}
	;

trap_config:	/* empty */
	|	trap_config trap_statement
	;

trap_statement:
		TOKEN_COMMUNITY quoted_string
		{
			if (trap.community) {
				yyerror("trap community duplicated");
				YYABORT;
			}
			trap.community = $2;
		}
	|	TOKEN_ENTERPRISE VAR_OID
		{
			oid buf[100];
			int len = 100;
			if (trap.objid) {
				yyerror("trap enterprise duplicated");
				YYABORT;
			}
			str2oid(buf, &len, $2);
			if (len < 4) {
				yyerror("bad trap enterprise");
				YYABORT;
			}
			trap.objid = (oid *)malloc(len * sizeof(oid));
			if (trap.objid) {
				memcpy(trap.objid, buf, len * sizeof(oid));
				trap.objid_len = len;
			}
		}
	|	TOKEN_SPECIFIC TOKEN_NUMBER
		{
			if (trap.specific) {
				yyerror("trap specific code duplicated");
				YYABORT;
			}
			trap.specific = $2;
		}
	;

group_list_config:	/* empty */
	|	group_list_config group_list_statement
	;

group_list_statement:
		TOKEN_PERMIT quoted_string
		{
			if (!add_group_entry(&group_list, 1, $2)) {
				YYABORT;
			}
		}
	|	TOKEN_DENY quoted_string
		{
			if (!add_group_entry(&group_list, 0, $2)) {
				YYABORT;
			}
		}
	;

chatscript_config:	/* empty */
	|	chatscript_config chatscript_statement
	;

chatscript_statement:
		TOKEN_SEND multiline_string
		{
			if ((chatscript.send || chatscript.expect) &&
			    !add_chatscript()) {
				YYABORT;
			}
			chatscript.send = $2;
		}
	|	TOKEN_EXPECT quoted_string
		{
			if (chatscript.expect && !add_chatscript()) {
				YYABORT;
			}
			chatscript.expect = $2;
			if (*$2) {
				yre_t prog;
				char *yre_error;

				if ((yre_error = yre_comp(&prog, $2)) != NULL) {
					yyerror("\"%s\": %s", $2, yre_error);
					YYABORT;
				}
			}
		}
	|	TOKEN_EXPECT quoted_string '{' reference_config '}'
		{
			if (chatscript.expect && !add_chatscript()) {
				YYABORT;
			}
			chatscript.expect = $2;
			if (*$2) {
				yre_t prog;
				char *yre_error;

				if ((yre_error = yre_comp(&prog, $2)) != NULL) {
					yyerror("\"%s\": %s", $2, yre_error);
					YYABORT;
				}
			} else {
				yyerror("empty regex pattern");
				YYABORT;
			}
			if (!first_reference) {
				yyerror("empty variable reference");
				YYABORT;
			}
			chatscript.var_ref = first_reference;
			first_reference = NULL;
		}
	;

reference_config:	/* empty */
	|	reference_config reference_statement
	;

reference_statement:
		VAR_NAME
		{
			if (!add_var_ref(0, &first_reference, $1)) {
				YYABORT;
			}
		}
	;

object_config:	/* empty */
	|	object_config object_statement
	;

object_statement:
		TOKEN_DESCRIPTION quoted_string
		{
			if (object.descr) {
				yyerror("object description duplicated");
				YYABORT;
			}
			object.descr = $2;
		}
	|	TOKEN_DATADIR legal_string
		{
			if (object.datadir) {
				yyerror("object datadir duplicated");
				YYABORT;
			}
			object.datadir = $2;
		}
	|	TOKEN_ADDRESS quoted_string
		{
			if (object.address) {
				yyerror("object address duplicated");
				YYABORT;
			}
			if (!gethostaddr(&object.ip_addr, $2)) {
				YYABORT;
			}
			object.address = $2;
		}
	|	TOKEN_POLLING TOKEN_NUMBER
		{
			if (object.polling) {
				yyerror("object polling duplicated");
				YYABORT;
			}
			if ($2 > 0 && $2 < POLLING_MIN) {
				yyerror("polling period too small (min %d sec.)", POLLING_MIN);
				YYABORT;
			}
			if ($2 > 0)
				object.polling = $2;
			else	object.polling = -1; /* disable polling */
		}
	|	TOKEN_SAVING TOKEN_NUMBER
		{
			if (object.saving) {
				yyerror("object saving period duplicated");
				YYABORT;
			}
			if ($2 > 0 && $2 < POLLING_MIN) {
				yyerror("object saving period too small (min %d sec.)", POLLING_MIN);
				YYABORT;
			}
			if ($2 > 0)
				object.saving = $2;
			else	object.saving = -1; /* disable saving */
		}
	|	TOKEN_TIMEOUT TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid timeout value (min 1 max %d sec.)",
					POLLING_MIN);
				YYABORT;
			}
			object.timeout = $2;
		}
	|	TOKEN_RETRIES TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > POLLING_MIN) {
				yyerror("invalid retries number (min 1 max %d)",
					POLLING_MIN);
				YYABORT;
			}
			object.retries = $2;
		}
	|	TOKEN_NETSTATE TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &object.ns_acl, $3)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&object.save_list, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE get_save
		{
			if (!add_save(&object.save_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_METHOD get_method '{' method_config '}'
		{
			if (method.address) {
				if ($2->address) free($2->address);
				$2->address = method.address;
				$2->ip_addr = method.ip_addr;
			}
			if (method.rport > 0)
				$2->rport = method.rport;
			if (method.timeout > 0)
				$2->timeout = method.timeout;
			if (method.retries > 0)
				$2->retries = method.retries;
			if (method.lport_min > 0)
				$2->lport_min = method.lport_min;
			if (method.lport_max > 0)
				$2->lport_max = method.lport_max;
			if (method.snmpreq) {
				if ($2->snmpreq) {
					free_var_ref($2->snmpreq->var_ref);
					free($2->snmpreq);
				}
				$2->snmpreq = method.snmpreq;
			}
			CLEAR_METHOD(&method);

			if (!add_method(&object.method_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_METHOD get_method
		{
			if (!add_method(&object.method_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_TRAP legal_string '{' trap_config '}'
		{
			trap.name = $2;
			trap.type = 6;	/* Enterprise Specific trap */
			if (!add_trap(&object.trap_list, &trap)) {
				YYABORT;
			}
			if (config.enable_traps != -1)
				config.enable_traps++;
		}
	|	TOKEN_TRAP get_trap
		{
			if (!add_trap(&object.trap_list, $2)) {
				YYABORT;
			}
			free($2);
			if (config.enable_traps != -1)
				config.enable_traps++;
		}
	|	TOKEN_SERVICE legal_string '{' subobject_config '}'
		{
			subobject.name = $2;
			if (!add_object(&object.service, &subobject)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE quoted_string
		{
			interface.name = $2;
			if (!add_interface(&object.interface, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE quoted_string '{' interface_config '}'
		{
			interface.name = $2;
			if (!add_interface(&object.interface, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > 65535) {
				yyerror("interface index out of range");
				YYABORT;
			}
			interface.index = $2;
			if (!add_interface(&object.interface, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE TOKEN_NUMBER '{' interface_config '}'
		{
			if ($2 < 1 || $2 > 65535) {
				yyerror("interface index out of range");
				YYABORT;
			}
			interface.index = $2;
			if (!add_interface(&object.interface, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE TOKEN_GROUP legal_string
		{
			if (!find_group_list(config.group_list, $3)) {
				yyerror("group name \"%s\" undefined", $3);
				YYABORT;
			}
			interface.name = $3;
			if (!add_interface(&object.ifgroup, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_INTERFACE TOKEN_GROUP legal_string '{' interface_config '}'
		{
			if (!find_group_list(config.group_list, $3)) {
				yyerror("group name \"%s\" undefined", $3);
				YYABORT;
			}
			interface.name = $3;
			if (!add_interface(&object.ifgroup, &interface)) {
				YYABORT;
			}
		}
	|	TOKEN_BGP TOKEN_NUMBER
		{
			if ($2 < 1 || $2 > 65535) {
				yyerror("bgp ASN out of range");
				YYABORT;
			}
			bgp_as.asn = $2;
			if (!add_bgp_as(&object.bgp, &bgp_as)) {
				YYABORT;
			}
		}
	|	TOKEN_BGP TOKEN_NUMBER '{' bgp_config '}'
		{
			if ($2 < 1 || $2 > 65535) {
				yyerror("bgp ASN out of range");
				YYABORT;
			}
			bgp_as.asn = $2;
			if (!add_bgp_as(&object.bgp, &bgp_as)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_TEMP
		{
			env_mon.type = TYPE_ENVTEMP;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_TEMP '{' env_config '}'
		{
			env_mon.type = TYPE_ENVTEMP;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_VOLT
		{
			env_mon.type = TYPE_ENVVOLT;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_VOLT '{' env_config '}'
		{
			env_mon.type = TYPE_ENVVOLT;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_FAN
		{
			env_mon.type = TYPE_ENVFAN;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_FAN '{' env_config '}'
		{
			env_mon.type = TYPE_ENVFAN;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_PS
		{
			env_mon.type = TYPE_ENVPS;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	TOKEN_ENV TOKEN_PS '{' env_config '}'
		{
			env_mon.type = TYPE_ENVPS;
			if (!add_env_mon(&object.env, &env_mon)) {
				YYABORT;
			}
		}
	|	init_variable
		{
			if (!add_variable(&object.var_list, $1)) {
				YYABORT;
			}
		}
	;

subobject_config:	/* empty */
	|	subobject_config subobject_statement
	;

subobject_statement:
		TOKEN_DESCRIPTION quoted_string
		{
			if (subobject.descr) {
				yyerror("service description duplicated");
				YYABORT;
			}
			subobject.descr = $2;
		}
	|	TOKEN_DATADIR legal_string
		{
			if (subobject.datadir) {
				yyerror("service datadir duplicated");
				YYABORT;
			}
			subobject.datadir = $2;
		}
	|	TOKEN_NETSTATE TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &subobject.ns_acl, $3)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&subobject.save_list, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE get_save
		{
			if (!add_save(&subobject.save_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_METHOD get_method '{' method_config '}'
		{
			if (method.address) {
				if ($2->address) free($2->address);
				$2->address = method.address;
				$2->ip_addr = method.ip_addr;
			}
			if (method.rport > 0)
				$2->rport = method.rport;
			if (method.timeout > 0)
				$2->timeout = method.timeout;
			if (method.retries > 0)
				$2->retries = method.retries;
			if (method.lport_min > 0)
				$2->lport_min = method.lport_min;
			if (method.lport_max > 0)
				$2->lport_max = method.lport_max;
			if (method.snmpreq) {
				if ($2->snmpreq) {
					free_var_ref($2->snmpreq->var_ref);
					free($2->snmpreq);
				}
				$2->snmpreq = method.snmpreq;
			}
			CLEAR_METHOD(&method);

			if (!add_method(&subobject.method_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_METHOD get_method
		{
			if (!add_method(&subobject.method_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_TRAP legal_string '{' trap_config '}'
		{
			trap.name = $2;
			trap.type = 6;	/* Enterprise Specific trap */
			if (!add_trap(&subobject.trap_list, &trap)) {
				YYABORT;
			}
			if (config.enable_traps != -1)
				config.enable_traps++;
		}
	|	TOKEN_TRAP get_trap
		{
			if (!add_trap(&subobject.trap_list, $2)) {
				YYABORT;
			}
			free($2);
			if (config.enable_traps != -1)
				config.enable_traps++;
		}
	|	init_variable
		{
			if (!add_variable(&subobject.var_list, $1)) {
				YYABORT;
			}
		}
	;

interface_config:	/* empty */
	|	interface_config interface_statement
	;

interface_statement:
		TOKEN_DESCRIPTION quoted_string
		{
			if (interface.descr) {
				yyerror("interface description duplicated");
				YYABORT;
			}
			interface.descr = $2;
		}
	|	TOKEN_DATADIR legal_string
		{
			if (interface.datadir) {
				yyerror("interface datadir duplicated");
				YYABORT;
			}
			interface.datadir = $2;
		}
	|	TOKEN_NETSTATE TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &interface.ns_acl, $3)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&interface.save_list, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE get_save
		{
			if (!add_save(&interface.save_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_BANDWIDTH TOKEN_NUMBER
		{
			if (interface.bandwidth) {
				yyerror("interface bandwidth duplicated");
				YYABORT;
			}
			interface.bandwidth = $2 * 1000; /* make bps */
		}
	|	init_variable
		{
			if (!add_variable(&interface.var_list, $1)) {
				YYABORT;
			}
		}
	;

bgp_config:	/* empty */
	|	bgp_config bgp_statement
	;

bgp_statement:
		TOKEN_DESCRIPTION quoted_string
		{
			if (bgp_as.descr) {
				yyerror("bgp description duplicated");
				YYABORT;
			}
			bgp_as.descr = $2;
		}
	|	TOKEN_DATADIR legal_string
		{
			if (bgp_as.datadir) {
				yyerror("bgp datadir duplicated");
				YYABORT;
			}
			bgp_as.datadir = $2;
		}
	|	TOKEN_NETSTATE TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &bgp_as.ns_acl, $3)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&bgp_as.save_list, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE get_save
		{
			if (!add_save(&bgp_as.save_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_PEER quoted_string
		{
			if (!add_bgp_peer(&bgp_as, $2)) {
				YYABORT;
			}
			free($2);
		}
	;

env_config:	/* empty */
	|	env_config env_statement
	;

env_statement:
		TOKEN_DESCRIPTION quoted_string
		{
			if (env_mon.descr) {
				yyerror("env description duplicated");
				YYABORT;
			}
			env_mon.descr = $2;
		}
	|	TOKEN_DATADIR legal_string
		{
			if (env_mon.datadir) {
				yyerror("env datadir duplicated");
				YYABORT;
			}
			env_mon.datadir = $2;
		}
	|	TOKEN_NETSTATE TOKEN_GROUP legal_string
		{
			if (!add_group_ref(config.group_list, &env_mon.ns_acl, $3)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE legal_string '{' save_config '}'
		{
			save.name = $2;
			if (!add_save(&env_mon.save_list, &save)) {
				YYABORT;
			}
		}
	|	TOKEN_SAVE get_save
		{
			if (!add_save(&env_mon.save_list, $2)) {
				YYABORT;
			}
			free($2);
		}
	|	TOKEN_GAUGE quoted_string
		{
			if (!add_env_gauge(&env_mon, $2)) {
				YYABORT;
			}
		}
	;

get_save:	legal_string optional_string
		{
			SAVE *save;
			if ((save = find_save(first_save, $1)) == NULL) {
				yyerror("save name \"%s\" undefined", $1);
				YYABORT;
			}
			free($1);
			if (($$ = dup_save_args(save, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	|	METHOD_ROUTER optional_string
		{
			extern SAVE save_router;
			if (($$ = dup_save_args(&save_router, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	|	TOKEN_INTERFACE optional_string
		{
			extern SAVE save_interface;
			if (($$ = dup_save_args(&save_interface, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	|	TOKEN_BGP optional_string
		{
			extern SAVE save_bgppeer;
			if (($$ = dup_save_args(&save_bgppeer, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	|	TOKEN_TEMP optional_string
		{
			extern SAVE save_envtemp;
			if (($$ = dup_save_args(&save_envtemp, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	|	TOKEN_VOLT optional_string
		{
			extern SAVE save_envvolt;
			if (($$ = dup_save_args(&save_envvolt, $2)) == NULL) {
				YYABORT;
			}
			if ($2) free($2);
		}
	;

get_method:	legal_string optional_string
		{
			METHOD *method;
			if ((method = find_method(first_method, $1)) == NULL) {
				yyerror("method name \"%s\" undefined", $1);
				YYABORT;
			}
			free($1);
			if (($$ = dup_method_args(method, $2)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_PING
		{
			extern METHOD method_ping;
			if (($$ = dup_method_args(&method_ping, 0)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_SNMP optional_string
		{
			extern METHOD method_snmp;
			if (($$ = dup_method_args(&method_snmp, $2)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_SNMP TOKEN_V2 optional_string
		{
			extern METHOD method_snmp;
			if (($$ = dup_method_args(&method_snmp, $3)) == NULL) {
				YYABORT;
			}
			$$->snmp_version = SNMP_VERSION_2;
		}
	|	METHOD_ROUTER optional_string
		{
			extern METHOD method_router;
			if (($$ = dup_method_args(&method_router, $2)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_ROUTER TOKEN_V2 optional_string
		{
			extern METHOD method_router;
			if (($$ = dup_method_args(&method_router, $3)) == NULL) {
				YYABORT;
			}
			$$->snmp_version = SNMP_VERSION_2;
		}
	|	METHOD_DNS optional_string
		{
			extern METHOD method_dns;
			if (($$ = dup_method_args(&method_dns, $2)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_RADIUS optional_string
		{
			extern METHOD method_radius;
			if (($$ = dup_method_args(&method_radius, $2)) == NULL) {
				YYABORT;
			}
		}
	|	METHOD_TACACS optional_string
		{
			extern METHOD method_tacacs;
			if (($$ = dup_method_args(&method_tacacs, $2)) == NULL) {
				YYABORT;
			}
		}
	;

get_trap:	legal_string optional_string
		{
			TRAP *trap;
			if ((trap = find_trap(first_trap, $1)) == NULL) {
				yyerror("trap name \"%s\" undefined", $1);
				YYABORT;
			}
			free($1);
			if (($$ = dup_trap(trap, $2)) == NULL) {
				YYABORT;
			}
		}
	|	TRAP_GENERIC optional_string
		{
			extern TRAP trap_generic;
			if (($$ = dup_trap(&trap_generic, $2)) == NULL) {
				YYABORT;
			}
		}
	;

init_variable:	make_variable
		{
			$$ = $1;
		}
	|	make_variable '=' TOKEN_NUMBER
		{
			int num = $3;
			set_var($1, INTEGER, &num, sizeof(num));
			$$ = $1;
		}
	|	make_variable '=' '-' TOKEN_NUMBER
		{
			int num = -$4;
			set_var($1, INTEGER, &num, sizeof(num));
			$$ = $1;
		}
	|	make_variable '=' TOKEN_STRING
		{
			set_var($1, STRING, $3, strlen($3));
			$$ = $1;
		}
	|	make_variable '=' VAR_OID
		{
			oid buf[100];
			int len = 100;
			str2oid(buf, &len, $3);
			if (len < 4) {
				yyerror("bad variable OID");
				YYABORT;
			}
			set_var($1, OBJID, buf, len * sizeof(oid));
			$$ = $1;
		}
	;

make_variable:	VAR_NAME
		{
			const SNMP_MIB_ENTRY *mib = get_mib_entry($1, 0);
			$$ = (VARIABLE *)malloc(sizeof(VARIABLE));
			if (!$$) {
				yyerror("Out of memory");
				YYABORT;
			}
			memset($$, 0, sizeof(VARIABLE));
			$$->name = strdup($1);
			if (!$$->name) {
				yyerror("Out of memory");
				YYABORT;
			}
			if (mib && mib->objid)
				set_objid($$, mib->objid, mib->objid_len);
		}
	|	make_variable VAR_OID
		{
			oid buf[100];
			int len = 100;
			str2oid(buf, &len, $2);
			if (len < 4) {
				yyerror("bad variable OID");
				YYABORT;
			}
			set_objid($1, buf, len);
		}
	;

quoted_string:	TOKEN_STRING
		{
			$$ = strdup($1);
			if (!$$) {
				yyerror("Out of memory");
				YYABORT;
			}
#ifdef	DEBUG_MEMORY
			assert(get_cksum());
#endif
		}
	;

legal_string:	TOKEN_STRING
		{
			char *cp;
			if (!$1 || !*$1) {
				yyerror("required string is empty");
				YYABORT;
			}
			if ((cp = strpbrk($1, ILLEGAL_CHARS)) != NULL) {
				yyerror("string \"%s\" contain '%c'", $1, *cp);
				YYABORT;
			}
			$$ = strdup($1);
			if (!$$) {
				yyerror("Out of memory");
				YYABORT;
			}
#ifdef	DEBUG_MEMORY
			assert(get_cksum());
#endif
		}
	;

multiline_string:	TOKEN_STRING
		{
			$$ = strdup($1);
			if (!$$) {
				yyerror("Out of memory");
				YYABORT;
			}
#ifdef	DEBUG_MEMORY
			assert(get_cksum());
#endif
		}
	|	multiline_string TOKEN_STRING
		{
			$$ = (char *)malloc(strlen($1) + strlen($2) + 1);
			if (!$$) {
				yyerror("Out of memory");
				YYABORT;
			}
			(void)strcpy($$, $1);
			(void)strcat($$, $2);
			free($1);
#ifdef	DEBUG_MEMORY
			assert(get_cksum());
#endif
		}
	;

optional_number:
		{
			$$ = -1;
		}
	|	TOKEN_NUMBER
	;

optional_string:
		{
			$$ = NULL;
		}
	|	quoted_string
	;

%%

void *
parse_config(arg)
	void *arg;
{
	int reconfig = (int)arg;
	int error;
	extern FILE *yyin;
	extern int yylineno;

	if (!config_file) return NULL;

	dprintf(("parse_config: entered\n"));

	if ((yyin = fopen(config_file, "r")) == NULL) {
		report(LOG_ERR, "parse_config: %s: %m", config_file);
		return NULL;
	}
	yylineno = 1;
	clear_config();
	error = yyparse();
	fclose(yyin);

	return (error ? NULL : set_config());
}
