/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "md5.h"

extern int errno;

/*
 * RADIUS specification according to RFC2138.
 */

#define	RADIUSSERVER_PORT	1645	/* 1812 suggested */
#define	HEADER_LEN		20
#define	MIN_PACKETSZ		HEADER_LEN
#define	MAX_PACKETSZ		4096
#define	AUTH_LEN		16
#define	AUTH_PASS_LEN		64
#define	TEST_NAME		NETMON

typedef	struct radius_header {
	u_char	code;		/* type of RADIUS packet */
/* standart RADIUS codes */
#define	ACCESS_REQUEST	1
#define	ACCESS_ACCEPT	2
#define	ACCESS_REJECT	3
	u_char	id;		/* request identification number */
	u_short	length;		/* length of packet */
	u_char	auth[AUTH_LEN];	/* authenticator */
	u_char	data[1];	/* attributes... */
/* standart RADIUS attribute-value pairs */
#define	ATTR_USER_NAME		1	/* string */
#define	ATTR_USER_PASS		2	/* string */
#define	ATTR_SERVICE_TYPE	6	/* integer */
#define		SERVICE_AUTHENTICATE_ONLY	8
#define	ATTR_REPLY_MESSAGE	18	/* string */
#define	ATTR_NAS_IDENTIFIER	32	/* string */
} HEADER;


static void radius_reply __P((int, SESSION *, int));

static int
radius_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	u_char *cp, buf[MAX_PACKETSZ];
	HEADER *hp = (HEADER *)buf;
	int len, length = HEADER_LEN;
	u_int32_t val;
	char *secret;

	/* build the RADIUS protocol packet header */
	memset(hp, 0, HEADER_LEN);
	hp->code = ACCESS_REQUEST;
	hp->id = (u_char)request->reqid;
	hp->length = htons(length);
	len = sizeof(TIMEVAL);
	if (len > AUTH_LEN) len = AUTH_LEN;
	memcpy(hp->auth, &sd->owner->last_request, len);

	/* pack an attribute-value pairs */
	cp = hp->data;
	/* User-Name */
	*cp++ = ATTR_USER_NAME;
	len = sizeof(TEST_NAME)-1;
	*cp++ = len + 2;
	memcpy(cp, TEST_NAME, len);
	cp += len;
	length += len + 2;

	secret = sd->method->argument;
	if (secret && *secret) {
		u_char passbuf[AUTH_PASS_LEN], digest[AUTH_PASS_LEN + AUTH_LEN];

		/* User-Password */
		*cp++ = ATTR_USER_PASS;
		*cp++ = AUTH_PASS_LEN + 2;

		len = strlen(secret);	/* secret as password */
		if (len > AUTH_PASS_LEN) len = AUTH_PASS_LEN;
		memset(passbuf, 0, AUTH_PASS_LEN);
		memcpy(passbuf, secret, len);

		/* calculate md5 digest */
		len = strlen(secret);	/* secret */
		if (len > AUTH_PASS_LEN) len = AUTH_PASS_LEN;
		memcpy(digest, secret, len);
		memcpy(digest + len, hp->auth, AUTH_LEN);
		md5_calc(cp, digest, len + AUTH_LEN);

		/* XOR the password into md5 digest */
		for (len = 0; len < AUTH_PASS_LEN; len++)
			*cp++ ^= passbuf[len];
		length += AUTH_PASS_LEN + 2;
	}

	/* Service Type */
	*cp++ = ATTR_SERVICE_TYPE;
	*cp++ = sizeof(u_int32_t) + 2;
	val = htonl(SERVICE_AUTHENTICATE_ONLY);
	memcpy(cp, &val, sizeof(u_int32_t));
	cp += sizeof(u_int32_t);
	length += sizeof(u_int32_t) + 2;

	/* Requestor Identifier */
	*cp++ = ATTR_NAS_IDENTIFIER;
	len = sizeof(NETMON)-1;
	*cp++ = len + 2;
	memcpy(cp, NETMON, len);
	cp += len;
	length += len + 2;

	hp->length = htons(length);

	len = send(sd->sock, (char *)buf, length, 0);
	if (len < 0 && errno == EINPROGRESS)
		len = 0;
	return len;
}

static int
radius_recv(sd)
	SESSION *sd;
{
	REQUEST *request;
	u_char buf[MAX_PACKETSZ];
	HEADER *hp;
	struct sockaddr_in from;
	int len;
	socklen_t length;

	if ((request = get_request(sd, 0 /* no reqid yet */)) == NULL) {
		trace(("\t radius_recv: internal error?\n"));
		return -1;
	}

	length = sizeof(from);
	if ((len = recvfrom(sd->sock, (char *)buf, sizeof(buf), 0,
			    (struct sockaddr *)&from, &length)) < 0) {
		if (errno != EAGAIN)
			radius_reply(errno, sd, request->reqid);
		return -1;
	}

	if (len < MIN_PACKETSZ) {
		trace(("\t radius_recv: undersized packet, len=%d\n", len));
		return -1;
	}
	hp = (HEADER *)buf;
	length = ntohs(hp->length);
	if (length < MIN_PACKETSZ || length > MAX_PACKETSZ) {
		trace(("\t radius_recv: bad packet length, len=%d\n", length));
		return -1;
	}	
	if (hp->id != (u_char)request->reqid) {
		trace(("\t radius_recv: invalid reqid\n"));
		return -1;
	}
	if (hp->code != ACCESS_ACCEPT && hp->code != ACCESS_REJECT) {
		trace(("\t radius_recv: not a response\n"));
		return -1;
	}

	if (sd->data_ptr) free(sd->data_ptr);
	if ((sd->data_ptr = (char *)malloc(length)) != NULL) {
		memmove(sd->data_ptr, buf, length);
		sd->data_int = length;
	} else 	sd->data_int = 0;

	return 0;
}

static void
radius_stop(method)
	METHOD *method;
{
	session_stop(method->sd, 1);
}

static void
radius_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int reqid;

	/* sanity check */
	if (!sd) return;

	radius_stop(method); /* stop the method is still running */

	/* open datagram socket near before UDP session */
	if ((sd->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		radius_reply(errno, sd, 0);
		return;
	}

	/* turn on non-blocking I/O */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		radius_reply(errno, sd, 0);
		return;
	}

	/*
	 * Make datagram socket `connected', so error on this socket
	 * will be returned asynchronously without timing out.
	 */
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		radius_reply(errno, sd, 0);
		return;
	}

	if ((reqid = session_start(sd, NULL)) != 0) {
		dprintf(("radius_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, reqid));
	}
}

static char *
radius_parse(buf, packet, length)
	char *buf;
	u_char *packet;
	int length;
{
	HEADER *hp = (HEADER *)packet;
	u_char *cp;
	char *tp;
	int len;

	/* extract attributes if any */
	*buf = '\0';
	tp = buf;
	cp = hp->data;
	length -= HEADER_LEN;
	for (len = length; length > 0; cp += len, length -= len) {
		len = cp[1];
		if (*cp == ATTR_REPLY_MESSAGE) {
			if (*buf) {
				memcpy(tp, "; ", 2);
				tp += 2;
			}
			memcpy(tp, &cp[2], len - 2);
			tp += len - 1;
			if (*tp) tp++;
		}
	}
	*tp = '\0';

	return buf;
}

static void
radius_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[MAX_PACKETSZ];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		diag = radius_parse(buf, sd->data_ptr, sd->data_int);
		if (!diag) {
			op = -1;
			diag = "Unknown error";
		} else if (!*diag) {
			diag = "No error";
		}
	} else	diag = strerror(op);

	dprintf(("radius_reply(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	radius_stop(method);

	method_finished(target, method, diag, !op);
}

static int
radius_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("radius_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = radius_send;
	template.recv = radius_recv;
	template.read = radius_reply;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_radius = {
	radius_init,	/* initialize radius protocol session */
	radius_start,	/* start radius session */
	radius_stop,	/* stop radius session */

	"RADIUS",	/* method name */
	NULL,		/* no argument by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_UDP,	/* network protocol */
	RADIUSSERVER_PORT, /* server port */
	0, 0,		/* timeout and retries undefined yet */
	{ 0, 0 },	/* no parameters used */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
