/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"

extern int errno;

#ifdef	O_NONBLOCK
#define	ASYNC_MODE	O_NONBLOCK
#elif	O_NDELAY
#define	ASYNC_MODE	O_NDELAY
#elif	FNDELAY
#define	ASYNC_MODE	FNDELAY
#elif	O_ASYNC
#define	ASYNC_MODE	O_ASYNC
#elif
#error the fcntl argument to turn ON/OFF non-blocking I/O is unknown
#endif

static int ReqId = 0;
static SESSION *first_session = NULL;	/* first network session in table */

SESSION *
session_create(template)
	SESSION *template;
{
	SESSION *curr_session;

	if ((curr_session = (SESSION *)malloc(sizeof(SESSION))) == NULL)
		return NULL;
	memset(curr_session, 0, sizeof(SESSION));

	/* user supplyed data */
	curr_session->owner = template->owner;
	curr_session->method = template->method;
	curr_session->sock = template->sock;
	curr_session->peer = template->peer;
	curr_session->timeout = template->timeout;
	curr_session->retries = template->retries;
	curr_session->connect = template->connect;
	curr_session->send = template->send;
	curr_session->recv = template->recv;
	curr_session->read = template->read;

	/* make chain of sessions */
	if (first_session) {
		SESSION *prev = first_session;
		while (prev->next) prev = prev->next;
		prev->next = curr_session;
	} else	first_session = curr_session;

	return curr_session;
}

/*
 * Return time offset for the session.
 */
int
session_offset(session, period)
	SESSION *session;
	int period;
{
	SESSION *sd;
	int off, curr, all;

	/* find current session index and number of all session */
	off = curr = all = 0;
	for (sd = first_session; sd; sd = sd->next) {
		if (!off) {
			if (sd != session) curr++;
			else off = 1;
		}
		all++;
	}
	off = period / all;
	if (off < 1) off = 1;
	return (off * curr);
}

/*
 * Check to see is there a busy session? Null arg match any first session.
 */
SESSION *
session_busy(session)
	SESSION *session;
{
	SESSION *sd;
	REQUEST *sr;

	for (sd = first_session; sd; sd = sd->next) {
		if (session && sd != session)
			continue;
		for (sr = sd->request; sr; sr = sr->next) {
			if (sr->reqid && sr->sequence)
				return sd;
		}
	}
	return NULL;
}

static void
free_requests(sd)
	SESSION *sd;
{
	REQUEST *curr, *next;

	curr = sd->request;
	while (curr) {
		next = curr->next;
		if (curr->snmpreq) {
			free_var_ref(curr->snmpreq->var_ref);
			free(curr->snmpreq);
			curr->snmpreq = NULL;
		}
		free(curr);
		curr = next;
	}
	sd->request = NULL;
}

void
session_free(free_sd)
	SESSION *free_sd;
{
	SESSION *sd, *prev, *next;

	if (free_sd == (SESSION *)0 || free_sd == (SESSION *)-1)
		return;

	sd = first_session;
	prev = NULL;
	while (sd) {
		if (sd == free_sd) {
			session_stop(sd, 1);
			free_requests(sd);
			next = sd->next;
			if (prev)
				prev->next = next;
			else	first_session = next;
			free(sd);
			sd = next;
		} else {
			prev = sd;
			sd = sd->next;
		}
	}
}

static void
set_request_timeout(request, timeout)
	register REQUEST *request;
	long timeout;
{
	if (request) {
		gettimeofday(&request->expire, NULL);
		request->expire.tv_usec += timeout;
		request->expire.tv_sec += request->expire.tv_usec / 1000000L;
		request->expire.tv_usec %= 1000000L;
	}
}

/*
 * Start a session. Handle multiply requests per session.
 */ 
int
session_start(sd, snmpreq)
	SESSION *sd;
	SNMP_REQ *snmpreq;
{
	REQUEST *curr, *prev = NULL, *next = NULL;

	if (sd->sock < 0) {
		if (sd->read) (*sd->read)(ENOTCONN, sd, 0);
		report(LOG_ERR, "session_start: Socket is not opened");
		return 0;
	}

	if (++ReqId < 1) ReqId = 1;

	/* search for first empty or last request slot */
	for (curr = sd->request; curr; curr = curr->next) {
		if (curr->reqid == 0) {
			next = curr->next;
			break;
		}
		prev = curr;
	}
	if (curr) {
		if (curr->snmpreq) {
			free_var_ref(curr->snmpreq->var_ref);
			free(curr->snmpreq);
			curr->snmpreq = NULL;
		}
	} else if ((curr = (REQUEST *)malloc(sizeof(REQUEST))) == NULL) {
		if (sd->read) (*sd->read)(errno, sd, ReqId);
		report(LOG_ERR, "session_start: Out of memory");
		return 0;
	}

	memset(curr, 0, sizeof(REQUEST));
	curr->snmpreq = snmpreq;
	curr->reqid = ReqId;
	curr->sequence = 1;
	if (next) curr->next = next;
	else if (prev) prev->next = curr;
	if (!sd->request) sd->request = curr;

	if (!sd->connect && sd->send) {
		if ((*sd->send)(sd, curr) < 0) {
			curr->reqid = 0;
			if (sd->read) (*sd->read)(errno, sd, ReqId);
			return 0;
		}
		sd->pkt_send++;	/* packet counter */
	}
	set_request_timeout(curr, sd->timeout);
	return ReqId;
}

void
session_stop(sd, close_socket)
	SESSION *sd;
	int close_socket;
{
	REQUEST *sr;

	/* sanity check */
	if (!sd) return;

	if (close_socket) {
		/* close the socket */
		if (sd->sock != -1) {
			close(sd->sock);
			sd->sock = -1;
		}
	}

	/* clear temporary buffer */
	if (sd->buf) {
		free(sd->buf);
		sd->buf = NULL;
	}

	/* clear last data */
	sd->data_int = 0;
	if (sd->data_ptr) {
		free(sd->data_ptr);
		sd->data_ptr = NULL;
	}

	/* reset statistics */
	memset(&sd->tv_sum, 0, sizeof(TIMEVAL));
	sd->pkt_send = sd->pkt_recv = 0;

	/* clear session requests */
	for (sr = sd->request; sr; sr = sr->next) {
		if (sr->snmpreq) {
			free_var_ref(sr->snmpreq->var_ref);
			free(sr->snmpreq);
			sr->snmpreq = NULL;
		}
		sr->reqid = 0;
		sr->sequence = 0;
		memset(&sr->expire, 0, sizeof(TIMEVAL));
	}
}

int
session_select(nfds, readfds, writefds, timeout, block)
	int *nfds;
	fd_set *readfds, *writefds;
	TIMEVAL *timeout;
	int *block;
{
	SESSION *sd;
	REQUEST *sr;
	TIMEVAL earliest, now;
	int active = 0, pending = 0;

	timerclear(&earliest);

	/*
	 * For each request outstanding, add it's socket to the readfds,
	 * and if it is the earliest timeout to expire, mark it as lowest.
	 */
	for (sd = first_session; sd; sd = sd->next) {
		if (sd->sock < 0)
			continue;

		active++;
		if (sd->sock + 1 > *nfds)
			*nfds = sd->sock + 1;

		if (!sd->connect)
			FD_SET(sd->sock, readfds);
		else	FD_SET(sd->sock, writefds);

		for (sr = sd->request; sr; sr = sr->next) {
			if (timerisset(&sr->expire)) {
				pending++;
				if (!timerisset(&earliest) ||
				    timercmp(&sr->expire, &earliest, <))
					earliest = sr->expire;
			}
		}
	}
	if (!pending)
		return active;

	/*
	 * Transforms earliest from an absolute time into a delta time, the
	 * time left until the select should timeout.
	 */
	gettimeofday(&now, NULL);
	tv_sub(&earliest, &now);

	/* if it was blocking before or our delta time is less, reset timeout */
	if (*block || timercmp(&earliest, timeout, <)) {
		*timeout = earliest;
		*block = 0;
	}
	return active;
}

REQUEST *
get_request(sd, reqid)
	SESSION *sd;
	int reqid;
{
	REQUEST *sr;

	if (reqid == 0) /* for single request per session (like tcp or icmp) */
		return sd->request;

	if (reqid < 0) /* invalid reqid */
		return NULL;

	for (sr = sd->request; sr; sr = sr->next) {
		if (sr->reqid == reqid)
			break;
	}
	return sr;
}

/*
 * Checks to see if any of the fd's set in the readfds belong to a session.
 */
void
session_operate(readfds, writefds)
	fd_set *readfds, *writefds;
{
	SESSION *sd;
	int error, reqid;
	REQUEST *sr;

	for (sd = first_session; sd; sd = sd->next) {
		if (sd->sock < 0)
			continue;

		if (!sd->connect && FD_ISSET(sd->sock, readfds)) {
			trace(("\t has to read socket %d\n", sd->sock));

			error = EINVAL;
			reqid = 0;
			if (sd->recv && sd->request) {
				reqid = sd->request->reqid;
				error = get_socket_error(sd->sock);
				if (!error && (reqid = (*sd->recv)(sd)) != -1) {
					sd->pkt_recv++;
					if ((sr = get_request(sd, reqid)) != NULL) {
						reqid = sr->reqid;
						sr->reqid = 0;
						sr->sequence = 0; /* sequence done */
						memset(&sr->expire, 0, sizeof(TIMEVAL));
					}
				}
			}
			if (reqid != -1 && sd->read)
				(*sd->read)(error, sd, reqid);
		}
		if (sd->connect && FD_ISSET(sd->sock, writefds)) {
			trace(("\t has to write socket %d\n", sd->sock));

			error = EINVAL;
			reqid = 0;
			if (sd->request) {
				reqid = sd->request->reqid;
				error = get_socket_error(sd->sock);
				if (!error) {
					if ((*sd->connect)(sd) != -1) {
						/* request id remain the same */
						if (sd->send &&
						    (*sd->send)(sd, sd->request) != -1) {
							sd->pkt_send++;
							set_request_timeout(sd->request, sd->timeout);
						} else {
							error = errno;
						}
					}
				}
			}
			if (error && sd->read) (*sd->read)(error, sd, reqid);
		}
	}
}

/*
 * Checks to see if any of the sessions have an outstanding request that
 * has timed out.  If it finds one (or more), and more retries available,
 * request will resent. If there are no more retries available, the
 * timeout function for the session is used to alert the user of the timeout.
 */
void
session_timeout()
{
	TIMEVAL now;
	SESSION *sd;
	REQUEST *sr;
	int reqid;

	gettimeofday(&now, NULL);

	/*
	 * For each request outstanding, check to see if it has expired.
	 */
	for (sd = first_session; sd; sd = sd->next) {
		if (sd->sock < 0)
			continue;

		for (sr = sd->request; sr; sr = sr->next) {
			if (timerisset(&sr->expire) &&
			    timercmp(&sr->expire, &now, <)) {
				/* this timer has expired */
				if (sr->sequence >= sd->retries) {
					/* no more retries */
					reqid = sr->reqid;
					sr->reqid = 0;
					sr->sequence = 0;
					memset(&sr->expire, 0, sizeof(TIMEVAL));
					if (sd->read) (*sd->read)(ETIMEDOUT, sd, reqid);
				} else { /* retransmit */
					sr->sequence++;
					if (sd->send &&
					    (*sd->send)(sd, sr) != -1) {
						sd->pkt_send++;
						set_request_timeout(sr, sd->timeout);
					} else {
						reqid = sr->reqid;
						sr->reqid = 0;
						sr->sequence = 0;
						memset(&sr->expire, 0, sizeof(TIMEVAL));
						if (sd->read) (*sd->read)(errno, sd, reqid);
					}
				}
			}
		}
	}
}

/*
 * Return socket error (like errno) in the session or 0 if no errors.
 */
int
get_socket_error(sock)
	int sock;
{
	int argbuf;
	socklen_t arglen;
	struct sockaddr peer;

	arglen = sizeof(argbuf);
	if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char *)&argbuf, &arglen) < 0)
		return errno;
	if (argbuf)
		return argbuf;

	arglen = sizeof(argbuf);
	if (getsockopt(sock, SOL_SOCKET, SO_TYPE, (char *)&argbuf, &arglen) < 0)
		return errno;
	if (argbuf == SOCK_STREAM) {
		arglen = sizeof(peer);
		if (getpeername(sock, &peer, &arglen) < 0)
			return errno;
	}

	return 0;
}

/*
 * Make socket blocked or non-blocked for sync/async I/O.
 */
int
set_socket_async(sock, on)
	int sock;
	int on; /* boolean */
{
	int mode;
	int prev; /* boolean */

	/* get current value of I/O mode */
	if ((mode = fcntl(sock, F_GETFL, 0)) < 0)
		return -1;

	prev = (mode & ASYNC_MODE) != 0;
	if (on != prev) {
		if (on)	mode |= ASYNC_MODE;
		else	mode &= ~ASYNC_MODE;
		if (fcntl(sock, F_SETFL, mode))
			return -1;
	}
	return prev;
}
