/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "snmp.h"
#include "mib.h"

extern int errno;

char *snmp_error[NUM_ERRORS] = {
	"No error",			/* NOERROR */
	"Response message too big",	/* TOOBIG */
	"No such variable name",	/* NOSUCHNAME */
	"Wrong type or length",		/* BADVALUE */
	"Access denied (read only)",	/* READONLY */
	"General failure",		/* GENERR */
/* extended error codes */
	"No access",			/* NOACCESS */
	"Wrong type",			/* WRONG_TYPE */
	"Wrong length",			/* WRONG_LENGTH */
	"Wrong encoding",		/* WRONG_ENCODING */
	"Wrong value",			/* WRONG_VALUE */
	"No creation",			/* NO_CREATION */
	"Inconsistent value",		/* INCONSISTENT_VALUE */
	"Resource unavailable",		/* RESOURCE_UNAVAILABLE */
	"Commit failed",		/* COMMIT_FAILED */
	"Undo failed",			/* UNDO_FAILED */
	"Authorization error",		/* AUTHORIZATION_ERROR */
	"Not writeable",		/* NOT_WRITEABLE */
	"Inconsistent name",		/* INCONSISTENT_NAME */
};

static void snmp_reply __P((int, SESSION *, int));
static VARIABLE *snmp_parse __P((SESSION *, SNMP_REQ *, u_char *, int));

/*
 * Routines to manipulate ASN-specific values.
 */

/* pack array of oids to unsigned long */
u_long
oid2long(from, len)
	const oid *from;
	int len;
{
	if (from) {
		if (len == 1)
			return (u_long)*from;
		if (len > 1) {
			u_long ul = 0;
			int i = 0;
			if (len > sizeof(u_long))
				i = len - sizeof(u_long);
			while (i < len) {
				ul = ul << 8;
				ul |= from[i++] & 0xFF;
			}
			return ul;
		}
	}
	return 0l;
}

oid *
long2oid(to, len, from)
	oid *to;
	int len;
	u_long from;
{
	if (to) {
		if (len == 1) {
			*to = (oid)from;
			return to;
		}
		if (len > 1) {
			int i;
			for (i = len; i > 0; i--) {
				to[i-1] = from & 0xFF;
				from = from >> 8;
			}
			return to;
		}
	}
	return 0;
}

static u_char *
insert_length(buf, size, length)
	u_char *buf;
	int *size;
	int length;
{
	u_char *begin = buf;

	if (length < 0x80) {
		if (*size < 1) return NULL;
		*buf++ = (u_char)length;
	} else if (length <= 0xFF) {
		if (*size < 2) return NULL;
		*buf++ = (u_char)(0x01 | ASN_LONG_LEN);
		*buf++ = (u_char)length;
	} else {
		if (*size < 3) return NULL;
		*buf++ = (u_char)(0x02 | ASN_LONG_LEN);
		*buf++ = (u_char)((length >> 8) & 0xFF);
		*buf++ = (u_char)(length & 0xFF);
	}
	*size -= buf - begin;
	return buf;
}

static u_char *
insert_header(buf, size, type, length)
	u_char *buf;
	int *size;
	int type;
	int length;
{
	if (*size < 1) return NULL;
	*buf++ = (u_char)type;
	*size -= 1;
	return insert_length(buf, size, length);
}

static u_char *
insert_sequence(buf, size, type, length)
	u_char *buf;
	int *size;
	int type;
	int length;
{
	if (*size - 4 < 0) return NULL;
	*size -= 4; 
	*buf++ = (u_char)type;
	*buf++ = (u_char)(0x02 | ASN_LONG_LEN);
	*buf++ = (u_char)((length >> 8) & 0xFF);
	*buf++ = (u_char)(length & 0xFF);
	return buf;
}

static u_char *
insert_string(buf, size, type, string, string_len)
	u_char *buf;
	int *size;
	int type;
	u_char *string;
	int string_len;
{
	if ((buf = insert_header(buf, size, type, string_len)) == NULL ||
	    *size < string_len)
		return NULL;
	memmove(buf, string, string_len);
	*size -= string_len;
	buf += string_len;
	return buf;
}

static u_char *
insert_integer(buf, size, type, integer)
	u_char *buf;
	int *size;
	int type;
	long integer;
{
	u_long mask;
	int integer_len = sizeof(integer);

	/* truncate bytes off of the most significant end */
	mask = ((u_long)0x1ff) << ((8 * (sizeof(integer) - 1)) - 1);
	/* mask is 0xFF800000 on a big-endian machine */
	while (((integer & mask) == 0 || (integer & mask) == mask) &&
	       integer_len > 1) {
		integer <<= 8;
		integer_len--;
	}
	if ((buf = insert_header(buf, size, type, integer_len)) == NULL ||
	    *size < integer_len)
		return NULL;
	*size -= integer_len;
	mask = ((u_long)0xff) << (8 * (sizeof(integer) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	while (integer_len--) {
		*buf++ = (u_char)((integer & mask) >> (8 * (sizeof(integer) - 1)));
		integer <<= 8;
	}
	return buf;
}

static u_char *
insert_objid(buf, size, type, objid, objid_len)
	u_char *buf;
	int *size;
	int type;
	oid *objid;
	int objid_len;
{
	int asn_len;
	oid *op = objid;
	u_char objid_size[MAX_OID_LEN];
	u_long first_objid_val, objid_val;
	int i;

	/* if there are at least 2 sub-identifiers */
	if (objid_len < 2) {
		/* there are not, so make OID have two with value of zero */
		objid_val = 0;
		objid_len = 2;
	} else { /* combine the first two values */
		objid_val = (op[0] * 40) + op[1];
		op += 2;
	}
	first_objid_val = objid_val;

	/* calculate the number of bytes needed to store the encoded value */
	for (i = 1, asn_len = 0;;) {
		if (objid_val < (unsigned)0x80) {
			objid_size[i] = 1;
			asn_len += 1;
		} else if (objid_val < (unsigned)0x4000) {
			objid_size[i] = 2;
			asn_len += 2;
		} else if (objid_val < (unsigned)0x200000) {
			objid_size[i] = 3;
			asn_len += 3;
		} else if (objid_val < (unsigned)0x10000000) {
			objid_size[i] = 4;
			asn_len += 4;
		} else {
			objid_size[i] = 5;
			asn_len += 5;
		}
		if (++i >= objid_len) break;
		objid_val = *op++;
	}
	if ((buf = insert_header(buf, size, type, asn_len)) == NULL ||
	    *size < asn_len)
		return NULL;
	for (i = 1, objid_val = first_objid_val, op = objid + 2;
	     i < objid_len; i++) {
		if (i != 1) objid_val = *op++;
		switch (objid_size[i]) {
		case 1:
			*buf++ = (u_char)objid_val;
			break;
		case 2:
			*buf++ = (u_char)((objid_val>>7) | 0x80);
			*buf++ = (u_char)(objid_val & 0x07f);
			break;
		case 3:
			*buf++ = (u_char)((objid_val>>14) | 0x80);
			*buf++ = (u_char)((objid_val>>7 & 0x7f) | 0x80);
			*buf++ = (u_char)(objid_val & 0x07f);
			break;
		case 4:
			*buf++ = (u_char)((objid_val>>21) | 0x80);
			*buf++ = (u_char)((objid_val>>14 & 0x7f) | 0x80);
			*buf++ = (u_char)((objid_val>>7 & 0x7f) | 0x80);
			*buf++ = (u_char)(objid_val & 0x07f);
			break;
		case 5:
			*buf++ = (u_char)((objid_val>>28) | 0x80);
			*buf++ = (u_char)((objid_val>>21 & 0x7f) | 0x80);
			*buf++ = (u_char)((objid_val>>14 & 0x7f) | 0x80);
			*buf++ = (u_char)((objid_val>>7 & 0x7f) | 0x80);
			*buf++ = (u_char)(objid_val & 0x07f);
			break;
		}
	}
	*size -= asn_len;
	return buf;
}

static int
instance_empty(var)
	const VARIABLE *var;
{
	if (var && var->instance && var->instance_len > 0) {
		int i = 0;
		while (i < var->instance_len) {
			if (var->instance[i++])
				return 0;
		}
	}
	return 1;
}

static u_char *
insert_variable(buf, size, var, cmd)
	u_char *buf;
	int *size;
	VARIABLE *var;
	int cmd;
{
	u_char *begin = buf;
	int length = *size;

	if (*size - 4 < 0) return NULL;
	buf += 4;
	*size -= 4;
	if (var && var->objid && var->objid_len) {
		oid objid[MAX_OID_LEN];
		int len = 0;
		while (len < var->objid_len) {
			if (var->objid[len] == 0)
				break;
			objid[len] = var->objid[len];
			len++;
		}
		if (cmd != GETNEXT_REQUEST || !instance_empty(var)) {
			if (var->instance && var->instance_len) {
				int i = 0;
				while (i < var->instance_len)
					objid[len++] = var->instance[i++];
			}
			while (len < var->objid_len) objid[len++] = 0;
		}
		buf = insert_objid(buf, size,
				   ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID,
				   objid, len);

	} else { /* something must be send */
		buf = insert_objid(buf, size,
				   ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID,
				   sysObjID, sysObjID_len);
	}
	if (!buf) return NULL;

	if ((buf = insert_header(buf, size, ASN_NULL, 0)) == NULL)
		return NULL;
	insert_sequence(begin, &length,
			ASN_SEQUENCE | ASN_CONSTRUCTOR, (buf - begin) - 4);
	return buf;
}

static u_char *
parse_length(buf, length)
	u_char *buf;
	u_long *length;
{
	u_char len = *buf++;

	if (len & ASN_LONG_LEN) {
		len &= ~ASN_LONG_LEN;	/* turn MSb off */
		if (!len) return NULL;
		if (len > sizeof(u_long)) return NULL;
		*length = 0;	/* protect against short lengths */
		memmove(length, buf, (int)len);
		*length = ntohl(*length);
		*length >>= (8 * ((sizeof(int)) - len));
		return buf + len;
	}
	/* short asn_len */
	*length = (long)len;
	return buf;
}

static u_char *
parse_header(buf, size, type)
	u_char *buf;
	int *size;
	int *type;
{
	u_char *bp = buf;
	int asn_len, header_len;

	/* this only works on data types < 30, i.e. no extension octets */
	if (IS_EXTENSION_ID(*bp)) return NULL;
	*type = (int)(*bp & 0xff);
	if ((bp = parse_length(bp + 1, &asn_len)) == NULL)
		return NULL;
	header_len = bp - buf;
	if (header_len + asn_len > *size)
		return NULL;	/* message overflow */
	*size = (int)asn_len;
	return bp;
}

static u_char *
parse_integer(buf, size, type, arg, arg_size)
	u_char *buf;
	int *size;
	int *type;
	void *arg;
	int arg_size;
{
	u_char *bp = buf;
	u_long asn_len;
	long value = 0;

	if (arg_size != sizeof(value)) return NULL;

	*type = (int)(*bp++ & 0xff);
	if ((bp = parse_length(bp, &asn_len)) == NULL)
		return NULL;
	if ((int)asn_len > arg_size)
		return NULL;	/* asn length too big */
	if (asn_len + (bp - buf) > *size)
		return NULL;	/* message overflow */
	*size -= (int)asn_len + (bp - buf);
	if (*bp & 0x80)
		value = -1; /* integer is negative */
	while (asn_len--) value = (value << 8) | *bp++;
	memmove(arg, &value, arg_size);
	return bp;
}

static u_char *
parse_unsigned(buf, size, type, arg, arg_size)
	u_char *buf;
	int *size;
	int *type;
	void *arg;
	int arg_size;
{
	u_char *bp = buf;
	u_long asn_len;
	u_int32_t value = 0;

	if (arg_size != sizeof(value)) return NULL;

	*type = (int)(*bp++ & 0xff);
	if ((bp = parse_length(bp, &asn_len)) == NULL)
		return NULL;
	if (((int)asn_len > arg_size + 1) ||
	    (((int)asn_len == arg_size + 1) && *bp != 0))
		return NULL;	/* asn length too big */
	if (asn_len + (bp - buf) > *size)
		return NULL;	/* message overflow */
	*size -= (int)asn_len + (bp - buf);
	if (*bp & 0x80)
		value = ~value; /* integer is negative */
	while (asn_len--) value = (value << 8) | *bp++;
	memmove(arg, &value, arg_size);
	return bp;
}

static u_char *
parse_unsigned64(buf, size, type, arg, arg_size)
	u_char *buf;
	int *size;
	int *type;
	void *arg;
	int arg_size;
{
	u_char *bp = buf;
	u_long asn_len;
	u_int64_t value = 0;

	if (arg_size != sizeof(value)) return NULL;

	*type = (int)(*bp++ & 0xff);
	if ((bp = parse_length(bp, &asn_len)) == NULL)
		return NULL;
	if (((int)asn_len > arg_size + 1) ||
	    (((int)asn_len == arg_size + 1) && *bp != 0))
		return NULL;	/* asn length too big */
	if (asn_len + (bp - buf) > *size)
		return NULL;	/* message overflow */
	*size -= (int)asn_len + (bp - buf);
	if (*bp & 0x80)
		value = ~value; /* integer is negative */
	while (asn_len--) value = (value << 8) | *bp++;
	memmove(arg, &value, arg_size);
	return bp;
}

static u_char *
parse_string(buf, size, type, arg, arg_size)
	u_char *buf;
	int *size;
	int *type;
	void *arg;
	int *arg_size;
{
	u_char *bp = buf;
	u_long asn_len;

	*type = (int)(*bp++ & 0xff);
	if ((bp = parse_length(bp, &asn_len)) == NULL)
		return NULL;
	if (asn_len + (bp - buf) > *size)
		return NULL;	/* message overflow */
	*size -= (int)asn_len + (bp - buf);
	buf = bp + asn_len;
	if (asn_len > *arg_size) asn_len = (u_long)*arg_size;
	memmove(arg, bp, (int)asn_len);
	*arg_size = (int)asn_len;
	return buf;
}

static u_char *
parse_objid(buf, size, type, objid, objid_len)
	u_char *buf;
	int *size;
	int *type;
	oid *objid;
	int *objid_len;
{
	u_char *bp = buf;
	u_long asn_len, subid;
	oid *op = objid + 1;

	*type = (int)(*bp++ & 0xff);
	if ((bp = parse_length(bp, &asn_len)) == NULL)
		return NULL;
	if (asn_len + (bp - buf) > *size)
		return NULL;	/* message overflow */
	*size -= (int)asn_len + (bp - buf);

	/* handle invalid object identifier encodings of the form 06 00 robustly */
	if (!asn_len)
		objid[0] = objid[1] = 0;

	(*objid_len)--;	/* account for expansion of first byte */
	while (asn_len > 0 && (*objid_len)-- > 0) {
		subid = 0;
		do {	/* shift and add in low order 7 bits */
			subid = (subid << 7) + (*(u_char *)bp & ~ASN_BIT8);
			asn_len--;
		} while (*(u_char *)bp++ & ASN_BIT8); /* last byte has high bit clear */
		*op++ = (oid)subid;
	}

	/*
	 * The first two subidentifiers are encoded into the first component
	 * with the value (X * 40) + Y, where:
	 *	X is the value of the first subidentifier.
	 *	Y is the value of the second subidentifier.
	 */
	subid = (u_long)objid[1];
	if (subid == 0x2B) {
		objid[0] = 1;
		objid[1] = 3;
	} else if (subid < 40) {
		objid[0] = 0;
		objid[1] = subid;
	} else if (subid < 80) {
		objid[0] = 1;
		objid[1] = subid - 40;
	} else if (subid < 120) {
		objid[0] = 2;
		objid[1] = subid - 80;
	} else {
		objid[1] = subid % 40;
		objid[0] = (subid - objid[1]) / 40;
	}
	*objid_len = (int)(op - objid);
	return bp;
}

#ifdef	unused
static VARIABLE *
match_var(snmp, objid, size)
	SNMP_REQ *snmp;
	const oid *objid;
	int *size; /* IN: objid total length, OUT: rest instance length */
{
	REFERENCE *ref;
	VARIABLE *var;
	int i, rest;

	/* sanity check */
	if (!snmp || !objid || !size || *size < 4)
		return 0;

	for (ref = snmp->var_ref; ref; ref = ref->next) {
		var = ref->var;
		if (!var || !var->name || !var->objid || var->objid_len < 4)
			continue;
		rest = 1;
		for (i = 0; i < *size && i < var->objid_len; i++) {
			if (var->objid[i] == 0)
				break;
			if (var->objid[i] != objid[i]) {
				rest = 0;
				break;
			}
		}
		if (!rest) continue;
		rest = *size - i;
		if (rest > 0 && rest == var->instance_len &&
		    (snmp->command == GETNEXT_REQUEST ||
		     (var->instance &&
		      !memcmp(&objid[i], var->instance, rest * sizeof(oid))))) {
			ref->found++;
			*size = rest;
			return var;
		}
	}
	*size = 0;
	return 0;
}
#endif /* unused */

/* parse variable-bindings sequence (extract variables if any) */
static VARIABLE *
snmp_parse(sd, snmpreq, data, size)
	SESSION *sd;
	SNMP_REQ *snmpreq;
	u_char *data;
	int size;
{
	u_char *header, *value, *cp = data;
	int var_index = 0, len, type, objid_len, instance_len, length;
	oid objid[MAX_OID_LEN], *instance;
	REFERENCE *ref = snmpreq ? snmpreq->var_ref : 0;
	VARIABLE *var, *first = 0, *last;
	long integer;
	u_int64_t integer64;
	char string[MAX_STR_LEN];

	dprintf(("snmp_parse: data %p, size %d\n", data, size));

	if (!data) size = 0;

	/* get each varBind sequence */
	while (size > 0) {
		len = size;
		header = cp;
		if ((cp = parse_header(cp, &len, &type)) == NULL ||
		    type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
			trace(("\t snmp_parse: bad variable-binding header\n"));
			break;
		}
		memset(objid, 0, sizeof(objid));
		objid_len = sizeof(objid) / sizeof(oid);
		if ((cp = parse_objid(cp, &len, &type, objid, &objid_len)) == NULL ||
		    type != (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID)) {
			trace(("\t snmp_parse: bad parse objid of variable-binding\n"));
			break;
		}
		value = cp;
		if ((cp = parse_header(cp, &len, &type)) == NULL) {
			trace(("\t snmp_parse: bad parse header of variable-binding value\n"));
			break;
		}
		cp += len;
		size -= cp - header;
		var_index++;

		/* sanity check */
		if (objid_len < 4 || objid_len >= MAX_OID_LEN)
			continue;

		instance = 0;
		instance_len = objid_len;
		if (snmpreq) {
			/* skip non-snmp variables if any */
			while (ref && (!ref->var || !ref->var->objid))
				ref = ref->next;
			if (!ref) {
				trace(("\t%2d %s (0x%02X/%d) - out of sequence\n",
				       var_index, oid2str(string, objid, objid_len), type, len));
				break;
			}
			var = ref->var;
			ref = ref->next; /* advance to next reference */
			if (objid_len > var->objid_len) {
				trace(("\t%2d %s (0x%02X/%d) - bad objid len\n",
				       var_index, oid2str(string, objid, objid_len), type, len));
				continue;
			}
			/* match it exclude instance */
			for (objid_len = 0; var->objid[objid_len] &&
			     objid_len < var->objid_len; objid_len++);
			if (memcmp(objid, var->objid, objid_len * sizeof(oid))) {
				trace(("\t%2d %s (0x%02X/%d) - unexpected objid\n",
				       var_index, oid2str(string, objid, objid_len), type, len));
				continue;
			}
			/* match instance if required */
			instance = &objid[objid_len];
			instance_len = var->objid_len - objid_len;
			if (instance_len != var->instance_len) {
				trace(("\t%2d %s (0x%02X/%d) - bad instance len\n",
				       var_index, oid2str(string, objid, objid_len), type, len));
				continue;
			}
			if (snmpreq->command == GET_REQUEST && var->instance &&
			    memcmp(instance, var->instance, instance_len * sizeof(oid))) {
				trace(("\t%2d %s (0x%02X/%d) - unexpected instance\n",
				       var_index, oid2str(string, objid, objid_len), type, len));
				continue;
			}
		} else { /* trap? */
			/* try appropriate name and instance for the objid */
			const char *name = get_mib_name(objid, &instance_len, 0);
			if (name) {
				instance = &objid[objid_len - instance_len];
			} else {
				name = oid2str(string, objid, objid_len);
				instance_len = 0;
			}
			if ((var = make_var(name, objid, objid_len, 0)) == 0)
				continue;

			if (first)
				last->next = var;
			else	first = var;
			last = var;
		}

		trace(("\t%2d %s", var_index, var->name));
		if (instance_len)
			trace((".%s", oid2str(string, instance, instance_len)));
		trace((" = "));

		len = cp - value;
		switch (type) {
		case ASN_INTEGER:
			length = sizeof(integer);
			if (parse_integer(value, &len, &type, &integer, length)) {
				set_var(var, INTEGER, &integer, length);
				set_instance(var, instance, instance_len);
				trace(("%s (%s)", var2str(string, var), type2str(var->type)));
			}
			break;
		case ASN_COUNTER:
		case ASN_GAUGE:
		case ASN_TIMETICKS:
		case ASN_UINTEGER:
			length = sizeof(integer);
			if (parse_unsigned(value, &len, &type, &integer, length)) {
				set_var(var, UNSIGNED, &integer, length);
				set_instance(var, instance, instance_len);
				trace(("%s (%s)", var2str(string, var), type2str(var->type)));
			}
			break;
		case ASN_COUNTER64:
			length = sizeof(integer64);
			if (parse_unsigned64(value, &len, &type, &integer64, length)) {
				set_var(var, UNSIGNED64, &integer64, length);
				set_instance(var, instance, instance_len);
				trace(("%s (%s)", var2str(string, var), type2str(var->type)));
			}
			break;
		case ASN_FLOAT:
			report(LOG_DEBUG, "snmp_parse($%s): type FLOAT is not implemented yet!",
			       var->name);
			break;
		case ASN_DOUBLE:
			report(LOG_DEBUG, "snmp_parse($%s): type DOUBLE is not implemented yet!",
			       var->name);
			break;
		case ASN_OCTET_STR:
		case ASN_IPADDRESS:
			length = sizeof(string);
			if (parse_string(value, &len, &type, string, &length)) {
				set_var(var, type == ASN_IPADDRESS ?
					IPADDRESS : STRING, string, length);
				set_instance(var, instance, instance_len);
				trace(("%s (%s)", var2str(string, var), type2str(var->type)));
			}
			break;
		case ASN_OBJECT_ID:
			length = sizeof(objid) / sizeof(oid);
			if (parse_objid(value, &len, &type, objid, &length)) {
				set_var(var, OBJID, objid, length * sizeof(oid));
				set_instance(var, instance, instance_len);
				trace(("%s (%s)", var2str(string, var), type2str(var->type)));
			}
			break;
		case ASN_NULL:
		case ASN_NOSUCHOBJECT:
		case ASN_NOSUCHINSTANCE:
		case ASN_ENDOFMIBVIEW:
			trace(("%s", type == ASN_NULL ? "Null" : "Exception"));
			if (snmpreq && snmpreq->error_status != NOERROR &&
			    snmpreq->error_index == var_index) {
				if (sd) {
					if (sd->data_ptr) free(sd->data_ptr);
					sd->data_ptr = strdup(var->name);
				}
				if (var->length)
					set_var(var, 0, 0, 0);
			} else if (!snmpreq || snmpreq->error_status == NOERROR) {
				char *err = "Null";
				if (type == ASN_NOSUCHOBJECT)
					err = "Exception: No such Object at this OID";
				else if (type == ASN_NOSUCHINSTANCE)
					err = "Exception: No such Instance at this OID";
				else if (type == ASN_ENDOFMIBVIEW)
					err = "Exception: End of MIB view";
				set_var(var, STRING, err, strlen(err));
			}
			break;
		default:
			trace(("- unknown type 0x%02X", type));
			report(LOG_DEBUG, "snmp_parse($%s): Unknown type 0x%02X",
			       var->name, type);
		}
		trace(("\n"));
	}
	return first;
}

/*
 * Compose and send SNMP V1 request.
 */
int
snmp_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	METHOD *method = sd->method;
	REFERENCE *ref;
	u_char *seq0, *seq0end, *seq1, *seq1end, *seq2, *seq2end;
	u_char *cp, buf[MAX_PACKETSZ];
	int cmd, length = MAX_PACKETSZ, cnt;

	/* sanity check */
	if (!method || !method->argument || sd->sock < 0) {
		errno = EINVAL;
		return -1;
	}

/*	dprintf(("snmp_send(%p/%p): reqid %d\n", sd, request, request->reqid));
*/

	/* build header SEQUENCE */
	seq0 = buf;
	if ((cp = insert_sequence(buf, &length,
				  ASN_SEQUENCE | ASN_CONSTRUCTOR,
				  0)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
        seq0end = cp;
	/* version */
	if ((cp = insert_integer(cp, &length,
				 ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER,
				 (long)method->snmp_version)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
	/* community */
	if ((cp = insert_string(cp, &length,
				ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR,
				method->argument, strlen(method->argument))) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}

	if (request->snmpreq && request->snmpreq->command)
		cmd = request->snmpreq->command;
	else	cmd = GET_REQUEST;

	/* build header PDU */
	seq1 = cp;
	/* snmp command */
	if ((cp = insert_sequence(cp, &length, cmd, 0)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
	seq1end = cp;
        /* request id */
	if ((cp = insert_integer(cp, &length,
				 ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER,
				 (long)request->reqid)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
	/* error status */
	if ((cp = insert_integer(cp, &length,
				 ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER,
				 (long)0)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
	/* error index */
	if ((cp = insert_integer(cp, &length,
				 ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER,
				 (long)0)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}

	/* build variable-bindings SEQUENCE */
	seq2 = cp;
	if ((cp = insert_sequence(cp, &length,
				  ASN_SEQUENCE | ASN_CONSTRUCTOR,
				  0)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}
	seq2end = cp;
	/* variables */
	cnt = 0;
	if (request->snmpreq) {
		for (ref = request->snmpreq->var_ref; ref; ref = ref->next) {
			/* sanity check */
			if (!ref->var || !ref->var->objid)
				continue;

			if ((cp = insert_variable(cp, &length, ref->var, cmd)) == NULL) {
				errno = EMSGSIZE;
				return -1;
			}
			cnt++;
		}
	}
	/* something must be send */
	if (!cnt && (cp = insert_variable(cp, &length, NULL, cmd)) == NULL) {
		errno = EMSGSIZE;
		return -1;
	}

	/* insert actual lengths of resulting fields */

	length = MAX_PACKETSZ;
	insert_sequence(seq2, &length,
		       ASN_SEQUENCE | ASN_CONSTRUCTOR, cp - seq2end);
	insert_sequence(seq1, &length, cmd, cp - seq1end);
	insert_sequence(seq0, &length,
		       ASN_SEQUENCE | ASN_CONSTRUCTOR, cp - seq0end);

	/* send the packet */

	cmd = send(sd->sock, (char *)buf, cp - buf, 0);
	if (cmd < 0 && errno == EINPROGRESS)
		cmd = 0;
	return cmd;
}

int
snmp_recv(sd)
	SESSION *sd;
{
	METHOD *method = sd->method;
	REQUEST *request;
	u_char *cp, buf[MAX_PACKETSZ];
	struct sockaddr_in from;
	int len, tmpval, type;
	socklen_t from_len;
	char community[MAX_STR_LEN];

	from_len = sizeof(from);
	if ((len = recvfrom(sd->sock, (char *)buf, sizeof(buf), 0,
			    (struct sockaddr *)&from, &from_len)) < 0) {
		if (errno != EAGAIN)
			snmp_reply(errno, sd, 0);
		return -1;
	}

	/* message is an ASN.1 SEQUENCE */
	if ((cp = parse_header(buf, &len, &type)) == NULL) {
		trace(("\t snmp_recv: bad packet header\n"));
		return -1;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
		trace(("\t snmp_recv: bad ASN packet type\n"));
		return -1;
	}

	/* check for SNMP version */
	if ((cp = parse_integer(cp, &len, &type, &tmpval, sizeof(tmpval))) == NULL ||
	    (tmpval != SNMP_VERSION_1 && tmpval != SNMP_VERSION_2)) {
		trace(("\t snmp_recv: bad packet version\n"));
		return -1;
	}

	/* extract community string */
	tmpval = sizeof(community)-1;
	if ((cp = parse_string(cp, &len, &type, community, &tmpval)) == NULL) {
		trace(("\t snmp_recv: bad parse of packet community\n"));
		return -1;
	}
	community[tmpval] = '\0';

	/* get the PDU type */
	if ((cp = parse_header(cp, &len, &type)) == NULL) {
		trace(("\t snmp_recv: bad parse of PDU type\n"));
		return -1;
	}

	if (type != GET_RESPONSE) {
		trace(("\t snmp_recv: not a response\n"));
		return -1;
	}

	/* check for request id */
	tmpval = 0;
	if ((cp = parse_integer(cp, &len, &type, &tmpval, sizeof(tmpval))) == NULL) {
		trace(("\t snmp_recv: bad parse of packet reqid\n"));
		return -1;
	}
	if (!tmpval || (request = get_request(sd, tmpval)) == NULL) {
		trace(("\t snmp_recv: invalid reqid=%d\n", tmpval));
		return -1;
	}

/*	dprintf(("snmp_recv(%p/%p): reqid %d\n", sd, request, tmpval));
*/
	/* sanity check */
	if (!method || !method->argument) {
		snmp_reply(EINVAL, sd, 0);
		return -1;
	}


	/* sanity check for community. Perhaps it required */
	if (strcmp(method->argument, community)) {
		snmp_reply(EACCES, sd, request->reqid);
		return -1;
	}

	/* error status */
	if ((cp = parse_integer(cp, &len, &type, &tmpval, sizeof(tmpval))) == NULL) {
		trace(("\t snmp_recv: bad parse of packet error-stat\n"));
		return -1;
	}
	if (tmpval < 0 || tmpval >= NUM_ERRORS)
		tmpval = GENERR;
	sd->data_int = tmpval;
	if (request->snmpreq)
		request->snmpreq->error_status = tmpval;

	/* error index */
	if ((cp = parse_integer(cp, &len, &type, &tmpval, sizeof(tmpval))) == NULL) {
		trace(("\t snmp_recv: bad parse of packet error-index\n"));
		return -1;
	}
	if (request->snmpreq)
		request->snmpreq->error_index = tmpval;

	/* get variable-bindings sequence */
	if ((cp = parse_header(cp, &len, &type)) == NULL) {
		trace(("\t snmp_recv: bad parse of variable-bindings sequence\n"));
		return -1;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
		trace(("\t snmp_recv: bad ASN type of variable-bindings sequence\n"));
		return -1;
	}

	if (request->snmpreq)
		snmp_parse(sd, request->snmpreq, cp, len);

	return request->reqid;
}

TRAP *
snmp_trap(sock, chksrc)
	int sock, chksrc;
{
	OBJECT *object;
	TRAP *trap;
	u_char *cp, buf[MAX_PACKETSZ];
	struct sockaddr_in from;
	struct in_addr agent_addr;
	int len, tmpval, type;
	socklen_t from_len;
	int objid_len, trap_type, specific_code;
	long time_stamp;
	char ipaddr[20], community[MAX_STR_LEN];
	oid objid[MAX_OID_LEN];
#ifdef	TRACE
	char tmpbuf[1024];
#endif

	from_len = sizeof(from);
	if ((len = recvfrom(sock, (char *)buf, sizeof(buf), 0,
			    (struct sockaddr *)&from, &from_len)) < 0) {
		trace(("\t snmp_trap: recvfrom: %s\n", strerror(errno)));
		return NULL;
	}

	dprintf(("snmp_trap(%s): len %d\n", intoa(ipaddr, from.sin_addr), len));

	/* message is an ASN.1 SEQUENCE */
	if ((cp = parse_header(buf, &len, &type)) == NULL) {
		trace(("\t snmp_trap: bad packet header\n"));
		return NULL;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
		trace(("\t snmp_trap: bad ASN packet type\n"));
		return NULL;
	}

	/* check for SNMP version */
	if ((cp = parse_integer(cp, &len, &type, &tmpval, sizeof(tmpval))) == NULL ||
	    tmpval != SNMP_VERSION_1) {
		trace(("\t snmp_trap: bad packet version\n"));
		return NULL;
	}

	/* extract community string */
	tmpval = sizeof(community)-1;
	if ((cp = parse_string(cp, &len, &type, community, &tmpval)) == NULL) {
		trace(("\t snmp_trap: bad parse of packet community\n"));
		return NULL;
	}
	community[tmpval] = '\0';

	/* get the PDU type */
	if ((cp = parse_header(cp, &len, &type)) == NULL) {
		trace(("\t snmp_trap: bad parse of PDU type\n"));
		return NULL;
	}

	if (type != TRAP_PDU) {
		trace(("\t snmp_trap: not a trap PDU\n"));
		return NULL;
	}

	/* enterprise */
	objid_len = sizeof(objid) / sizeof(oid);
	if ((cp = parse_objid(cp, &len, &type, objid, &objid_len)) == NULL) {
		trace(("\t snmp_trap: bad parse of packet enterprise\n"));
		return NULL;
	}

	/* agent-addr */
	tmpval = sizeof(agent_addr);
	if ((cp = parse_string(cp, &len, &type, (u_char *)&agent_addr, &tmpval)) == NULL) {
		trace(("\t snmp_trap: bad parse of packet agent-addr\n"));
		return NULL;
	}

	if (chksrc && memcmp(&from.sin_addr, &agent_addr, tmpval)) {
		report(LOG_NOTICE, "Trap from %s mismatch it agent-addr", intoa(ipaddr, from.sin_addr));
		return NULL;
	}

	/* generic trap type */
	if ((cp = parse_integer(cp, &len, &type, &trap_type, sizeof(trap_type))) == NULL) {
		trace(("\t snmp_trap: bad parse of trap type\n"));
		return NULL;
	}

	/* specific trap code */
	if ((cp = parse_integer(cp, &len, &type, &specific_code, sizeof(specific_code))) == NULL) {
		trace(("\t snmp_trap: bad parse of specific type\n"));
		return NULL;
	}

	/* time stamp  */
	if ((cp = parse_unsigned(cp, &len, &type, &time_stamp, sizeof(time_stamp))) == NULL) {
		trace(("\t snmp_trap: bad parse of time stamp\n"));
		return NULL;
	}

	/* get variable-bindings sequence */
	if ((cp = parse_header(cp, &len, &type)) == NULL) {
		trace(("\t snmp_trap: bad parse of variable-bindings sequence\n"));
		return NULL;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
		trace(("\t snmp_trap: bad ASN type of variable-bindings sequence\n"));
		return NULL;
	}

	trace(("\t enterprise %s, agent %s, type %d/%d, uptime %u\n",
	       oid2str(tmpbuf, objid, objid_len), intoa(ipaddr, agent_addr),
	       trap_type, specific_code, time_stamp));


	/* trap authorization */
	if ((object = trap_auth(&agent_addr, community, objid, objid_len, trap_type, specific_code)) == NULL)
		return NULL;

	if ((trap = (TRAP *)malloc(sizeof(TRAP))) == NULL) return NULL;
	memset(trap, 0, sizeof(TRAP));

	trap->community = strdup(community);
	if ((trap->objid = (oid *)malloc(objid_len * sizeof(oid))) == NULL) {
		free(trap);
		return NULL;
	}
	memcpy(trap->objid, objid, objid_len * sizeof(oid));
	trap->objid_len = objid_len;
	trap->type = trap_type;
	trap->specific = specific_code;
	trap->uptime = time_stamp;
	trap->owner = object;
	trap->var_list = snmp_parse(0, 0, cp, len);

	return trap;
}

static void
snmp_stop(method)
	METHOD *method;
{
	session_stop(method->sd, 1);
}

static void
snmp_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int reqid;

	/* sanity check */
	if (!sd) return;

	snmp_stop(method); /* stop the method if still running */

	/* open datagram socket near before UDP session */
	if ((sd->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		snmp_reply(errno, sd, 0);
		return;
	}

	/* turn on non-blocking I/O */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		snmp_reply(errno, sd, 0);
		return;
	}

	/*
	 * Make datagram socket `connected', so error on this socket
	 * will be returned asynchronously without timing out.
	 */
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		snmp_reply(errno, sd, 0);
		return;
	}

	if ((reqid = session_start(sd, dup_snmp_request(method->snmpreq))) != 0) {
		dprintf(("snmp_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, reqid));
	}
}

static void
snmp_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[1024];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		op = sd->data_int; /* just snmp error status */
		if (op >= 0 && op < NUM_ERRORS)
			diag = snmp_error[op];
		else	diag = "Unknown error";
		if (op != NOERROR &&
		    sd->data_ptr && strlen(sd->data_ptr) < 100) {
			(void)strcpy(buf, sd->data_ptr);
			(void)strcat(buf, ": ");
			(void)strcat(buf, diag);
			diag = buf;
		}
	} else	diag = strerror(op);

	dprintf(("snmp_reply(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	dump_var_list(target->var_list);

	snmp_stop(method);

	method_finished(target, method, diag, !op);
}

static int
snmp_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("snmp_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = snmp_send;
	template.recv = snmp_recv;
	template.read = snmp_reply;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_snmp = {
	snmp_init,	/* initialize SNMP session */
	snmp_start,	/* start SNMP session */
	snmp_stop,	/* stop SNMP session */

	"SNMP",		/* method name */
	"public",	/* community by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_UDP,	/* network protocol */
	SNMPSERVER_PORT,/* server port */
	0, 0,		/* timeout and retries undefined yet */
	{ SNMP_VERSION_1,/* version number */
	  0 },		 /* no parameter used */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
