/*
 *	Copyright (c) 2000-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "md5.h"

extern int errno;

/*
 * TACACS+ specification according to Cisco Draft V1.78.
 */

#define	TACACSSERVER_PORT	49
#define	HEADER_LEN		12
#define	DIGEST_LEN		16
#define	MAX_PACKETSZ		1024
#define	TEST_NAME		NETMON
#define	TEST_PORT		"tty0"

typedef	struct tacacs_header {
	u_char	version;		/* major & minor version numbers */
#define	MAJOR_VERSION		0xc0	/* mandatory */
#define	MINOR_VERSION		0x01	/* for PAP authentication */
	u_char	type;		/* type of TACACS packet */
#define	TYPE_AUTHEN		1	/* authentication */
#define	TYPE_AUTHOR		2	/* authorization */
#define	TYPE_ACCT		3	/* accounting */
	u_char	seq_no;		/* sequence number */
	u_char	flags;		/* bit mapped */
#define	FLAG_UNENCRYPTED	1	/* if set - body unencrypted */
	u_int32_t session_id;
	u_int32_t length;	/* length of packet body */
	u_char	body[1];
} HEADER;

#define	AUTHEN_LOGIN		1
#define	AUTHEN_TYPE_PAP		2
#define	AUTHEN_STATUS_PASS	1
#define	AUTHEN_STATUS_FAIL	2

static void tacacs_reply __P((int, SESSION *, int));

/*
 * Check to see if an TCP connection established at this session.
 */
static int
tacacs_connect(sd)
	SESSION *sd;
{
	REQUEST *request;
	METHOD *method = sd->method;

	/* sanity check */
	if (!method) return -1;

	if ((request = get_request(sd, 0 /* no reqid */)) == NULL) {
		trace(("\t tacacs_connect: internal error?\n"));
		return -1;
	}

	dprintf(("tacacs_connect(%s/%s)\n", sd->owner->name, method->name));

	/*
	 * OK, we are connected to a peer.
	 */

	/* if no secret-key configured just close request */
	if (!method->argument || !*method->argument) {
		sd->data_int = 0;
		tacacs_reply(0, sd, request->reqid);
		return -1;
	}

	/*
	 * Prepare us for TACACS authentication.
	 */

	/* turn off non-blocking I/O after connecting */
	if (set_socket_async(sd->sock, FALSE) < 0) {
		tacacs_reply(errno, sd, request->reqid);
		return -1;
	}

	/* no more `awaiting' connection for the session */
	sd->connect = NULL;

	/* set defaults timeout and retries */
	sd->timeout = method->timeout * 1000000L; /* make miscroseconds */
	sd->retries = 1;	/* only single attempt */

	return 0;
}

static void
tacacs_crypt(packet, secret)
	u_char *packet;
	char *secret;
{
	HEADER *hp = (HEADER *)packet;
	u_char *bp, *eob, *cp, digest[DIGEST_LEN], buf[MAX_PACKETSZ];
	int len, i;

	if (hp->flags & FLAG_UNENCRYPTED)
		return;

	bp = hp->body;
	eob = bp + ntohl(hp->length);

	if (bp >= eob) /* body is empty */
		return;

	/* prepare for md5 digest */
	cp = buf;
	memcpy(cp, &hp->session_id, sizeof(hp->session_id));
	cp += sizeof(hp->session_id);
	memcpy(cp, secret, strlen(secret));
	cp += strlen(secret);
	memcpy(cp, &hp->version, sizeof(hp->version));
	cp += sizeof(hp->version);
	memcpy(cp, &hp->seq_no, sizeof(hp->seq_no));
	cp += sizeof(hp->seq_no);

	/* calculate first md5 digest */
	len = cp - buf;
	md5_calc(digest, buf, len);
	len += DIGEST_LEN;

	/* XOR the packet body into md5 digest */
	for (;;) {
		for (i = 0; i < DIGEST_LEN; i++) {
			*bp++ ^= digest[i];
			if (bp >= eob) return;
		}

		/* concatenate previous md5 digest and complement ones */
		memcpy(cp, digest, DIGEST_LEN);
		md5_calc(digest, buf, len);
	}
}

static int
tacacs_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	u_char *cp, buf[MAX_PACKETSZ];
	HEADER *hp = (HEADER *)buf;
	METHOD *method = sd->method;
	int len;
	char *secret;

	/* sanity check */
	if (!method || !method->argument || !*method->argument) {
		errno = EINVAL;
		return -1;
	}
	secret = method->argument;

	/* build the TACACS+ protocol packet header */
	memset(hp, 0, HEADER_LEN);
	hp->version = (MAJOR_VERSION | MINOR_VERSION);
	hp->type = TYPE_AUTHEN;
	hp->seq_no = 1;	/* START packet */

	hp->session_id = htonl(request->reqid);
/*	len = sizeof(TIMEVAL);
	if (len > sizeof(hp->session_id) len = sizeof(hp->session_id);
	memcpy(&hp->session_id, &sd->owner->last_request, len);
*/

	/* build START packet body for the PAP Authentication */
	cp = hp->body;
	*cp++ = AUTHEN_LOGIN;		/* action */
	*cp++ = 0;			/* priv_lvl */
	*cp++ = AUTHEN_TYPE_PAP;	/* authen_type */
	*cp++ = 0;			/* service */
	*cp++ = sizeof(TEST_NAME)-1;	/* user_len */
	*cp++ = sizeof(TEST_PORT)-1;	/* port_len */
	*cp++ = 0;			/* rem_addr_len */
	*cp++ = strlen(secret);		/* data_len (password length) */
	len = sizeof(TEST_NAME)-1;
	memcpy(cp, TEST_NAME, len);
	cp += len;
	len = sizeof(TEST_PORT)-1;
	memcpy(cp, TEST_PORT, len);
	cp += len;
	len = strlen(secret);
	memcpy(cp, secret, len); /* the secret used as user password */
	cp += len;

	len = cp - hp->body;
	hp->length = htonl(len);
	len += HEADER_LEN;

	tacacs_crypt(buf, secret); /* encrypt packet */

	dprintf(("tacacs_send(%s/%s): %d bytes\n",
		 sd->owner->name, method->name, len));

	return write(sd->sock, buf, len);
}

static int
tacacs_recv(sd)
	SESSION *sd;
{
	METHOD *method = sd->method;
	REQUEST *request;
	u_char buf[MAX_PACKETSZ];
	HEADER *hp;
	int len;

	if ((request = get_request(sd, 0 /* no reqid */)) == NULL) {
		trace(("\t tacacs_recv: internal error?\n"));
		return -1;
	}

	if ((len = read(sd->sock, buf, sizeof(buf))) < 1) {
		if (!len) { /* connection closed by peer? */
			sd->data_int = 0;
#ifdef	ECONNABORTED
			errno =	ECONNABORTED;
#elif	ECONNRESET
			errno =	ECONNRESET;
#else
			errno = EPIPE;
#endif
		}
		tacacs_reply(errno, sd, request->reqid);
		return -1;
	}

	dprintf(("tacacs_recv(%s/%s): %d bytes\n",
		 sd->owner->name, sd->method->name, len));

	if (len < HEADER_LEN + 6) {
		trace(("\t tacacs_recv: undersized packet, len=%d\n", len));
		return -1;
	}
	hp = (HEADER *)buf;
	if (ntohl(hp->length) != len - HEADER_LEN) {
		trace(("\t tacacs_recv: wrong body length\n"));
		return -1;
	}
	if (ntohl(hp->session_id) != request->reqid) {
		trace(("\t tacacs_recv: invalid session id\n"));
		return -1;
	}
	if (hp->type != TYPE_AUTHEN || hp->seq_no != 2) {
		trace(("\t tacacs_recv: not a response\n"));
		return -1;
	}

	if (method && method->argument)
		tacacs_crypt(buf, method->argument); /* decrypt packet */

	len = ntohl(hp->length);
	if (sd->data_ptr) free(sd->data_ptr);
	if ((sd->data_ptr = (char *)malloc(len)) != NULL) {
		memmove(sd->data_ptr, hp->body, len);
		sd->data_int = len;
	} else 	sd->data_int = 0;

	return 0;
}

static void
tacacs_stop(method)
	METHOD *method;
{
	session_stop(method->sd, 1);
}

static void
tacacs_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int reqid;

	/* sanity check */
	if (!sd) return;

	tacacs_stop(method); /* stop the method is still running */

	/* open stream socket near before TCP session */
	if ((sd->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		tacacs_reply(errno, sd, 0);
		return;
	}

	/* turn on non-blocking I/O before connecting */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		tacacs_reply(errno, sd, 0);
		return;
	}
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0 &&
	    errno != EINPROGRESS) {
		tacacs_reply(errno, sd, 0);
		return;
	}

	/* timeout and retries for connection */
	sd->timeout = method->timeout * 1000000L; /* make miscroseconds */
	sd->retries = method->retries;
	sd->connect = tacacs_connect;	/* `awaiting' connection */

	if ((reqid = session_start(sd, NULL)) != 0) {
		dprintf(("tacacs_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, reqid));
	}
}

/*
 * Parse REPLY packet. Return server_msg if any.
 */
static char *
tacacs_parse(buf, body, length)
	char *buf;
	u_char *body;
	int length;
{
	u_char *cp;
	short len;
	int i;

	*buf = '\0';
	if (length < 1) return buf;
	cp = body + 2;
	memcpy(&len, cp, 2);
	len = ntohs(len);

	if (!len) {
		if (*body != AUTHEN_STATUS_PASS)
			(void)strcpy(buf, "Authentication failed");
		return buf;
	}
	if (len > length - 6)
		return strcpy(buf, "Invalid secret key?");

	cp += 4;
	if (len < 0 || len > 100) {
		len = 100;
		if (len > length - 6) len = length - 6;
	}
	for (i = 0; i < len && isprint(cp[i]); i++);
	if (i < len)
		return strcpy(buf, "Invalid secret key?");

	memcpy(buf, cp, len);
	buf[len] = '\0';

	return buf;
}

static void
tacacs_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[MAX_PACKETSZ];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		diag = tacacs_parse(buf, sd->data_ptr, sd->data_int);
		if (!diag) {
			op = -1;
			diag = "Unknown error";
		} else if (!*diag) {
			diag = "No error";
		}
	} else	diag = strerror(op);

	dprintf(("tacacs_reply(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	tacacs_stop(method);

	method_finished(target, method, diag, !op);
}

static int
tacacs_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("tacacs_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.connect = tacacs_connect;
	template.send = tacacs_send;
	template.recv = tacacs_recv;
	template.read = tacacs_reply;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_tacacs = {
	tacacs_init,	/* initialize tacacs protocol session */
	tacacs_start,	/* start tacacs session */
	tacacs_stop,	/* stop tacacs session */

	"TACACS+",	/* method name */
	NULL,		/* no argument by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_TCP,	/* network protocol */
	TACACSSERVER_PORT, /* server port */
	0, 0,		/* timeout and retries undefined yet */
	{ 0, 0 },	/* no parameters used */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
