/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "regex.h"

extern int errno;

static void tcp_close __P((int, SESSION *, int));

/*
 * Check to see if an TCP connection established at this session.
 */
static int
tcp_connect(sd)
	SESSION *sd;
{
	REQUEST *request;
	METHOD *method = sd->method;

	/* sanity check */
	if (!method) return -1;

	if ((request = get_request(sd, 0 /* no reqid */)) == NULL) {
		trace(("\t tcp_connect: internal error?\n"));
		return -1;
	}

	dprintf(("tcp_connect(%s/%s)\n", sd->owner->name, method->name));

	/*
	 * OK, we are connected to a peer.
	 */

	/* if no chat-script configured just close request */
	if (!method->chatscript) {
		sd->data_int = 1;
		tcp_close(0, sd, request->reqid);
		return -1;
	}

	/*
	 * Prepare us for chat-script.
	 */

	/* turn off non-blocking I/O after connecting */
	if (set_socket_async(sd->sock, FALSE) < 0) {
		tcp_close(errno, sd, request->reqid);
		return -1;
	}

	/* no more `awaiting' connection for the session */
	sd->connect = NULL;

	/* set defaults timeout and retries for chat-script */
	sd->timeout = method->timeout * 1000000L; /* make miscroseconds */
	sd->retries = 1;	/* only single attempt */

	/* set chat-script for the session descriptor */
	sd->cs = method->chatscript;

	return 0;
}

/*
 * Send the data through TCP session.
 */
static int
tcp_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	CHATSCRIPT *cs;

	/* sanity check */
	if (!sd || !sd->cs) {
		errno = EINVAL;
		return -1;
	}

	/*
	 * Handle multiply send without expect in chat-script pairs.
	 */
	for (cs = sd->cs; cs; cs = cs->next) {
		if (!cs->send || !*cs->send)
			break;	/* nothing to send */

		if (write(sd->sock, cs->send, strlen(cs->send)) < 0)
			return -1;

		trace(("\t tcp_send: \"%s\"\n", cs->send));

		if (cs->expect && *cs->expect)
			break;	/* expect follow */
	}

	/* set pair for expect */
	sd->cs = cs;

	if (!sd->cs) { /* end of chat-script */
		sd->data_int = 1;
		tcp_close(0, sd, request->reqid);
	}

	return 0;
}

int
match_expect(sd, cs, data)
	SESSION *sd;
	CHATSCRIPT *cs;
	char *data;
{
	yre_t prog;
	int len;
	REFERENCE *ref;
	char *top, *cp, buf[1024], tag[10];

	dprintf(("match_expect(%s/%s): \"%s\"\n",
		 sd->owner->name, sd->method->name, data));

	if (yre_comp(&prog, cs->expect)) {
		errno = EINVAL;
		return -1;
	}
	if (!yre_exec(&prog, data) || !yre_subs(&prog, "&", buf)) {
		trace(("\t not matched \"%s\"\n", cs->expect));
		return 0; /* expect not matched */
	}

	trace(("\t matched \"%s\"\n", buf));

	/* concatenate diagnostic string */

	len = strlen(buf);
	if (sd->data_ptr && *sd->data_ptr)
		len += strlen(sd->data_ptr);
	if ((top = malloc(len + 2 /* coma and \0 */)) == NULL) {
		errno = ENOMEM;
		return -1;
	}
	cp = top;
	if (sd->data_ptr && *sd->data_ptr) {
		cp += strlen(strcpy(cp, sd->data_ptr));
		*cp++ = ',';
		free(sd->data_ptr);
	}
	(void)strcpy(cp, buf);
	sd->data_int = strlen(top);
	sd->data_ptr = top;

	/* extract value of variables if any */

	len = 1;
	for (ref = cs->var_ref; ref; ref = ref->next) {
		sprintf(tag, "\\%d", len++);
		if (yre_subs(&prog, tag, buf)) {
			trace(("\t substitute %s \"%s\"\n", tag, buf));
			if (ref->var)
				str2var(ref->var, buf);
		}
	}

	return 1;	/* expect matched */
}

/*
 * Receive data through TCP session.
 */
static int
tcp_recv(sd)
	SESSION *sd;
{
	REQUEST *request;
	CHATSCRIPT *cs;
	int len, rest;
	char *cp, *line, buf[BUFSIZ+1];

	if ((request = get_request(sd, 0 /* no reqid */)) == NULL) {
		trace(("\t tcp_recv: internal error?\n"));
		return -1;
	}

	/* sanity check */
	if (!sd || !sd->cs) {
		tcp_close(EINVAL, sd, request->reqid);
		return -1;
	}

	if (sd->buf) { /* previous line was truncated */
		rest = strlen(strcpy(buf, sd->buf));
		free(sd->buf);
		sd->buf = NULL;
	} else 	rest = 0;

	if ((len = read(sd->sock, &buf[rest], (sizeof(buf)-1) - rest)) < 1) {
		if (!len) { /* connection closed by peer? */
			sd->data_int = 0;
			errno = 0;
		}
		tcp_close(errno, sd, request->reqid);
		return -1;
	}
	buf[rest + len] = '\0';

	trace(("\t tcp_recv: \"%s\" (%d)\n", buf, rest + len));

	/*
	 * Handle multiply expect without send in chat-script pairs.
	 */

	if (strchr(buf, '\n') == NULL) { /* non linear data */
		cs = sd->cs;
		do {
			if (cs->expect && *cs->expect && !cs->matched) {
				if ((len = match_expect(sd, cs, buf)) < 0) {
					tcp_close(errno, sd, request->reqid);
					return -1;
				}
				cs->matched = len;
			}
			cs = cs->next;
		} while (cs && (!cs->send || !*cs->send));
		cp = NULL;
	} else {
		rest = 0;
		/* consider received data as array of lines */
		for (cp = buf; (line = strchr(cp, '\n')) != NULL; cp = line) {
			if (line > cp && line[-1] == '\r') line[-1] = '\0';
			*line++ = '\0';
			trace(("\t line #%d \"%s\"\n", ++rest, cp));
			cs = sd->cs;
			do {
				if (cs->expect && *cs->expect && !cs->matched) {
					if ((len = match_expect(sd, cs, cp)) < 0) {
						tcp_close(errno, sd, request->reqid);
						return -1;
					}
					cs->matched = len;
				}
				cs = cs->next;
			} while (cs && (!cs->send || !*cs->send));
		}
		/* cp point to possibly truncated line */
	}

	/*
	 * If any of expects (without send) not matched then tailing data.
	 */
	cs = sd->cs;
	do {
		if (cs->expect && *cs->expect && !cs->matched) {
			/* need to expect rest of data for match */
			if (cp && *cp) /* last line has truncated */
				sd->buf = strdup(cp);
			sd->data_int = -1; /* try to tailing data */
			return -1;
		}
		cs = cs->next;
	} while (cs && (!cs->send || !*cs->send));

	if (cp && *cp && cs)
		return -1; /* need to drain rest of incoming data */

	/* set pair for send */
	sd->cs = cs;

	if (sd->cs) {	/* send next string */
		sd->data_int = 0;
		if (tcp_send(sd, request) < 0)
			tcp_close(errno, sd, request->reqid);
		return -1;
	}

	/* end of chat-script */
	return 0;
}

void
tcp_stop(method)
	METHOD *method;
{
	CHATSCRIPT *cs;

	/* clear chat-script matches */
	for (cs = method->chatscript; cs; cs = cs->next)
		cs->matched = 0;

	session_stop(method->sd, 1);
}

void
tcp_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int tmpval;

	/* sanity check */
	if (!sd) return;

	tcp_stop(method); /* stop the method if still running */

	/* open stream socket near before TCP session */
	if ((sd->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		tcp_close(errno, sd, 0);
		return;
	}

	/* allocate local port if required */
	if (method->lport_min) {
		struct sockaddr_in sin;

		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		tmpval = method->lport_min;
		do {
			sin.sin_port = htons((u_short)tmpval);
			if (!bind(sd->sock, (struct sockaddr *)&sin, sizeof(sin))) {
				tmpval = 0;
				break;
			}
			if (errno != EADDRINUSE) {
				tcp_close(errno, sd, 0);
				return;
			}
		} while (++tmpval <= method->lport_max);

		if (tmpval) {
			tcp_close(EAGAIN, sd, 0);
			return;
		}
	}

	/* turn on non-blocking I/O before connecting */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		tcp_close(errno, sd, 0);
		return;
	}
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0 &&
	    errno != EINPROGRESS) {
		tcp_close(errno, sd, 0);
		return;
	}

	/* timeout and retries for connection */
	sd->timeout = method->timeout * 1000000L; /* make miscroseconds */
	sd->retries = method->retries;
	sd->connect = tcp_connect;	/* `awaiting' connection */

	if ((tmpval = session_start(sd, NULL)) != 0) {
		dprintf(("tcp_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, tmpval));
	}
}

static void
tcp_close(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[100];
	int tmpval;

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		op = !sd->data_int;
		if (op)
			diag = "Chat-script not matched";
		else if (sd->data_ptr) {
			diag = strncpy(buf, sd->data_ptr, sizeof(buf));
			buf[sizeof(buf)-1] = '\0';
		} else	diag = "";
	} else if (op == ETIMEDOUT) {
		if (sd->data_int == -1)
			diag = "Chat-script not matched";
		else if ((tmpval = get_socket_error(sd->sock)) != 0)
			diag = strerror(tmpval);
		else	diag = strerror(op);
	} else	diag = strerror(op);

	dprintf(("tcp_close(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	dump_var_list(target->var_list);

	tcp_stop(method);

	method_finished(target, method, diag, !op);
}

int
tcp_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("tcp_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.connect = tcp_connect;
	template.send = tcp_send;
	template.recv = tcp_recv;
	template.read = tcp_close;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}
