/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"

extern int errno;

static void udp_close __P((int, SESSION *, int));

/*
 * Send the data through UDP session.
 */
static int
udp_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	CHATSCRIPT *cs;
	int sent = 0;

	/*
	 * Handle multiply send without expect in chat-script pairs.
	 */
	for (cs = sd->cs; cs; cs = cs->next) {
		if (!cs->send || !*cs->send)
			break;	/* nothing to send */

		if (send(sd->sock, cs->send, strlen(cs->send)+1, 0) < 0 &&
		    errno != EINPROGRESS) {
			return -1;
		}
		sent++;
		trace(("\t udp_send: \"%s\"\n", cs->send));

		if (cs->expect && *cs->expect)
			break;	/* expect follow */
	}

	/* set pair for expect */
	sd->cs = cs;

	/*
	 * If no chat-script configured, just send something
	 * to detect remote port reachable.
	 */
	if (!sd->cs && !sent) { /* no chat-script or end of chat-script */
		if (send(sd->sock, "", 1, 0) < 0 && errno != EINPROGRESS)
			return -1;
		trace(("\t udp_send: \"\"\n"));
	}

	return 0;
}

/*
 * Receive data through UDP session.
 */
static int
udp_recv(sd)
	SESSION *sd;
{
	REQUEST *request;
	CHATSCRIPT *cs;
	struct sockaddr_in from;
	int len, tmpval;
	socklen_t from_len;
	char *cp, *line, buf[BUFSIZ+1];

	/* if no chat-script configured, just report OK */
	if (!sd->cs) {
		sd->data_int = 1;
		return 0;
	}

	if ((request = get_request(sd, 0 /* no reqid */)) == NULL) {
		trace(("\t udp_recv: internal error?\n"));
		return -1;
	}

	from_len = sizeof(from);
	if ((len = recvfrom(sd->sock, (char *)buf, sizeof(buf)-1, 0,
			    (struct sockaddr *)&from, &from_len)) < 0) {
		if (errno != EAGAIN)
			udp_close(errno, sd, request->reqid);
		return -1;
	}
	buf[len] = '\0';

	trace(("\t udp_recv: \"%s\" (%d)\n", buf, len));

	/*
	 * Handle multiply expect without send in chat-script pairs.
	 */

	if (strchr(buf, '\n') == NULL) { /* non linear data */
		cs = sd->cs;
		do {
			if (cs->expect && *cs->expect && !cs->matched) {
				if ((len = match_expect(sd, cs, buf)) < 0) {
					udp_close(errno, sd, request->reqid);
					return -1;
				}
				cs->matched = len;
			}
			cs = cs->next;
		} while (cs && (!cs->send || !*cs->send));
	} else {
		/* consider received data as array of lines */
		for (cp = buf; (line = strchr(cp, '\n')) != NULL; cp = line) {
			if (line > cp && line[-1] == '\r') line[-1] = '\0';
			*line++ = '\0';
			cs = sd->cs;
			do {
				if (cs->expect && *cs->expect && !cs->matched) {
					if ((len = match_expect(sd, cs, cp)) < 0) {
						udp_close(errno, sd, request->reqid);
						return -1;
					}
					cs->matched = len;
				}
				cs = cs->next;
			} while (cs && (!cs->send || !*cs->send));
		}
		/* truncated line has been discarded */
	}

	/*
	 * If any of expects (without send) not matched -- break chat-script.
	 */
	cs = sd->cs;
	len = tmpval = 0;
	do {
		if (cs->expect && *cs->expect) {
			len++;
			tmpval += cs->matched;
		}
		cs = cs->next;
	} while (cs && (!cs->send || !*cs->send));

	if (len != tmpval) {
		sd->data_int = 0; /* chat-script aborted */
		return 0;
	}

	/* set pair for send */
	sd->cs = cs;

	if (sd->cs) {	/* send next string */
		if (udp_send(sd) < 0)
			udp_close(errno, sd, request->reqid);
		return -1;
	}

	/* end of chat-script */
	return 0;
}

void
udp_stop(method)
	METHOD *method;
{
	CHATSCRIPT *cs;

	/* clear chat-script matches */
	for (cs = method->chatscript; cs; cs = cs->next)
		cs->matched = 0;

	session_stop(method->sd, 1);
}

void
udp_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int tmpval;

	/* sanity check */
	if (!sd) return;

	udp_stop(method); /* stop the method if still running */

	/* open datagram socket near before UDP session */
	if ((sd->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		udp_close(errno, sd, 0);
		return;
	}

	/* allocate local port if required */
	if (method->lport_min) {
		struct sockaddr_in sin;

		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		tmpval = method->lport_min;
		do {
			sin.sin_port = htons((u_short)tmpval);
			if (!bind(sd->sock, (struct sockaddr *)&sin, sizeof(sin))) {
				tmpval = 0;
				break;
			}
			if (errno != EADDRINUSE) {
				udp_close(errno, sd, 0);
				return;
			}
		} while (++tmpval <= method->lport_max);

		if (tmpval) {
			udp_close(EAGAIN, sd, 0);
			return;
		}
	}

	/* turn on non-blocking I/O */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		udp_close(errno, sd, 0);
		return;
	}

	/*
	 * Make datagram socket `connected', so error on this socket
	 * will be returned asynchronously without timing out.
	 */
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		udp_close(errno, sd, 0);
		return;
	}

	sd->timeout = method->timeout * 1000000L; /* make microseconds */
	sd->retries = method->retries;
	sd->cs = method->chatscript;

	if ((tmpval = session_start(sd, NULL)) != 0) {
		dprintf(("udp_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, tmpval));
	}
}

static void
udp_close(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[100];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		op = !sd->data_int;
		if (op)
			diag = "Chat-script not matched";
		else if (sd->data_ptr) {
			diag = strncpy(buf, sd->data_ptr, sizeof(buf));
			buf[sizeof(buf)-1] = '\0';
		} else	diag = "";
	} else	diag = strerror(op);

	dprintf(("udp_close(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	dump_var_list(target->var_list);

	udp_stop(method);

	method_finished(target, method, diag, !op);
}

int
udp_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("udp_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = udp_send;
	template.recv = udp_recv;
	template.read = udp_close;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	return 0;
}
