/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/* Access to internal structure and variables */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <netdb.h>
#include <ctype.h>

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"
#include "mib.h"

extern CONFIG *cf;

static char empty[] = "\"\"";
static char null[] = "0";
static char buf[BUFSIZ];
static char *strbuf = NULL;

static char *obj_states[2] = {
	"UP", "DOWN" };
static char *if_states[5] = {
	"UP", "DOWN", "TESTING", "UNKNOWN", "DORMANT" };
static char *bgp_states[6] = {
	"IDLE", "CONNECT", "ACTIVE", "OPENSENT", "OPENCONFIRM", "ESTABLISHED" };
static char *env_states[5] = {
	"NORMAL", "WARNING", "CRITICAL", "SHUTDOWN", "NOTPRESENT" };

static struct states_ent {
	int size;
	char **name;
} states[4] = {
	{ 2, obj_states },
	{ 5, if_states },
	{ 6, bgp_states },
	{ 5, env_states },
};

char *
state2str(what, state)
	int what;
	int state;
{
	static char buf[20];
	struct states_ent *sp;
	char *cp;

	sp = &states[what];
	if (!state)
		cp = "NONE";
	else if (state > 0 && state <= sp->size)
		cp = sp->name[state-1];
	else	cp = "ERROR";

	buf[0] = '"';
	(void)strcpy(&buf[1], cp);
	(void)strcat(buf, "\"");
	return buf;
}

static char *
setstrbuf(len)
	int len;
{
	static int strbuf_len = 0;

	if (len < 1) return NULL;
	if (len > strbuf_len) {
		if (strbuf) free(strbuf);
		if ((strbuf = (char *)malloc(len + 1)) != NULL)
			strbuf_len = len;
		else	strbuf_len = 0;
	}
	if (strbuf) *strbuf = '\0';
	return strbuf;
}

static char *
sprint_time()
{
	TIMEVAL tv;

	gettimeofday(&tv, NULL);
	if (cf->timefmt) {
#ifdef	HAVE_PTHREAD
		pthread_mutex_lock(&localtime_lock);
#endif
		strftime(buf, sizeof(buf), cf->timefmt, localtime((time_t *)&tv.tv_sec));
#ifdef	HAVE_PTHREAD
		pthread_mutex_unlock(&localtime_lock);
#endif
	} else	sprintf(buf, "%u", (u_int)tv.tv_sec);

	return buf;
}

static char *
sprint_datadir(save)
	SAVE *save;
{
	char *cp;

	while (save) {
		if (save->method == SAVE_FILE && save->file) {
			sprintf(buf, "\"%s\"", save->file);
			if ((cp = strrchr(buf, '/')) != NULL) {
				*cp++ = '"';
				*cp = '\0';
			}
			return buf;
		}
		save = save->next;
	}
	return empty;
}

static int
get_var_subname(token)
	char *token;
{
	int i;
	static char *subnames[4] = {
		"ave", "min", "max", "delta" };

	for (i = 0; i < 4; i++) {
		if (!strcasecmp(token, subnames[i]))
			return i;
	}
	return -1;
}

static char *
soft_variables(var, name)
	register VARIABLE *var;
	char *name;
{
	int len;
	char *cp;

	for (; var; var = var->next) {
		/* sanity check */
		if (!var->name) continue;

		len = strlen(var->name);
		if (strncasecmp(var->name, name, len))
			continue; /* not matched */

		cp = &name[len];
		if (*cp == '\0') /* matched exactly */
			return var2str(buf, var);

		if (*cp != '.') continue;
		cp++;

		if ((var->type == INTEGER || var->type == UNSIGNED ||
		     var->type == UNSIGNED64 || var->type == FLOATING) &&
		    (len = get_var_subname(cp)) != -1) {
			STATISTIC *st = var->statistic;
			u_int64_t value = 0;

			if (st && st->count) {
				switch (len) {
				case 0:	value = st->sum / st->count; break;
				case 1: value = st->min; break;
				case 2: value = st->max; break;
				case 3: value = st->delta; break;
				}
			}
			sprintf(buf, "%llu", value);
			return buf;
		}
#ifdef	USE_VAR_STATE
		if ((var->type == INTEGER || var->type == UNSIGNED) &&
		    !strcasecmp(cp, "state")) {
			if (var->value.integer)
				len = *var->value.integer;
			else	len = -1;
			if (!strncasecmp(var->name, "bgp", 3))
				return state2str(BGP_STATE, len);
			if (!strncasecmp(var->name, "ciscoEnv", 8))
				return state2str(ENV_STATE, len);
			return state2str(IF_STATE, len);
		}
#endif
		if (!strcasecmp(cp, "old"))
			return var2str(buf, var->old);
	}
	/*
	report(LOG_DEBUG, "\"%s\": No such variable", name);
	*/
	return null;
}

/*
 * Config file main settings.
 */

static char *
main_rootdir()
{
	if (!cf->rootdir) return empty;
	sprintf(buf, "\"%s\"", cf->rootdir);
	return buf;
}

static char *
main_timefmt()
{
	if (!cf->timefmt) return empty;
	sprintf(buf, "\"%s\"", cf->timefmt);
	return buf;
}

static char *
main_polling()
{
	sprintf(buf, "%d", cf->polling);
	return buf;
}

static char *
main_timeout()
{
	sprintf(buf, "%d", cf->timeout);
	return buf;
}

static char *
main_retries()
{
	sprintf(buf, "%d", cf->retries);
	return buf;
}

static char *
main_saving()
{
	sprintf(buf, "%d", cf->saving);
	return buf;
}

VAR_TABLE main_table[] = {
	/* name		old?	function	*/

	{ "ROOTDIR",	0,	main_rootdir	},
	{ "TIMEFMT",	0,	main_timefmt	},
	{ "POLLING",	0,	main_polling	},
	{ "TIMEOUT",	0,	main_timeout	},
	{ "RETRIES",	0,	main_retries	},
	{ "SAVING",	0,	main_saving	},


	{ "TIME",	0,	sprint_time	},
	{ 0 },
};


/*
 * Object settings and variables matched by object name.
 */

static char *
object_name(object)
	OBJECT *object;
{
	sprintf(buf, "\"%s\"", object->name);
	return buf;
}

static char *
object_descr(object)
	OBJECT *object;
{
	if (!object->descr) return empty;
	sprintf(buf, "\"%s\"", object->descr);
	return buf;
}

static char *
object_address(object)
	OBJECT *object;
{
	sprintf(buf, "\"%s\"", object->address);
	return buf;
}

static char *
object_polling(object)
	OBJECT *object;
{
	sprintf(buf, "%d", object->polling);
	return buf;
}

static char *
object_timeout(object)
	OBJECT *object;
{
	sprintf(buf, "%d", object->timeout);
	return buf;
}

static char *
object_retries(object)
	OBJECT *object;
{
	sprintf(buf, "%d", object->retries);
	return buf;
}

static char *
object_saving(object)
	OBJECT *object;
{
	sprintf(buf, "%d", object->saving);
	return buf;
}

static char *
object_datadir(object)
	OBJECT *object;
{
	return sprint_datadir(object->save_list);
}

static char *
object_state(object)
	OBJECT *object;
{
	return state2str(OBJ_STATE, object->state);
}

static char *
object_prevstate(object)
	OBJECT *object;
{
	return state2str(OBJ_STATE, object->prev_state);
}

static char *
object_lastrequest(object)
	OBJECT *object;
{
	sprintf(buf, "%u", tv_round(&object->last_request));
	return buf;
}

static char *
object_prevrequest(object)
	OBJECT *object;
{
	sprintf(buf, "%u", tv_round(&object->prev_request));
	return buf;
}

static char *
object_lastreply(object)
	OBJECT *object;
{
	sprintf(buf, "%u", tv_round(&object->last_reply));
	return buf;
}

static char *
object_prevreply(object)
	OBJECT *object;
{
	sprintf(buf, "%u", tv_round(&object->prev_reply));
	return buf;
}

static char *
object_lastchange(object)
	OBJECT *object;
{
	sprintf(buf, "%u", tv_round(&object->last_change));
	return buf;
}

static char *
object_netpath(object)
	OBJECT *object;
{
	int len = 0;
	char *cp;
	NEXTHOP *hop, *last = NULL;

	hop = object->nexthop;
	while (hop) {
		if (hop->target) {
			len += strlen(hop->target->name) + 2;
			if (hop->interface) {
				cp = hop->interface->name;
				if (!cp) cp = itoa(hop->interface->index);
				len += strlen(cp) + 2;
			}
		}
		last = hop;
		hop = hop->next;
	}
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	while (last) {
		if (strbuf[1]) *cp++ = ',';
		if (last->target) {
			if (last->interface)
				cp += sprintf(cp, "%s!%s", last->target->name,
					      last->interface->name ?
					      last->interface->name :
					      itoa(last->interface->index));
			else	cp += strlen(strcpy(cp, last->target->name));
		}
		last = last->prev;
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
object_hopcount(object)
	OBJECT *object;
{
	sprintf(buf, "%d", object->hop_count);
	return buf;
}

static char *
object_variables(object, name)
	OBJECT *object;
	char *name;
{
	return soft_variables(object->var_list, name);
}

VAR_TABLE object_table[] = {
	/* name		old?	function		*/

	{ "NAME",	0,	object_name		},
	{ "DESCRIPTION",0,	object_descr		},
	{ "ADDRESS",	0,	object_address		},
	{ "POLLING",	0,	object_polling		},
	{ "TIMEOUT",	0,	object_timeout		},
	{ "RETRIES",	0,	object_retries		},
	{ "SAVING",	0,	object_saving		},
	{ "DATADIR",	0,	object_datadir		},

	{ "STATE",	0,	object_state		},
	{ "STATE",	1,	object_prevstate	},
	{ "QUERYTIME",	0,	object_lastrequest	},
	{ "QUERYTIME",	1,	object_prevrequest	},
	{ "REPLYTIME",	0,	object_lastreply	},
	{ "REPLYTIME",	1,	object_prevreply	},
	{ "LASTCHANGE",	0,	object_lastchange	},

	{ "NETPATH",	0,	object_netpath		},
	{ "HOPCOUNT",	0,	object_hopcount		},

	{ "",		0,	object_variables	},
	{ 0 },
};

VAR_TABLE service_table[] = {
	/* name		old?	function		*/

	{ "NAME",	0,	object_name		},
	{ "DESCRIPTION",0,	object_descr		},
	{ "DATADIR",	0,	object_datadir		},

	{ "STATE",	0,	object_state		},
	{ "STATE",	1,	object_prevstate	},
	{ "QUERYTIME",	0,	object_lastrequest	},
	{ "QUERYTIME",	1,	object_prevrequest	},
	{ "REPLYTIME",	0,	object_lastreply	},
	{ "REPLYTIME",	1,	object_prevreply	},
	{ "LASTCHANGE",	0,	object_lastchange	},

	{ "",		0,	object_variables	},
	{ 0 },
};


/*
 * Interface specific settings and variables matched by interface name
 * (or index).
 */

static char *
iface_name(iface)
	INTERFACE *iface;
{
	sprintf(buf, "\"%s\"", iface->name ? iface->name : itoa(iface->index));
	return buf;
}

static char *
iface_descr(iface)
	INTERFACE *iface;
{
	if (!iface->descr) return empty;
	sprintf(buf, "\"%s\"", iface->descr);
	return buf;
}

static char *
iface_bandwidth(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", iface->bandwidth);
	return buf;
}

static char *
iface_datadir(iface)
	INTERFACE *iface;
{
	return sprint_datadir(iface->save_list);
}

static char *
iface_state(iface)
	INTERFACE *iface;
{
	return state2str(IF_STATE, iface->state);
}

static char *
iface_prevstate(iface)
	INTERFACE *iface;
{
	return state2str(IF_STATE, iface->prev_state);
}

static char *
iface_lastrequest(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", tv_round(&iface->last_request));
	return buf;
}

static char *
iface_prevrequest(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", tv_round(&iface->prev_request));
	return buf;
}

static char *
iface_lastreply(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", tv_round(&iface->last_reply));
	return buf;
}

static char *
iface_prevreply(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", tv_round(&iface->prev_reply));
	return buf;
}

static char *
iface_lastchange(iface)
	INTERFACE *iface;
{
	sprintf(buf, "%u", tv_round(&iface->last_change));
	return buf;
}

static char *
iface_address(iface)
	INTERFACE *iface;
{
	int len = 0;
	IFADDR *addr;
	char *cp, ipaddr[100];

	for (addr = iface->ifaddr; addr; addr = addr->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (addr = iface->ifaddr; addr; addr = addr->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, intoa(ipaddr, addr->ip_addr)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static struct IANAifType_ent {
	char *name;
	int type;
	char *descr;
} IANAifType[] = {
 { "other", ifType_other, "" },
 { "regular1822", ifType_regular1822, "" },
 { "hdh1822", ifType_hdh1822, "" },
 { "ddnX25", ifType_ddnX25, "" },
 { "rfc877x25", ifType_rfc877x25, "" },
 { "ethernetCsmacd", ifType_ethernetCsmacd, "" },
 { "iso88023Csmacd", ifType_iso88023Csmacd, "" },
 { "iso88024TokenBus", ifType_iso88024TokenBus, "" },
 { "iso88025TokenRing", ifType_iso88025TokenRing, "" },
 { "iso88026Man", ifType_iso88026Man, "" },
 { "starLan", ifType_starLan, "" },
 { "proteon10Mbit", ifType_proteon10Mbit, "" },
 { "proteon80Mbit", ifType_proteon80Mbit, "" },
 { "hyperchannel", ifType_hyperchannel, "" },
 { "fddi", ifType_fddi, "" },
 { "lapb", ifType_lapb, "" },
 { "sdlc", ifType_sdlc, "" },
 { "ds1", ifType_ds1, "DS1-MIB" },
 { "e1", ifType_e1, "Obsolete see DS1-MIB" },
 { "basicISDN", ifType_basicISDN, "" },
 { "primaryISDN", ifType_primaryISDN, "" },
 { "propPointToPointSerial", ifType_propPointToPointSerial, "proprietary serial" },
 { "ppp", ifType_ppp, "" },
 { "softwareLoopback", ifType_softwareLoopback, "" },
 { "eon", ifType_eon, "CLNP over IP" },
 { "ethernet3Mbit", ifType_ethernet3Mbit, "" },
 { "nsip", ifType_nsip, "XNS over IP" },
 { "slip", ifType_slip, "generic SLIP" },
 { "ultra", ifType_ultra, "ULTRA technologies" },
 { "ds3", ifType_ds3, "DS3-MIB" },
 { "sip", ifType_sip, "SMDS, coffee" },
 { "frameRelay", ifType_frameRelay, "DTE only." },
 { "rs232", ifType_rs232, "" },
 { "para", ifType_para, "parallel-port" },
 { "arcnet", ifType_arcnet, "arcnet" },
 { "arcnetPlus", ifType_arcnetPlus, "arcnet plus" },
 { "atm", ifType_atm, "ATM cells" },
 { "miox25", ifType_miox25, "" },
 { "sonet", ifType_sonet, "SONET or SDH" },
 { "x25ple", ifType_x25ple, "" },
 { "iso88022llc", ifType_iso88022llc, "" },
 { "localTalk", ifType_localTalk, "" },
 { "smdsDxi", ifType_smdsDxi, "" },
 { "frameRelayService", ifType_frameRelayService, "FRNETSERV-MIB" },
 { "v35", ifType_v35, "" },
 { "hssi", ifType_hssi, "" },
 { "hippi", ifType_hippi, "" },
 { "modem", ifType_modem, "Generic modem" },
 { "aal5", ifType_aal5, "AAL5 over ATM" },
 { "sonetPath", ifType_sonetPath, "" },
 { "sonetVT", ifType_sonetVT, "" },
 { "smdsIcip", ifType_smdsIcip, "SMDS InterCarrier Interface" },
 { "propVirtual", ifType_propVirtual, "proprietary virtual/internal" },
 { "propMultiplexor", ifType_propMultiplexor, "proprietary multiplexing" },
 { "ieee80212", ifType_ieee80212, "100BaseVG" },
 { "fibreChannel", ifType_fibreChannel, "Fibre Channel" },
 { "hippiInterface", ifType_hippiInterface, "HIPPI interfaces" },
 { "frameRelayInterconnect", ifType_frameRelayInterconnect, "Obsolete use either" },
 { "aflane8023", ifType_aflane8023, "ATM Emulated LAN for 802.3" },
 { "aflane8025", ifType_aflane8025, "ATM Emulated LAN for 802.5" },
 { "cctEmul", ifType_cctEmul, "ATM Emulated circuit" },
 { "fastEther", ifType_fastEther, "Fast Ethernet 100BaseT" },
 { "isdn", ifType_isdn, "ISDN and X.25" },
 { "v11", ifType_v11, "CCITT V.11/X.21" },
 { "v36", ifType_v36, "CCITT V.36" },
 { "g703at64k", ifType_g703at64k, "CCITT G703 at 64Kbps" },
 { "g703at2mb", ifType_g703at2mb, "Obsolete see DS1-MIB" },
 { "qllc", ifType_qllc, "SNA QLLC" },
 { "fastEtherFX", ifType_fastEtherFX, "Fast Ethernet 100BaseFX" },
 { "channel", ifType_channel, "channel" },
 { "ieee80211", ifType_ieee80211, "radio spread spectrum" },
 { "ibm370parChan", ifType_ibm370parChan, "IBM System 360/370 OEMI Channel" },
 { "escon", ifType_escon, "IBM Enterprise Systems Connection" },
 { "dlsw", ifType_dlsw, "Data Link Switching" },
 { "isdns", ifType_isdns, "ISDN S/T interface" },
 { "isdnu", ifType_isdnu, "ISDN U interface" },
 { "lapd", ifType_lapd, "Link Access Protocol D" },
 { "ipSwitch", ifType_ipSwitch, "IP Switching Objects" },
 { "rsrb", ifType_rsrb, "Remote Source Route Bridging" },
 { "atmLogical", ifType_atmLogical, "ATM Logical Port" },
 { "ds0", ifType_ds0, "Digital Signal Level 0" },
 { "ds0Bundle", ifType_ds0Bundle, "group of ds0s on the same ds1" },
 { "bsc", ifType_bsc, "Bisynchronous Protocol" },
 { "async", ifType_async, "Asynchronous Protocol" },
 { "cnr", ifType_cnr, "Combat Net Radio" },
 { "iso88025Dtr", ifType_iso88025Dtr, "ISO 802.5r DTR" },
 { "eplrs", ifType_eplrs, "Ext Pos Loc Report Sys" },
 { "arap", ifType_arap, "Appletalk Remote Access Protocol" },
 { "propCnls", ifType_propCnls, "Proprietary Connectionless Protocol" },
 { "hostPad", ifType_hostPad, "CCITT-ITU X.29 PAD Protocol" },
 { "termPad", ifType_termPad, "CCITT-ITU X.3 PAD Facility" },
 { "frameRelayMPI", ifType_frameRelayMPI, "Multiproto Interconnect over FR" },
 { "x213", ifType_x213, "CCITT-ITU X213" },
 { "adsl", ifType_adsl, "Asymmetric Digital Subscriber Loop" },
 { "radsl", ifType_radsl, "Rate-Adapt. Digital Subscriber Loop" },
 { "sdsl", ifType_sdsl, "Symmetric Digital Subscriber Loop" },
 { "vdsl", ifType_vdsl, "Very H-Speed Digital Subscrib. Loop" },
 { "iso88025CRFPInt", ifType_iso88025CRFPInt, "ISO 802.5 CRFP" },
 { "myrinet", ifType_myrinet, "Myricom Myrinet" },
 { "voiceEM", ifType_voiceEM, "voice recEive and transMit" },
 { "voiceFXO", ifType_voiceFXO, "voice Foreign Exchange Office" },
 { "voiceFXS", ifType_voiceFXS, "voice Foreign Exchange Station" },
 { "voiceEncap", ifType_voiceEncap, "voice encapsulation" },
 { "voiceOverIp", ifType_voiceOverIp, "voice over IP encapsulation" },
 { "atmDxi", ifType_atmDxi, "ATM DXI" },
 { "atmFuni", ifType_atmFuni, "ATM FUNI" },
 { "atmIma", ifType_atmIma, "ATM IMA" },
 { "pppMultilinkBundle", ifType_pppMultilinkBundle, "PPP Multilink Bundle" },
 { "ipOverCdlc", ifType_ipOverCdlc, "IBM ipOverCdlc" },
 { "ipOverClaw", ifType_ipOverClaw, "IBM Common Link Access to Workstn" },
 { "stackToStack", ifType_stackToStack, "IBM stackToStack" },
 { "virtualIpAddress", ifType_virtualIpAddress, "IBM VIPA" },
 { "mpc", ifType_mpc, "IBM multi-protocol channel support" },
 { "ipOverAtm", ifType_ipOverAtm, "IBM ipOverAtm" },
 { "iso88025Fiber", ifType_iso88025Fiber, "ISO 802.5j Fiber Token Ring" },
 { "tdlc", ifType_tdlc, "IBM twinaxial data link control" },
 { "gigabitEthernet", ifType_gigabitEthernet, "Gigabit Ethernet" },
 { "hdlc", ifType_hdlc, "HDLC" },
 { "lapf", ifType_lapf, "LAP F" },
 { "v37", ifType_v37, "V.37" },
 { "x25mlp", ifType_x25mlp, "Multi-Link Protocol" },
 { "x25huntGroup", ifType_x25huntGroup, "X25 Hunt Group" },
 { "trasnpHdlc", ifType_trasnpHdlc, "Transp HDLC" },
 { "interleave", ifType_interleave, "Interleave channel" },
 { "fast", ifType_fast, "Fast channel" },
 { "ip", ifType_ip, "IP for APPN HPR in IP networks" },
 { "docsCableMaclayer", ifType_docsCableMaclayer, "CATV Mac Layer" },
 { "docsCableDownstream", ifType_docsCableDownstream, "CATV Downstream interface" },
 { "docsCableUpstream", ifType_docsCableUpstream, "CATV Upstream interface" },
 { "a12MppSwitch", ifType_a12MppSwitch, "Avalon Parallel Processor" },
 { "tunnel", ifType_tunnel, "Encapsulation interface" },
 { "coffee", ifType_coffee, "coffee pot" },
 { "ces", ifType_ces, "Circuit Emulation Service" },
 { "atmSubInterface", ifType_atmSubInterface, "ATM Sub Interface" },
 { "l2vlan", ifType_l2vlan, "Layer 2 Virtual LAN using 802.1Q" },
 { "l3ipvlan", ifType_l3ipvlan, "Layer 3 Virtual LAN using IP" },
 { "l3ipxvlan", ifType_l3ipxvlan, "Layer 3 Virtual LAN using IPX" },
 { "digitalPowerline", ifType_digitalPowerline, "IP over Power Lines" },
 { "mediaMailOverIp", ifType_mediaMailOverIp, "Multimedia Mail over IP" },
 { "dtm", ifType_dtm, "Dynamic syncronous Transfer Mode" },
 { "dcn", ifType_dcn, "Data Communications Network" },
 { "ipForward", ifType_ipForward, "IP Forwarding Interface" },
 { "msdsl", ifType_msdsl, "Multi-rate Symmetric DSL" },
 { "ieee1394", ifType_ieee1394, "IEEE1394 High Performance Serial Bus" },
 { "if_gsn", ifType_if_gsn, "HIPPI-6400" },
 { "dvbRccMacLayer", ifType_dvbRccMacLayer, "DVB-RCC MAC Layer" },
 { "dvbRccDownstream", ifType_dvbRccDownstream, "DVB-RCC Downstream Channel" },
 { "dvbRccUpstream", ifType_dvbRccUpstream, "DVB-RCC Upstream Channel" },
 { "atmVirtual", ifType_atmVirtual, "ATM Virtual Interface" },
 { "mplsTunnel", ifType_mplsTunnel, "MPLS Tunnel Virtual Interface" },
 { "srp", ifType_srp, "Spatial Reuse Protocol" },
 { "voiceOverAtm", ifType_voiceOverAtm, "Voice Over ATM" },
 { "voiceOverFrameRelay", ifType_voiceOverFrameRelay, "Voice Over Frame Relay" },
 { "idsl", ifType_idsl, "Digital Subscriber Loop over ISDN" },
 { "compositeLink", ifType_compositeLink, "Avici Composite Link Interface" },
 { "ss7SigLink", ifType_ss7SigLink, "SS7 Signaling Link" },
 { "propWirelessP2P", ifType_propWirelessP2P, "Prop. P2P wireless interface" },
 { "frForward", ifType_frForward, "Frame Forward Interface" },
 { "rfc1483", ifType_rfc1483, "Multiprotocol over ATM AAL5" },
 { "usb", ifType_usb, "USB Interface" },
 { "ieee8023adLag", ifType_ieee8023adLag, "IEEE 802.3ad Link Aggregate" },
 { "bgppolicyaccounting", ifType_bgppolicyaccounting, "BGP Policy Accounting" },
 { "frf16MfrBundle", ifType_frf16MfrBundle, "FRF .16 Multilink Frame Relay" },
 { "h323Gatekeeper", ifType_h323Gatekeeper, "H323 Gatekeeper" },
 { "h323Proxy", ifType_h323Proxy, "H323 Voice and Video Proxy" },
 { "mpls", ifType_mpls, "MPLS" },
 { "mfSigLink", ifType_mfSigLink, "Multi-frequency signaling link" },
 { "hdsl2", ifType_hdsl2, "High Bit-Rate DSL - 2nd generation" },
 { "shdsl", ifType_shdsl, "Multirate HDSL2" },
 { "ds1FDL", ifType_ds1FDL, "Facility Data Link 4Kbps on a DS1" },
 { "pos", ifType_pos, "Packet over SONET/SDH Interface" },
 { "dvbAsiIn", ifType_dvbAsiIn, "DVB-ASI Input" },
 { "dvbAsiOut", ifType_dvbAsiOut, "DVB-ASI Output" },
 { "plc", ifType_plc, "Power Line Communtications" },
 { "nfas", ifType_nfas, "Non Facility Associated Signaling" },
 { "tr008", ifType_tr008, "TR008" },
 { "gr303RDT", ifType_gr303RDT, "Remote Digital Terminal" },
 { "gr303IDT", ifType_gr303IDT, "Integrated Digital Terminal" },
 { "isup", ifType_isup, "ISUP" },
 { "propDocsWirelessMaclayer", ifType_propDocsWirelessMaclayer, "prop/Maclayer" },
 { "propDocsWirelessDownstream", ifType_propDocsWirelessDownstream, "prop/Downstream" },
 { "propDocsWirelessUpstream", ifType_propDocsWirelessUpstream, "prop/Upstream" },
 { "hiperlan2", ifType_hiperlan2, "HIPERLAN Type 2 Radio Interface" },
 { "propBWAp2Mp", ifType_propBWAp2Mp, "PropBroadbandWirelessAccesspt2multipt" },
 { "sonetOverheadChannel", ifType_sonetOverheadChannel, "SONET Overhead Channel" },
 { "digitalWrapperOverheadChannel", ifType_digitalWrapperOverheadChannel, "Digital Wrapper" },
 { "aal2", ifType_aal2, "ATM adaptation layer 2" },
 { "radioMAC", ifType_radioMAC, "MAC layer over radio links" },
 { "atmRadio", ifType_atmRadio, "ATM over radio links" },
 { "imt", ifType_imt, "Inter Machine Trunks" },
 { "mvl", ifType_mvl, "Multiple Virtual Lines DSL" },
 { "reachDSL", ifType_reachDSL, "Long Reach DSL" },
 { "frDlciEndPt", ifType_frDlciEndPt, "Frame Relay DLCI End Point" },
 { "atmVciEndPt", ifType_atmVciEndPt, "ATM VCI End Point" },
 { "opticalChannel", ifType_opticalChannel, "Optical Channel" },
 { "opticalTransport", ifType_opticalTransport, "Optical Transport" },
 { "propAtm", ifType_propAtm, "Proprietary ATM" },
 { "voiceOverCable", ifType_voiceOverCable, "Voice Over Cable Interface" },
 { "infiniband", ifType_infiniband, "Infiniband" },
 { "teLink", ifType_teLink, "TE Link" },
 { "q2931", ifType_q2931, "Q.2931" },
 { "virtualTg", ifType_virtualTg, "Virtual Trunk Group" },
 { "sipTg", ifType_sipTg, "SIP Trunk Group" },
 { "sipSig", ifType_sipSig, "SIP Signaling" },
 { "docsCableUpstreamChannel", ifType_docsCableUpstreamChannel, "CATV Upstream Channel" },
 { "econet", ifType_econet, "Acorn Econet" },
 { "pon155", ifType_pon155, "FSAN 155Mb Symetrical PON interface" },
 { "pon622", ifType_pon622, "FSAN622Mb Symetrical PON interface" },
 { "bridge", ifType_bridge, "Transparent bridge interface" },
 { "linegroup", ifType_linegroup, "Interface common to multiple lines" },
 { "voiceEMFGD", ifType_voiceEMFGD, "voice E&M Feature Group D" },
 { "voiceFGDEANA", ifType_voiceFGDEANA, "voice FGD Exchange Access North American" },
 { "voiceDID", ifType_voiceDID, "voice Direct Inward Dialing" },
 { "mpegTransport", ifType_mpegTransport, "MPEG transport interface" },
 { "sixToFour", ifType_sixToFour, "6to4 interface" },
 { "gtp", ifType_gtp, "GPRS Tunneling Protocol" },
 { "pdnEtherLoop1", ifType_pdnEtherLoop1, "Paradyne EtherLoop 1" },
 { "pdnEtherLoop2", ifType_pdnEtherLoop2, "Paradyne EtherLoop 2" },
 { "opticalChannelGroup", ifType_opticalChannelGroup, "Optical Channel Group" },
{ 0 }
};

static char *
iface_ianaiftype(iface)
	INTERFACE *iface;
{
	VARIABLE *var = get_var(iface->var_list, "ifType");
	if (var && var->type == INTEGER && var->value.integer) {
		struct IANAifType_ent *e = IANAifType;
		for (; e && e->name; e++) {
			if (e->type == *var->value.integer) {
				if (e->descr && *e->descr)
					sprintf(buf, "\"%s (%s)\"", e->name, e->descr);
				else	sprintf(buf, "\"%s\"", e->name);
				return buf;
			}
		}
	}
	return empty;
}

static char *
iface_variables(iface, name)
	INTERFACE *iface;
	char *name;
{
	return soft_variables(iface->var_list, name);
}

VAR_TABLE iface_table[] = {
	/* name		old?	function		*/

	{ "NAME",	0,	iface_name		},
	{ "DESCRIPTION",0,	iface_descr		},
	{ "BANDWIDTH",	0,	iface_bandwidth		},
	{ "DATADIR",	0,	iface_datadir		},

	{ "STATE",	0,	iface_state		},
	{ "STATE",	1,	iface_prevstate		},
	{ "QUERYTIME",	0,	iface_lastrequest	},
	{ "QUERYTIME",	1,	iface_prevrequest	},
	{ "REPLYTIME",	0,	iface_lastreply		},
	{ "REPLYTIME",	1,	iface_prevreply		},
	{ "LASTCHANGE",	0,	iface_lastchange	},
	{ "ADDRESS",	0,	iface_address		},
	{ "IANAIFTYPE",	0,	iface_ianaiftype	},

	{ "",		0,	iface_variables		},
	{ 0 },
};

char *
get_bgppeer_name(buf, in)
	char *buf;
	struct in_addr in;
{
	OBJECT *target;
	IFADDR *ifaddr;
	char ipaddr[20];

	for (target = cf->target; target; target = target->next) {
		for (ifaddr = target->ifaddr; ifaddr; ifaddr = ifaddr->next) {
			if (!memcmp(&in, &ifaddr->ip_addr, sizeof(struct in_addr)))
				return target->name;
		}
	}
	(void)strcpy(buf, intoa(ipaddr, in));
	return buf;
}

static char *
peer_state(peer)
	BGP_PEER *peer;
{
	return state2str(BGP_STATE, peer->state);
}

static char *
peer_prevstate(peer)
	BGP_PEER *peer;
{
	return state2str(BGP_STATE, peer->prev_state);
}

static char *
peer_lastchange(peer)
	BGP_PEER *peer;
{
	sprintf(buf, "%u", tv_round(&peer->last_change));
	return buf;
}

/*
 * BGP specific settings and variables matched by AS number.
 */

static char *
bgp_asn(bgp)
	BGP_AS *bgp;
{
	sprintf(buf, "%d", bgp->asn);
	return buf;
}

static char *
bgp_name(bgp)
	BGP_AS *bgp;
{
	sprintf(buf, "\"AS%d\"", bgp->asn);
	return buf;
}

static char *
bgp_descr(bgp)
	BGP_AS *bgp;
{
	if (!bgp->descr) return empty;
	sprintf(buf, "\"%s\"", bgp->descr);
	return buf;
}

static char *
bgp_datadir(bgp)
	BGP_AS *bgp;
{
	return sprint_datadir(bgp->save_list);
}

static char *
bgp_state(bgp)
	BGP_AS *bgp;
{
	int len = 0;
	BGP_PEER *peer;
	char *cp;

	for (peer = bgp->peer; peer; peer = peer->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (peer = bgp->peer; peer; peer = peer->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, peer_state(peer)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
bgp_prevstate(bgp)
	BGP_AS *bgp;
{
	int len = 0;
	BGP_PEER *peer;
	char *cp;

	for (peer = bgp->peer; peer; peer = peer->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (peer = bgp->peer; peer; peer = peer->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, peer_prevstate(peer)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
bgp_peer(bgp)
	BGP_AS *bgp;
{
	int len = 0;
	BGP_PEER *peer;
	char *cp, tmpbuf[100];

	for (peer = bgp->peer; peer; peer = peer->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (peer = bgp->peer; peer; peer = peer->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, get_bgppeer_name(tmpbuf, peer->ip_addr)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
bgp_lastchange(bgp)
	BGP_AS *bgp;
{
	int len = 0;
	BGP_PEER *peer;
	char *cp;

	for (peer = bgp->peer; peer; peer = peer->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (peer = bgp->peer; peer; peer = peer->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, peer_lastchange(peer)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

VAR_TABLE bgp_table[] = {
	/* name		old?	function		*/

	{ "ASN",	0,	bgp_asn			},
	{ "NAME",	0,	bgp_name		},
	{ "DESCRIPTION",0,	bgp_descr		},
	{ "DATADIR",	0,	bgp_datadir		},

	{ "MEMBERS",	0,	bgp_peer		},
	{ "STATE",	0,	bgp_state		},
	{ "STATE",	1,	bgp_prevstate		},
	{ "LASTCHANGE",	0,	bgp_lastchange		},

	{ 0 },
};


/*
 * BGP Peer specific settings and variables.
 */

static char *
peer_name(peer)
	BGP_PEER *peer;
{
	char tmpbuf[100];
	sprintf(buf, "\"%s\"", get_bgppeer_name(tmpbuf, peer->ip_addr));
	return buf;
}

static char *
peer_asn(peer)
	BGP_PEER *peer;
{
	if (!peer->parent) return empty;
	return bgp_asn(peer->parent);
}

static char *
peer_descr(peer)
	BGP_PEER *peer;
{
	if (!peer->parent) return empty;
	return bgp_descr(peer->parent);
}

static char *
peer_datadir(peer)
	BGP_PEER *peer;
{
	if (!peer->parent) return empty;
	return bgp_datadir(peer->parent);
}

static char *
peer_lastrequest(peer)
	BGP_PEER *peer;
{
	sprintf(buf, "%u", tv_round(&peer->last_request));
	return buf;
}

static char *
peer_prevrequest(peer)
	BGP_PEER *peer;
{
	sprintf(buf, "%u", tv_round(&peer->prev_request));
	return buf;
}

static char *
peer_lastreply(peer)
	BGP_PEER *peer;
{
	sprintf(buf, "%u", tv_round(&peer->last_reply));
	return buf;
}

static char *
peer_prevreply(peer)
	BGP_PEER *peer;
{
	sprintf(buf, "%u", tv_round(&peer->prev_reply));
	return buf;
}

static char *
peer_variables(peer, name)
	BGP_PEER *peer;
	char *name;
{
	return soft_variables(peer->var_list, name);
}

VAR_TABLE peer_table[] = {
	/* some data partially inherited from parent bgp_as */

	{ "NAME",	0,	peer_name		},
	{ "ASN",	0,	peer_asn		},
	{ "DESCRIPTION",0,	peer_descr		},
	{ "DATADIR",	0,	peer_datadir		},
	{ "STATE",	0,	peer_state		},
	{ "STATE",	1,	peer_prevstate		},
	{ "QUERYTIME",	0,	peer_lastrequest	},
	{ "QUERYTIME",	1,	peer_prevrequest	},
	{ "REPLYTIME",	0,	peer_lastreply		},
	{ "REPLYTIME",	1,	peer_prevreply		},
	{ "LASTCHANGE",	0,	peer_lastchange		},

	{ "",		0,	peer_variables		},
	{ 0 },
};

static char *
envgauge_state(gauge)
	ENV_GAUGE *gauge;
{
	return state2str(ENV_STATE, gauge->state);
}

static char *
envgauge_prevstate(gauge)
	ENV_GAUGE *gauge;
{
	return state2str(ENV_STATE, gauge->prev_state);
}

static char *
envgauge_lastchange(gauge)
	ENV_GAUGE *gauge;
{
	u_long sec = gauge->parent ? tv_round(&gauge->parent->last_change) : 0;
	sprintf(buf, "%u", sec);
	return buf;
}

/*
 * Environment Monitor variables.
 */

static char *
envmon_name(envmon)
	ENV_MON *envmon;
{
	sprintf(buf, "\"%s\"", get_name_by_type(envmon->type));
	return buf;
}

static char *
envmon_descr(envmon)
	ENV_MON *envmon;
{
	if (!envmon->descr) return empty;
	sprintf(buf, "\"%s\"", envmon->descr);
	return buf;
}

static char *
envmon_datadir(envmon)
	ENV_MON *envmon;
{
	return sprint_datadir(envmon->save_list);
}

static char *
envmon_state(envmon)
	ENV_MON *envmon;
{
	int len = 0;
	ENV_GAUGE *gauge;
	char *cp;

	for (gauge = envmon->gauge; gauge; gauge = gauge->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (gauge = envmon->gauge; gauge; gauge = gauge->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, envgauge_state(gauge)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
envmon_prevstate(envmon)
	ENV_MON *envmon;
{
	int len = 0;
	ENV_GAUGE *gauge;
	char *cp;

	for (gauge = envmon->gauge; gauge; gauge = gauge->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (gauge = envmon->gauge; gauge; gauge = gauge->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, envgauge_prevstate(gauge)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
envmon_gauge(envmon)
	ENV_MON *envmon;
{
	int len = 0;
	ENV_GAUGE *gauge;
	char *cp;

	for (gauge = envmon->gauge; gauge; gauge = gauge->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (gauge = envmon->gauge; gauge; gauge = gauge->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += sprintf(cp, "\"%s\"", gauge->name);
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

static char *
envmon_lastchange(envmon)
	ENV_MON *envmon;
{
	int len = 0;
	ENV_GAUGE *gauge;
	char *cp;

	for (gauge = envmon->gauge; gauge; gauge = gauge->next)
		len += 100;
	if ((cp = setstrbuf(len)) == NULL)
		return empty;
	*cp++ = '"';
	*cp = '\0';
	for (gauge = envmon->gauge; gauge; gauge = gauge->next) {
		if (strbuf[1]) *cp++ = ',';
		cp += strlen(strcpy(cp, envgauge_lastchange(gauge)));
	}
	*cp++ = '"';
	*cp = '\0';
	return strbuf;
}

VAR_TABLE envmon_table[] = {
	/* name		old?	function		*/

	{ "NAME",	0,	envmon_name		},
	{ "DESCRIPTION",0,	envmon_descr		},
	{ "DATADIR",	0,	envmon_datadir		},

	{ "MEMBERS",	0,	envmon_gauge		},
	{ "STATE",	0,	envmon_state		},
	{ "STATE",	1,	envmon_prevstate	},
	{ "LASTCHANGE",	0,	envmon_lastchange	},

	{ 0 },
};


/*
 * Environment temperature variables.
 */

static char *
envgauge_name(gauge)
	ENV_GAUGE *gauge;
{
	sprintf(buf, "\"%s\"", gauge->name);
	return buf;
}

static char *
envgauge_descr(gauge)
	ENV_GAUGE *gauge;
{
	if (!gauge->parent) return empty;
	return envmon_descr(gauge->parent);
}

static char *
envgauge_datadir(gauge)
	ENV_GAUGE *gauge;
{
	if (!gauge->parent) return empty;
	return envmon_datadir(gauge->parent);
}

static char *
envgauge_lastrequest(gauge)
	ENV_GAUGE *gauge;
{
	u_long sec = gauge->parent ? tv_round(&gauge->parent->last_request) : 0;
	sprintf(buf, "%u", sec);
	return buf;
}

static char *
envgauge_prevrequest(gauge)
	ENV_GAUGE *gauge;
{
	u_long sec = gauge->parent ? tv_round(&gauge->parent->prev_request) : 0;
	sprintf(buf, "%u", sec);
	return buf;
}

static char *
envgauge_lastreply(gauge)
	ENV_GAUGE *gauge;
{
	u_long sec = gauge->parent ? tv_round(&gauge->parent->last_reply) : 0;
	sprintf(buf, "%u", sec);
	return buf;
}

static char *
envgauge_prevreply(gauge)
	ENV_GAUGE *gauge;
{
	u_long sec = gauge->parent ? tv_round(&gauge->parent->prev_reply) : 0;
	sprintf(buf, "%u", sec);
	return buf;
}

static char *
envgauge_variables(gauge, name)
	ENV_GAUGE *gauge;
	char *name;
{
	return soft_variables(gauge->var_list, name);
}

VAR_TABLE envgauge_table[] = {
	/* name		old?	function		*/

	{ "NAME",	0,	envgauge_name		},
	{ "DESCRIPTION",0,	envgauge_descr		},
	{ "DATADIR",	0,	envgauge_datadir	},
	{ "STATE",	0,	envgauge_state		},
	{ "STATE",	1,	envgauge_prevstate	},
	{ "QUERYTIME",	0,	envgauge_lastrequest	},
	{ "QUERYTIME",	1,	envgauge_prevrequest	},
	{ "REPLYTIME",	0,	envgauge_lastreply	},
	{ "REPLYTIME",	1,	envgauge_prevreply	},
	{ "LASTCHANGE",	0,	envgauge_lastchange	},

	{ "",		0,	envgauge_variables	},
	{ 0 },
};

static char *
return_null()
{
	return null;
}

static char *
(*do_variable(vtab, name))()
	VAR_TABLE *vtab;
	char *name;
{
	VAR_TABLE *tp;
	char *cp;
	int len;

	if (!strcasecmp(name, "TIME"))
		return sprint_time;

	if ((cp = strrchr(name, '.')) != NULL && !strcasecmp(++cp, "old"))
		len = strlen(name) - 4;
	else	len = 0;

	for (tp = vtab; tp && tp->func; tp++) {
		if (!*tp->name)
			return tp->func;
		if (!len && !strcasecmp(tp->name, name))
			return tp->func;
		if (len && tp->old && !strncasecmp(tp->name, name, len))
			return tp->func;
	}
	/*
	report(LOG_DEBUG, "\"%s\": No such variable", name);
	*/
	return return_null;
}

/*
 * Insert requested variable(s) into user's format string.
 */
char *
insert_variables(dst_buf, dst_len, src, type, object)
	char *dst_buf;
	int dst_len;
	register char *src;
	int type;
	void *object;
{
	register char *dst;
	int len, parent;
	char *vb, *ve, *vt[5], var[100], *dst_save, *next;
	VAR_TABLE *vtab;
	int found_type;
	void *found_obj;

	dst = dst_buf;
	dst_len--;
next_char:
	while (src && *src && dst < dst_buf + dst_len) {
		if (*src != '$') {
			*dst++ = *src++;
			continue;
		}
		vb = src + 1;
		parent = 0;
		while (*vb == '$') {
			parent++;
			vb++;
		}
		if (*vb == '{') {
			if ((ve = strchr(++vb, '}')) == NULL) {
				*dst++ = *src++;
				continue;
			}
			len = ve - vb;
			ve++;
		} else {
			for (ve = vb; *ve && (isalnum(*ve) ||
					      *ve == '_' || *ve == '.'); ve++);
			len = ve - vb;
		}
		if (len < 2 || len > sizeof(var)-1) {
			*dst++ = *src++;
			continue;
		}
		memcpy(var, vb, len);
		var[len] = '\0';
		len = 0;
		next = var;
		while ((vb = my_strsep(&next, "!")) != NULL) {
			vt[len++] = vb;
			if (len > 3) break;
		}
		if (len < 1 || len > 3) {
			*dst++ = *src++;
			continue;
		}
		found_type = type;
		found_obj = object;

		/* old style: pointing directly to the object by name */
		if (len != 1) {
			OBJECT *target, *service;
			INTERFACE *iface;
			BGP_AS *bgp;

			found_type = 0;
			found_obj = 0;
			for (target = cf->target; target; target = target->next) {
				if (!strcasecmp(vt[0], target->name))
					break;
			}
			if (target) {
				found_type = TYPE_OBJECT;
				found_obj = target;
				if (len != 2) {
					for (iface = target->interface; iface; iface = iface->next) {
						if ((iface->name && !strcasecmp(iface->name, vt[1])) ||
						    (iface->index > 0 && iface->index == atoi(vt[1]))) {
							break;
						}
					}
					if (iface) {
						found_type = TYPE_INTERFACE;
						found_obj = iface;
					} else { /* not an iface; maybe bgp? */
						bgp = NULL;
						if (!strncmp(vt[1], "AS", 2)) {
							vb = vt[1] + 2;
							for (bgp = target->bgp; bgp; bgp = bgp->next) {
								if (bgp->asn == atoi(vb))
									break;
							}
						}
						if (bgp) {
							found_type = TYPE_BGPAS;
							found_obj = bgp;
						} else {/* or maybe service? */
							for (service = target->service; service; service = service->next) {
								if (!strcasecmp(service->name, vt[1]))
									break;
							}
							found_type = TYPE_SERVICE;
							found_obj = service;
						}
					}
				}
			}
			if (!found_type || !found_obj) {
				*dst++ = *src++;
				continue;
			}
		}

		/* reparent if required */
		while (found_type && found_obj && parent--) {
			switch (found_type) {
			case TYPE_OBJECT:
				found_type = 0;
				found_obj = 0;
				break;
			case TYPE_SERVICE:
				found_type = TYPE_OBJECT;
				found_obj = ((OBJECT *)found_obj)->parent;
				break;
			case TYPE_INTERFACE:
				found_type = TYPE_OBJECT;
				found_obj = ((INTERFACE *)found_obj)->parent;
				break;
			case TYPE_BGPAS:
				found_type = TYPE_OBJECT;
				found_obj = ((BGP_AS *)found_obj)->parent;
				break;
			case TYPE_BGPPEER:
				found_type = TYPE_BGPAS;
				found_obj = ((BGP_PEER *)found_obj)->parent;
				break;
			case TYPE_ENVMON:
				found_type = TYPE_OBJECT;
				found_obj = ((ENV_MON *)found_obj)->parent;
				break;
			case TYPE_ENVTEMP:
			case TYPE_ENVVOLT:
			case TYPE_ENVFAN:
			case TYPE_ENVPS:
				found_type = TYPE_ENVMON;
				found_obj = ((ENV_GAUGE *)found_obj)->parent;
			}
		}

		if (!found_obj) /* for sanity */
			found_type = 0;

		/* set table according type */
		switch (found_type) {
			case TYPE_OBJECT:	vtab = object_table;	break;
			case TYPE_SERVICE:	vtab = service_table;	break;
			case TYPE_INTERFACE:	vtab = iface_table;	break;
			case TYPE_BGPAS:	vtab = bgp_table;	break;
			case TYPE_BGPPEER:	vtab = peer_table;	break;
			case TYPE_ENVMON:	vtab = envmon_table;	break;
			case TYPE_ENVTEMP:
			case TYPE_ENVVOLT:
			case TYPE_ENVFAN:
			case TYPE_ENVPS:	vtab = envgauge_table;	break;
			default:		vtab = main_table;
		}

		dst_save = dst;
		next = vt[len-1];
		while ((vb = my_strsep(&next, " \t\r\n")) != NULL) {
#ifdef	HAVE_PTHREAD
			pthread_mutex_lock(&variables_lock); /* unsafe calls follow */
#endif
			vb = (*do_variable(vtab, vb))(found_obj, vb);
			if (!vb) {
#ifdef	HAVE_PTHREAD
				pthread_mutex_unlock(&variables_lock);
#endif
				dst = dst_save;
				*dst++ = *src++;
				goto next_char;
			}
			if (dst + strlen(vb) + 1 >= dst_buf + dst_len) {
#ifdef	HAVE_PTHREAD
				pthread_mutex_unlock(&variables_lock);
#endif
				*dst = '\0';
				return dst_buf;
			}
			if (!len) *dst++ = ' ';
			else len = 0;
			dst += strlen(strcpy(dst, vb));
#ifdef	HAVE_PTHREAD
			pthread_mutex_unlock(&variables_lock);
#endif
		}
		src = ve;
	}
	*dst = '\0';
	return dst_buf;
}
